@@ -1,6 +1,6 @@
 ;;;; batch-driver.scm - Driver procedure for the compiler
 ;
-; Copyright (c) 2000-2003, Felix L. Winkelmann
+; Copyright (c) 2000-2004, Felix L. Winkelmann
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -28,8 +28,8 @@
 ; felix@call-with-current-continuation.org
 ;
 ; Felix L. Winkelmann
-; Steinweg 1A
-; 37130 Gleichen, OT Weissenborn
+; Unter den Gleichen 1
+; 37130 Gleichen
 ; Germany
 
 
@@ -44,20 +44,20 @@
   installation-home compiler-cleanup-hook
   foreign-type-table-size file-io-only
   unit-name insert-timer-checks used-units
-  debugging perform-lambda-lifting!
-  posix-is-available register-ffi-macro ffi-include-path
+  debugging perform-lambda-lifting! disable-stack-overflow-checking no-c-syntax-checks
+  register-ffi-macro ffi-include-path-list
   foreign-declarations emit-trace-info block-compilation analysis-database-size line-number-database-size
   target-heap-size target-stack-size try-harder default-installation-home target-heap-growth target-heap-shrinkage
   default-default-target-heap-size default-default-target-stack-size verbose-mode original-program-size
-  target-initial-heap-size
+  target-initial-heap-size split-level default-user-pass-2
   current-program-size line-number-database-2 foreign-lambda-stubs immutable-constants foreign-variables
   rest-parameters-promoted-to-vector inline-table inline-table-used constant-table constants-used mutable-constants
   broken-constant-nodes inline-substitutions-enabled enable-sharp-greater-read-syntax
   emit-profile profile-lambda-list profile-lambda-index profile-info-vector-name
-  direct-call-ids foreign-type-table first-analysis
+  direct-call-ids foreign-type-table first-analysis optimizers-used
   initialize-compiler canonicalize-expression expand-foreign-lambda update-line-number-database scan-toplevel-assignments
   perform-cps-conversion analyze-expression simplifications perform-high-level-optimizations perform-pre-optimization!
-  reorganize-recursive-bindings substitution-table simplify-named-call
+  reorganize-recursive-bindings substitution-table simplify-named-call emit-unsafe-marker
   perform-closure-conversion prepare-for-code-generation compiler-source-file create-foreign-stub expand-foreign-lambda*
   transform-direct-lambdas! source-filename compressed-literals literal-compression-threshold
   debugging-chicken warnings-enabled bomb check-signature posq stringify symbolify flonum? build-lambda-list
@@ -67,13 +67,14 @@
   build-node-graph build-expression-tree fold-boolean inline-lambda-bindings match-node expression-has-side-effects?
   simple-lambda-node? compute-database-statistics print-program-statistics output gen gen-list 
   pprint-expressions-to-file foreign-type-check estimate-foreign-result-size scan-used-variables scan-free-variables
-  topological-sort print-version print-usage initialize-analysis-database
-  product copyright compiler-features default-declarations units-used-by-default words-per-flonum default-debugging-declarations
+  topological-sort print-version print-usage initialize-analysis-database dump-exported-globals
+  copyright compiler-features default-declarations units-used-by-default words-per-flonum default-debugging-declarations
   default-profiling-declarations default-optimization-passes compressed-literals-initializer
   foreign-string-result-reserve parameter-limit eq-inline-operator optimizable-rest-argument-operators
   membership-test-operators membership-unfold-limit valid-compiler-options valid-compiler-options-with-argument
-  chop-separator chop-extension display-real-name-table display-line-number-database
+  chop-separator chop-extension display-real-name-table display-line-number-database explicit-use-flag
   generate-code make-variable-list make-argument-list generate-foreign-stubs foreign-type-declaration
+  export-list module-export-list compute-module-exports
   foreign-argument-conversion foreign-result-conversion parse-script-header}
 
 
@@ -84,23 +85,38 @@
 ;;; Compile a complete source file:
 
 (define (compile-source-file filename . options)
+  (define (option-arg p)
+    (if (null? (cdr p))
+	(quit "missing argument to `-~A' option" (car p))
+	(cadr p) ) )
   (initialize-compiler)
+  (set! explicit-use-flag (memq 'explicit-use options))
   (let ([initforms `((##core#declare
 		      ,@(map (lambda (x) `(quote ,x))
 			     (append 
 			      default-declarations
-			      (if (memq 'explicit-use options) 
+			      (if explicit-use-flag
 				  '()
 				  `((uses ,@units-used-by-default)) ) ) ) ) ) ]
         [verbose (memq 'verbose options)]
 	[outfile (cond [(memq 'output-file options) 
 			=> (lambda (node)
-			     (let ([oname (cadr node)])
+			     (let ([oname (option-arg node)])
 			       (if (symbol? oname)
 				   (symbol->string oname)
 				   oname) ) ) ]
 		       [(memq 'to-stdout options) #f]
 		       [else (make-pathname #f (if filename (pathname-file filename) "out") "c")] ) ]
+        [partitions (cond [(memq 'split options)
+                           => (lambda (node)
+                                (let ([oname (option-arg node)])
+                                  (cond
+                                    ((symbol? oname)
+                                     (string->number (symbol->string oname)))
+                                    ((string? oname)
+                                     (string->number oname))
+                                    (else oname))))]
+                          [else 1])]
 	[home (chop-separator (or (getenv "CHICKEN_HOME") default-installation-home ""))]
 	[ipath (map chop-separator (string-split (or (getenv "CHICKEN_INCLUDE_PATH") "") ";"))]
 	[strict (memq 'strict options)]
@@ -119,16 +135,27 @@
 	[hshrink (memq 'heap-shrinkage options)]
 	[kwstyle (memq 'keyword-style options)]
 	[lcthreshold (memq 'compress-literals options)]
-	[srfi7 (memq 'srfi-7 options)]
-	[hygienic-at-rt (memq 'hygienic-at-run-time options)]
+	[hygienic (or (memq 'hygienic options) (memq 'syntax options))]
+	[hygienic-at-rt (or (memq 'hygienic-at-run-time options) (memq 'syntax-at-run-time options))]
 	[uunit (memq 'unit options)]
 	[a-only (memq 'analyze-only options)]
 	[dynamic (memq 'dynamic options)]
-	[script-extra-units #f]
 	[dumpnodes #f]
 	[ffi-mode (memq 'ffi options)]
+	[ofilelist #f]
+	[quiet (memq 'quiet options)]
 	[ssize (or (memq 'nursery options) (memq 'stack-size options))] )
 
+    (define (outfile->list outfile num)
+      (cond
+       [(not outfile) #f]
+       [(= num 1) (list outfile)]
+       [else
+	(let ([file (pathname-file outfile)])
+	  (list-tabulate 
+	   num
+	   (lambda (i) (make-pathname #f (sprintf "~A~A.c" file i))) ) ) ] ) )
+
     (define (cputime) (##sys#fudge 6))
 
     (define (print-header mode dbgmode)
@@ -177,7 +204,7 @@
 
     (define (collect-options opt)
       (let loop ([opts options])
-	(cond [(memq opt opts) => (lambda (p) (cons (cadr p) (loop (cddr p))))]
+	(cond [(memq opt opts) => (lambda (p) (cons (option-arg p) (loop (cddr p))))]
 	      [else '()] ) ) )
 
     (define (begin-time)
@@ -209,6 +236,7 @@
       (if (or s s0 strictr) 
 	  (begin
 	    (strict-reader #t)
+	    (keyword-style #:none)
 	    (case-sensitive #f) )
 	  (enable-sharp-greater-read-syntax) ) )
 
@@ -222,8 +250,10 @@
 		   (foreign-parse ,s) ) ) )
 	  (##sys#read in infohook) ) )
 
-    (when (and uunit (pair? (cdr uunit)))
-      (set! unit-name (string->c-identifier (stringify (cadr uunit)))) )
+    (when (or (< partitions 1) (not outfile))
+        (set! partitions 1))
+    (when uunit
+      (set! unit-name (string->c-identifier (stringify (option-arg uunit)))) )
     (set! debugging-chicken 
       (append-map
        (lambda (do)
@@ -242,20 +272,30 @@
     (when (memq 'unsafe options) 
       (set! unsafe #t)
       (##match#set-error-control #:unspecified) )
+    (when (memq 'unsafe-libraries options)
+      (set! emit-unsafe-marker #t) )
     (when (memq 'disable-interrupts options) (set! insert-timer-checks #f))
     (when (memq 'fixnum-arithmetic options) (set! number-type 'fixnum))
     (when (memq 'block options) (set! block-compilation #t))
+    (when (memq 'disable-c-syntax-checks options) (set! no-c-syntax-checks #t))
     (when (memq 'case-insensitive options) 
       (when verbose (printf "Identifiers and symbols are case insensitive~%~!"))
       (register-feature! 'case-insensitive)
       (case-sensitive #f) )
+    (and-let* ([slevel (memq 'split-level options)])
+      (set! split-level 
+	(let ([n (string->number (option-arg slevel))])
+	  (if (and n (<= 0 n 2))
+	      n
+	      (quit "invalid argument to `-split-level' option") ) ) ) )
     (when kwstyle
-      (cond [(string=? "prefix" (cadr kwstyle))
-	     (keyword-style #:prefix) ]
-	    [(string=? "suffix" (cadr kwstyle))
-	     (keyword-style #:suffix) ] ) )
+      (let ([val (option-arg kwstyle)])
+	(cond [(string=? "prefix" val) (keyword-style #:prefix)]
+	      [(string=? "none" val) (keyword-style #:none)]
+	      [(string=? "suffix" val) (keyword-style #:suffix)]
+	      [else (quit "invalid argument to `-keyword-style' option")] ) ) )
     (when lcthreshold
-      (let ([t (cadr lcthreshold)])
+      (let ([t (option-arg lcthreshold)])
 	(set! literal-compression-threshold 
 	  (or (string->number t)
 	      (quit "invalid argument to `-compress-literals' option: ~A" t) ) ) ) )
@@ -266,7 +306,8 @@
 	      ##sys#include-pathnames
 	      ipath
 	      (if home (list home) '()) ) )
-    (when (and outfile filename (string=? outfile filename))
+    (set! ofilelist (outfile->list outfile partitions))
+    (when (and outfile filename (any (cut string=? <> filename) ofilelist))
       (quit "source- and output-filename are the same") )
 
     ;; Handle feature options:
@@ -275,25 +316,26 @@
 
     ;; Handle FFI defines and include-paths:
     (for-each register-ffi-macro (collect-options 'ffi-define))
-    (set! ffi-include-path (collect-options 'ffi-include-path))
+    (set! ffi-include-path-list (append (collect-options 'ffi-include-path) ffi-include-path-list))
 
     ;; Load extensions:
     (set! ##sys#features (cons #:compiler-extension ##sys#features))
     (let ([extends (collect-options 'extend)])
       (when verbose
-	(printf "Loading extensions...~%~!")
+	(printf "Loading compiler extensions...~%~!")
 	(load-verbose #t) )
-      (for-each (lambda (f) (load (##sys#resolve-include-filename f))) extends) )
+      (for-each (lambda (f) (load (##sys#resolve-include-filename f #t))) extends) )
     (set! ##sys#features (delete #:compiler-extension ##sys#features eq?))
 
     (set! ##sys#features (cons '#:compiling ##sys#features))
     ;; Install highlevel macros:
-    (if (memq 'hygienic options)
+    (if hygienic
 	(begin
 	  (when verbose 
-	    (printf "Using hygienic macros~!") )
+	    (printf "Using hygienic macros~%~!") )
 	  (install-macros) )
 	(begin
+	  (set! ##sys#features (cons #:match (cons #:match-support ##sys#features)))
 	  (##sys#provide 'match) 
 	  (when (or strict strict-srfi0 strict-letrec) (##sys#enable-strict-letrec) ) ) )
 
@@ -309,25 +351,26 @@
 	  (set! initforms
 	    (append initforms (map (lambda (r) `(##sys#require ',r)) rs)) ) ) ) )
 
-    (when (memq 'compile-time-macros-only options)
-      (set! ##sys#enable-runtime-macros #f) )
+    (when (memq 'run-time-macros options)
+      (set! ##sys#enable-runtime-macros #t) )
     (strictness strict strict-srfi0)
     (set! target-heap-size
       (if hsize
-	  (arg-val (cadr hsize))
+	  (arg-val (option-arg hsize))
 	  (and-let* ([hsize default-default-target-heap-size]
 		     [(not (zero? hsize))] )
 	    hsize) ) )
-    (set! target-initial-heap-size (and hisize (arg-val (cadr hisize))))
-    (set! target-heap-growth (and hgrowth (arg-val (cadr hgrowth))))
-    (set! target-heap-shrinkage (and hshrink (arg-val (cadr hshrink))))
+    (set! target-initial-heap-size (and hisize (arg-val (option-arg hisize))))
+    (set! target-heap-growth (and hgrowth (arg-val (option-arg hgrowth))))
+    (set! target-heap-shrinkage (and hshrink (arg-val (option-arg hshrink))))
     (set! target-stack-size
       (if ssize
-	  (arg-val (cadr ssize))
+	  (arg-val (option-arg ssize))
 	  (and-let* ([ssize default-default-target-stack-size]
 		     [(not (zero? ssize))] )
 	    ssize) ) )
     (set! emit-trace-info (not (memq 'no-trace options)))
+    (set! disable-stack-overflow-checking (memq 'disable-stack-overflow-checks options))
     (when (memq 'm debugging-chicken) (set-gc-report! #t))
     (when (memq 'usual-integrations options)
       (set! standard-bindings default-standard-bindings)
@@ -344,16 +387,17 @@
 	(printf "Generating profile~%~!" emit-profile) ) )
 
     (cond ((memq 'version options)
-	   (print-version)
+	   (print-version #t)
 	   (newline) )
-	  ((memq 'help options) (print-usage))
+	  ((or (memq 'help options) (memq '-help options)) (print-usage))
 	  ((not filename)
-	   (print-version)
-	   (display "\n\nEnter \"chicken -help\" for information on how to use it.\n") )
+	   (unless quiet
+	     (print-version #t)
+	     (display "\n\nEnter \"chicken -help\" for information on how to use it.\n") ) )
 	  (else
 
 	   ;; Display header:
-	   (unless (memq 'quiet options)
+	   (unless quiet
 	     (printf "compiling `~a' ...~%" filename) )
 	   (set! source-filename filename)
 	   (debugging 'r "options" options)
@@ -392,45 +436,43 @@
 			  (cond [ffi-mode]
 				[(char=? #\# c0)
 				 (read-char in)
-				 (unless (char=? #\! (peek-char in))
-				   (quit "source file may not begin with `#' character if not a script") )
-				 (read-char in)
-				 (let ([header (open-output-string)]
-				       [cleanup #t] )
-				   (do ([c (read-char in) (read-char in)])
-				       ((char=? c #\newline)
-					(set! ##sys#read-line-counter (add1 ##sys#read-line-counter)) )
-				     (if (eof-object? c)
-					 (quit "unexpected end of file - incorrect header")
-					 (write-char c header) ) )
-				   (when verbose (printf "Compiling UNIX script~%~!"))
-				   (case (parse-script-header (get-output-string header) in)
-				     [(script)
-				      (set! cleanup #f)
-				      (set! script-extra-units #t) ]
-				     [(chicken) (set! script-extra-units #t)]
-				     [(chicken-hygienic) 
-				      (set! script-extra-units #t)
-				      (install-macros) ]
-				     [(srfi7) 
-				      (set! srfi7 #t)
-				      (strictness #t #f)
-				      (install-macros) ]
-				     [(r5rs)
-				      (install-macros)
-				      (strictness #t #f) ]
-				     [(srfi0)
-				      (install-macros)
-				      (strictness #t #t) ]
-				     [(#f) (warning "unrecognized script header - might be corrupt")]
-				     [else (strictness #t #f)] ) ; r4rs & ieee
-				   (when cleanup
-				     (set! cleanup-forms
-				       (cons '(##sys#script-main '0 '0 main)
-					     cleanup-forms) ) ) ) ]
+				 (if (not (char=? #\! (peek-char in)))
+				     (if (string=? "-" f)
+					 (quit "source input from pipe may not begin with `#' character if not a script") 
+					 (begin
+					   (close-input-port in)
+					   (set! in (check-and-open-input-file f)) ) )
+				     (begin
+				       (read-char in)
+				       (let ([header (open-output-string)]
+					     [cleanup #t] )
+					 (do ([c (read-char in) (read-char in)])
+					     ((char=? c #\newline)
+					      (set! ##sys#read-line-counter (add1 ##sys#read-line-counter)) )
+					   (if (eof-object? c)
+					       (quit "unexpected end of file - incorrect header")
+					       (write-char c header) ) )
+					 (when verbose (printf "Compiling UNIX script~%~!"))
+					 (case (parse-script-header (get-output-string header) in)
+					   [(script)
+					    (set! cleanup #f) ]
+					   [(chicken)]
+					   [(chicken-hygienic) 
+					    (install-macros) ]
+					   [(r5rs)
+					    (install-macros)
+					    (strictness #t #f) ]
+					   [(srfi0)
+					    (install-macros)
+					    (strictness #t #t) ]
+					   [(#f) (warning "unrecognized script header - might be corrupt")]
+					   [else (strictness #t #f)] ) ; r4rs & ieee
+					 (when cleanup
+					   (set! cleanup-forms
+					     (cons '(##sys#script-main '0 '0 main)
+						   cleanup-forms) ) ) ) ) ) ]
 				[(char=? #\@ c0)
 				 (read-line in)
-				 (set! script-extra-units #t)
 				 (when verbose (printf "Compiling Windows batch file~%~!")) ] )
 
 			  (do ([x (read-form in) (read-form in)])
@@ -438,33 +480,12 @@
 			       (close-checked-input-file in f) )
 			    (set! forms (cons x forms)) ) ) ) ] ) ) )
 
-	   ;; Visit files:
-	   (for-each visit (collect-options 'visit))
-
 	   ;; Start compilation passes:
 	   (let ([proc (user-preprocessor-pass)])
 	     (when proc
 	       (when verbose (printf "User preprocessing pass...~%~!"))
 	       (set! forms (cons (first forms) (map proc (cdr forms)))) ) )
 
-	   (when srfi7
-	     (set! forms
-	       (list
-		(##sys#expand-srfi-7-program 
-		 (chop-separator filename)
-		 (car forms)
-		 ##sys#syntax-error-hook) ) ) )
-
-	   (set! script-extra-units
-	     (if script-extra-units
-		 (map (compose string->c-identifier stringify)
-		      (append
-		       '(extras srfi-1 srfi-4 lolevel srfi-18 srfi-37 format tinyclos regex script-utils tcp)
-		       (if posix-is-available
-			   '(posix)
-			   '() ) ) )
-		 '() ) )
-
 	   (print "source" '|1| forms)
 	   (begin-time)
 	   (let ([us (collect-options 'uses)])
@@ -475,10 +496,7 @@
 		  [plen (length profile-lambda-list)]
 		  [exps (append
 			 (map (lambda (ic) `(set! ,(cdr ic) ',(car ic))) immutable-constants)
-			 (map (lambda (n) `(##core#callunit ,n))
-			      (begin
-				(set! used-units (append used-units script-extra-units))
-				used-units) ) 
+			 (map (lambda (n) `(##core#callunit ,n)) used-units)
 			 (if hygienic-at-rt
 			     '((##core#callunit "syntax_case")
 			       (##syncase#install-macro-package '#t '#t) )
@@ -497,7 +515,9 @@
 				  ',(cdr pl) ) )
 			      profile-lambda-list)
 			 (let ([is (fold (lambda (clf r)
-					   `(let ([,(gensym) (set! ,(car clf) (##sys#read-from-string ',(cdr clf)))])
+					   `(let ([,(gensym) 
+						   (set! ,(car clf)
+						     (##sys#read-from-string ',(cdr clf)))])
 					      ,r) )
 					 '(##core#undefined) 
 					 compressed-literals) ] )
@@ -506,11 +526,14 @@
 					      (lambda () ,is) ) )
 			       (list is) ) )
 			 exps0
+			 (if hygienic
+			     (compute-module-exports module-export-list export-list)
+			     '() )
 			 (if (and (not unit-name) (not dynamic))
 			     cleanup-forms
 			     '() ) ) ] )
 
-	     (when (debugging '|N| "real name table:")
+   	     (when (debugging '|N| "real name table:")
 	       (display-real-name-table) )
 	     (when (debugging 'n "line number database:")
 	       (display-line-number-database) )
@@ -543,7 +566,7 @@
 			    'lambda '(())
 			    (list (build-node-graph
 				   (canonicalize-begin-body exps) ) ) ) ] 
-		    [proc (user-pass-2)] )
+		    [proc (or (user-pass-2) (and optimizers-used default-user-pass-2))] )
 	       (when proc
 		 (when verbose (printf "Secondary user pass...~%"))
 		 (begin-time)
@@ -552,9 +575,12 @@
 		   (print-db "analysis (u)" '|0| db 0)
 		   (end-time "pre-analysis (u)")
 		   (begin-time)
-		   (proc node0 db)
+		   (proc node0 
+			 (lambda (k p) (cut get db <> <>))
+			 (lambda (k p x) (cut put! db <> <>)) )
 		   (end-time "secondary user pass")
-		   (print-node "secondary user pass" '|U| node0) ) )
+		   (print-node "secondary user pass" '|U| node0) )
+		 (set! first-analysis #t) )
 
 	       (when (memq 'lambda-lift options)
 		 (begin-time)
@@ -567,6 +593,7 @@
 		   (end-time "lambda lifting")
 		   (print-node "lambda lifted" '|L| node0) )
 		 (set! first-analysis #t) )
+
 	       (set! ##sys#line-number-database #f)
 	       (set! constant-table #f)
 	       (set! inline-table #f)
@@ -619,6 +646,8 @@
 			    (let ([node3 (perform-closure-conversion node2 db)])
 			      (end-time "closure conversion")
 			      (print-db "final-analysis" '|8| db i)
+			      (when (debugging 'e "exported toplevel variables:")
+				(dump-exported-globals db) )
 			      (let ([upap (user-post-analysis-pass)])
 				(when upap 
 				  (upap db
@@ -627,20 +656,33 @@
 			      (when a-only (exit 0))
 			      (print-node "closure-converted" '|9| node3)
 
+                              (when (and verbose
+					 outfile
+					 (not (> partitions 1)) )
+                                (printf "files to be generated: ~A~%" (string-intersperse ofilelist ", ")) )
+                              
 			      (begin-time)
-			      (receive (node literals lambdas) (prepare-for-code-generation node3 db)
+			      (receive (node literals lambdas) (prepare-for-code-generation node3 db partitions)
 				(end-time "preparation")
 
-				(let ([out (if outfile (open-output-file outfile) (current-output-port))])
-				  (begin-time)
-				  (generate-code literals lambdas out filename db)
-				  (end-time "code generation")
-
-				  (when outfile (close-output-port out))
-				  (when (memq 't debugging-chicken) (##sys#display-times (##sys#stop-timer)))
-				  (compiler-cleanup-hook)
-				  (when verbose 
-				    (printf "compilation finished.~%~!") ) ) ) ) ] ) ) ) ) ) ) ) ) ) )
+                                (begin-time)
+                                (let loopfile [(lof ofilelist)
+                                               (file-partition 0)]
+                                  (let* ([outfile (and lof (car lof))]
+					 [out (if outfile (open-output-file outfile) (current-output-port))] )
+				    (unless quiet
+				      (printf "generating `~A' ...~%" outfile) )
+                                    (generate-code 
+				     literals lambdas out filename dynamic db
+				     (if (= partitions 1) #f file-partition))
+                                    (when outfile (close-output-port out))
+                                    (when (and lof (pair? (cdr lof)))
+                                      (loopfile (cdr lof) (+ file-partition 1)))))
+                                (end-time "code generation")
+                                (when (memq 't debugging-chicken) (##sys#display-times (##sys#stop-timer)))
+                                (compiler-cleanup-hook)
+                                (when verbose 
+                                  (printf "compilation finished.~%~!") ) ) ) ] ) ) ) ) ) ) ) ) ) )
 
 
 ;;; Parse script header:
@@ -656,11 +698,8 @@
 	  [(cmp-suffix "-script-meta")
 	   (read-line in)
 	   'script]
-	  [(cmp-suffix "scheme-chicken") 'chicken]
-	  [(cmp-suffix "scheme-chicken-hygienic") 'chicken-hygienic]
 	  [(cmp-suffix "scheme-r4rs") 'r4rs]
 	  [(cmp-suffix "scheme-r5rs") 'r5rs]
-	  [(cmp-suffix "scheme-srfi-7") 'srfi7]
 	  [(cmp-suffix "scheme-srfi-0") 'srfi0]
 	  [(cmp-suffix "scheme-ieee-1178-1990") 'ieee]
 	  [else #f] ) ) )
