@@ -1,6 +1,6 @@
 ;;; c-backend.scm - C-generating backend for the CHICKEN compiler
 ;
-; Copyright (c) 2000-2003, Felix L. Winkelmann
+; Copyright (c) 2000-2004, Felix L. Winkelmann
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -28,8 +28,8 @@
 ; felix@call-with-current-continuation.org
 ;
 ; Felix L. Winkelmann
-; Steinweg 1A
-; 37130 Gleichen, OT Weissenborn
+; Unter den Gleichen 1
+; 37130 Gleichen
 ; Germany
 
 
@@ -39,39 +39,47 @@
 #{compiler
   build-information compiler-arguments process-command-line
   default-standard-bindings default-extended-bindings side-effecting-standard-bindings
-  non-foldable-standard-bindings foldable-standard-bindings non-foldable-extended-bindings foldable-extended-bindings
+  non-foldable-standard-bindings foldable-standard-bindings non-foldable-extended-bindings 
+  foldable-extended-bindings
   standard-bindings-that-never-return-false side-effect-free-standard-bindings-that-never-return-false
   installation-home optimization-iterations debugging cleanup
-  foreign-type-table-size file-io-only
+  foreign-type-table-size file-io-only namespace-table
   unit-name insert-timer-checks used-units inlining external-variables
   foreign-declarations emit-trace-info block-compilation analysis-database-size line-number-database-size
   target-heap-size target-stack-size try-harder default-installation-home target-heap-growth target-heap-shrinkage
   default-default-target-heap-size default-default-target-stack-size verbose-mode original-program-size
-  target-initial-heap-size
+  target-initial-heap-size disable-stack-overflow-checking
   current-program-size line-number-database-2 foreign-lambda-stubs immutable-constants
-  rest-parameters-promoted-to-vector inline-table inline-table-used constant-table constants-used mutable-constants
-  broken-constant-nodes inline-substitutions-enabled
+  rest-parameters-promoted-to-vector inline-table inline-table-used constant-table constants-used 
+  mutable-constants
+  broken-constant-nodes inline-substitutions-enabled data-declarations
   direct-call-ids foreign-type-table first-analysis block-variable-literal?
-  initialize-compiler canonicalize-expression expand-foreign-lambda update-line-number-database scan-toplevel-assignments
-  perform-cps-conversion analyze-expression simplifications perform-high-level-optimizations perform-pre-optimization!
+  initialize-compiler canonicalize-expression expand-foreign-lambda update-line-number-database 
+  scan-toplevel-assignments
+  perform-cps-conversion analyze-expression simplifications perform-high-level-optimizations 
+  perform-pre-optimization!
   reorganize-recursive-bindings substitution-table simplify-named-call find-inlining-candidates perform-inlining!
-  perform-closure-conversion prepare-for-code-generation compiler-source-file create-foreign-stub expand-foreign-lambda*
-  transform-direct-lambdas! target-include-file
+  perform-closure-conversion prepare-for-code-generation compiler-source-file create-foreign-stub 
+  expand-foreign-lambda*
+  transform-direct-lambdas! target-include-file emit-unsafe-marker
   debugging-chicken warnings-enabled bomb check-signature posq stringify symbolify flonum? build-lambda-list
   string->c-identifier c-ify-string words check-and-open-input-file close-checked-input-file fold-inner constant?
   collapsable-literal? immediate? canonicalize-begin-body extract-mutable-constants string->expr get get-all
   put! collect! count! get-line get-line-2 find-lambda-container display-analysis-database varnode qnode 
-  build-node-graph build-expression-tree fold-boolean inline-lambda-bindings match-node expression-has-side-effects?
+  build-node-graph build-expression-tree fold-boolean inline-lambda-bindings match-node 
+  expression-has-side-effects?
   simple-lambda-node? compute-database-statistics print-program-statistics output gen gen-list 
-  pprint-expressions-to-file foreign-type-check estimate-foreign-result-size scan-used-variables scan-free-variables
+  pprint-expressions-to-file foreign-type-check estimate-foreign-result-size scan-used-variables 
+  scan-free-variables
   topological-sort print-version print-usage initialize-analysis-database
-  generate-external-variables real-name real-name2
-  product copyright compiler-features default-declarations units-used-by-default words-per-flonum
+  generate-external-variables real-name real-name2 unique-id
+  copyright compiler-features default-declarations units-used-by-default words-per-flonum
   foreign-string-result-reserve parameter-limit eq-inline-operator optimizable-rest-argument-operators
   membership-test-operators membership-unfold-limit valid-compiler-options valid-compiler-options-with-argument
   default-optimization-iterations generate-foreign-callback-header generate-foreign-callback-stub-prototypes
   generate-code make-variable-list make-argument-list generate-foreign-stubs foreign-type-declaration
-  foreign-argument-conversion foreign-result-conversion}
+  foreign-argument-conversion foreign-result-conversion quick-namespace-list setup-quick-namespace-list
+  namespace-lookup compute-namespace-size}
 
 (include "tweaks")
 
@@ -93,10 +101,38 @@
    (lambda (x) (display x output))
    (intersperse lst #\space) ) )
 
+(define (compute-namespace-size n)
+  37)					;*** FIX
+
+
+;;; Unique id/prefix:
+
+(define unique-id
+  (string->c-identifier
+   (sprintf "C_~X_~A_" (random #x1000000) (current-seconds)) ) )
+
+
+;;; Check name for namespace:
+
+(define quick-namespace-list '())
+
+(define (setup-quick-namespace-list)
+  (for-each 
+   (lambda (ns)
+     (set! quick-namespace-list (append (cdr ns) quick-namespace-list)) )
+   namespace-table) )
+
+(define (namespace-lookup sym)
+  (and (memq sym quick-namespace-list)
+       (let loop ([nslist namespace-table] [i 0])
+	 (cond [(null? nslist) (bomb "symbol not in namespace" sym)]
+	       [(memq sym (cdar nslist)) i]
+	       [else (loop (cdr nslist) (add1 i))] ) ) ) )
+
 
 ;;; Generate target code:
 
-(define (generate-code literals lambdas out source-file db)
+(define (generate-code literals lambdas out source-file dynamic db file-partition)
 
   (define (find-lambda id)
     (or (find (lambda (ll) (eq? id (lambda-literal-id ll))) lambdas)
@@ -104,6 +140,9 @@
 
   (define (slashify s) (string-translate (->string s) "\\" "/"))
   
+  (define (prefix-id)
+    (if file-partition unique-id "") )	
+
   (define (expression node temps ll)
 
     (define (expr n i)
@@ -117,9 +156,10 @@
 	     ((char) (gen "C_make_character(" (char->integer (second params)) #\)))
 	     ((nil) (gen "C_SCHEME_END_OF_LIST"))
 	     ((fix) (gen "C_fix(" (second params) #\)))
+	     ((eof) (gen "C_SCHEME_END_OF_FILE"))
 	     (else (bomb "bad immediate")) ) )
 
-	  ((##core#literal) (gen "lf[" (first params) #\]))
+	  ((##core#literal) (gen (prefix-id) "lf[" (first params) #\]))
 
 	  ((if)
 	   (gen #t "if(C_truep(")
@@ -130,7 +170,10 @@
 	   (expr (caddr subs) i)
 	   (gen #\}) )
 
-	  ((##core#proc) (gen "(C_word)" (first params)))
+	  ((##core#proc)
+	   (gen "(C_word)"
+		(if (and (pair? (cdr params)) (cadr params)) "" (prefix-id))
+		(first params)) )
 
 	  ((##core#bind) 
 	   (let loop ((bs subs) (i i) (count (first params)))
@@ -207,17 +250,17 @@
 		 [block (third params)] )
 	     (cond [block
 		    (if safe
-			(gen "lf[" index "]")
-			(gen "C_retrieve2(lf[" index "]," (c-ify-string (symbol->string (fourth params))) #\)) ) ]
-		   [safe (gen "*((C_word*)lf[" index "]+1)")]
-		   [else (gen "C_retrieve(lf[" index "])")] ) ) )
+			(gen (prefix-id) "lf[" index "]")
+			(gen "C_retrieve2(" (prefix-id) "lf[" index "]," (c-ify-string (symbol->string (fourth params))) #\)) ) ]
+		   [safe (gen "*((C_word*)" (prefix-id) "lf[" index "]+1)")]
+		   [else (gen "C_retrieve(" (prefix-id) "lf[" index "])")] ) ) )
 
 	  ((##core#setglobal)
 	   (let ([index (first params)]
 		 [block (second params)] )
 	     (if block
-		 (gen "C_mutate(&lf[" index "],")
-		 (gen "C_mutate((C_word*)lf[" index "]+1,") )
+		 (gen "C_mutate(&" (prefix-id) "lf[" index "],")
+		 (gen "C_mutate((C_word*)" (prefix-id) "lf[" index "]+1,") )
 	     (expr (car subs) i)
 	     (gen #\)) ) )
 
@@ -225,11 +268,11 @@
 	   (let ([index (first params)]
 		 [block (second params)] )
 	     (cond [block
-		    (gen "lf[" index "]=")
+		    (gen (prefix-id) "lf[" index "]=")
 		    (expr (car subs) i)
 		    (gen #\;) ]
 		   [else
-		    (gen "C_set_block_item(lf[" index "],0,")
+		    (gen "C_set_block_item(" (prefix-id) "lf[" index "],0,")
 		    (expr (car subs) i)
 		    (gen #\)) ] ) ) )
 
@@ -248,7 +291,13 @@
 		  (fn (car subs)) )
 	     (when (and emit-trace-info name) (gen #t "C_trace(\"" (slashify name) "\");"))
 	     (cond ((eq? '##core#proc (node-class fn))
-		    (gen #t (first (node-parameters fn)) #\( nf ",0,") 
+		    (let ([fpars (node-parameters fn)])
+		      (gen #t 
+			   (if (and (pair? (cdr fpars)) (cadr fpars))
+			       ""
+			       (prefix-id) )
+			   (first fpars)
+			   #\( nf ",0,") )
 		    (expr-args args i)
 		    (gen ");") )
 		   (call-id
@@ -272,7 +321,7 @@
 			     (gen #t #\t nc #\=)
 			     (expr fn i)
 			     (gen #\;) )
-			   (gen #t call-id #\()
+			   (gen #t (prefix-id) call-id #\()
 			   (unless customizable (gen nf #\,))
 			   (unless empty-closure (gen #\t nc #\,))
 			   (expr-args args i)
@@ -309,7 +358,7 @@
 		       ts (iota n 1 1) )
 		      (gen #t "goto loop;") ) )
 		   (else
-		    (gen call-id #\()
+		    (gen (prefix-id) call-id #\()
 		    (unless empty-closure (gen "t0,"))
 		    (expr-args subs i)
 		    (gen #\)) ) ) ) )
@@ -324,7 +373,7 @@
 		  (allocating (not (zero? demand)))
 		  (empty-closure (zero? (lambda-literal-closure-size (find-lambda call-id))))
 		  (fn (car subs)) )
-	     (gen call-id #\()
+	     (gen (prefix-id) call-id #\()
 	     (when allocating 
 	       (gen "C_a_i(&a," demand #\))
 	       (when (or (not empty-closure) (pair? args)) (gen #\,)) )
@@ -418,25 +467,32 @@
        args) )
 
     (expr node temps) )
-    
+  
   (define (header)
-    (gen "/* Generated from " source-file " by " product #t
-	 "   " build-information #t
-	 "   command line: ")
-    (gen-list compiler-arguments)
-    (gen #t)
-    (cond [unit-name (gen "   unit: " unit-name)]
-	  [else 
-	   (gen "   used units: ")
-	   (gen-list used-units) ] )
-    (unless unit-name
-      (gen #t "   default installation home: " (or default-installation-home "not specified") #t)
-      (gen "   default heap size: " default-default-target-heap-size #t)
-      (gen "   default nursery (stack) size: " default-default-target-stack-size) )
-    (gen #t "*/" #t #t "#include \"" target-include-file "\"")
-    (when (pair? foreign-declarations)
-      (gen #t)
-      (for-each (lambda (decl) (gen #t decl)) foreign-declarations) ) )
+    (define (pad0 n)
+      (if (< n 10)
+	  (string-append "0" (number->string n))
+	  n) )
+    (match (##sys#decode-seconds (current-seconds) #f)
+      [#(_ min hour mday mon year _ _ _ _)
+       (gen "/* Generated from " source-file " by the Chicken compiler" #t
+	    "   " (+ 1900 year) #\- (pad0 (add1 mon)) #\- (pad0 mday) #\space (pad0 hour) #\: (pad0 min) #t
+	    "   " build-information #t
+	    "   command line: ")
+       (gen-list compiler-arguments)
+       (gen #t)
+       (cond [unit-name (gen "   unit: " unit-name)]
+	     [else 
+	      (gen "   used units: ")
+	      (gen-list used-units) ] )
+       (unless unit-name
+	 (gen #t "   default installation home: " (or default-installation-home "not specified") #t)
+	 (gen "   default heap size: " default-default-target-heap-size #t)
+	 (gen "   default nursery (stack) size: " default-default-target-stack-size) )
+       (gen #t "*/" #t #t "#include \"" target-include-file "\"")
+       (when (pair? foreign-declarations)
+	 (gen #t)
+	 (for-each (lambda (decl) (gen #t decl)) foreign-declarations) ) ] ) ) 
   
   (define (trailer)
     (gen #t "/* end of file */" #t) )
@@ -447,7 +503,13 @@
       (for-each 
        (lambda (uu) (gen #t "C_externimport void C_" uu "_toplevel(C_word c,C_word d,C_word k) C_noret;"))
        used-units)
-      (unless (zero? n) (gen #t #t "static C_word lf[" n "];")) ) )
+      (unless (zero? n)
+        (cond [(not file-partition)
+               (gen #t #t "static C_TLS C_word lf[" n"];")]
+              [(zero? file-partition)
+               (gen #t #t "C_TLS C_word " (prefix-id) "lf[" n"];")]
+              [else
+                (gen #t #t "C_externimport C_TLS C_word " (prefix-id) "lf[" n"];")]))))
   
   (define (prototypes)
     (let ([large-signatures '()])
@@ -459,6 +521,10 @@
 		[empty-closure (and customizable (zero? (lambda-literal-closure-size ll)))]
 		[varlist (intersperse (make-variable-list (if empty-closure (sub1 n) n) "t") #\,)]
 		[id (lambda-literal-id ll)]
+                [partition (lambda-literal-partition ll)]
+                [in-partition (or (not file-partition) (= file-partition partition))]
+                [in-partition* (and in-partition file-partition)]
+                [out-partition (and file-partition (not (= file-partition partition)))]
 		[rest (lambda-literal-rest-argument ll)]
 		[rest-mode (lambda-literal-rest-argument-mode ll)]
 		[direct (lambda-literal-direct ll)] 
@@ -472,12 +538,26 @@
 		(set! large-signatures (lset-adjoin = large-signatures (add1 s))) ) )
 	    (lambda-literal-callee-signatures ll) )
 	   (cond [(not (eq? 'toplevel id))
-		  (gen "static " (if direct "C_word " "void "))
+		  (cond [in-partition* (gen "C_externexport ")]
+			[in-partition (gen "static ")]                        
+			[else (gen "C_externimport ")])
+		  (gen (if direct "C_word " "void "))
 		  (when customizable (gen "C_fcall "))
-		  (gen id) ]
+		  (gen (prefix-id) id) ]
 		 [else
-		  (gen "C_externexport void ")
-		  (gen "C_" (if unit-name (string-append unit-name "_toplevel") "toplevel")) ] )
+		  (unless (null? data-declarations)
+		    (gen "C_externexport C_char *C_module_data_hook(C_char *tag) {")
+		    (for-each
+		     (lambda (ddef)
+		       (gen #t "if(!C_strcmp(tag, " (c-ify-string (car ddef)) "))"
+			    #t "return " (c-ify-string (cdr ddef)) ";") )
+		     data-declarations)
+		    (gen " }" #t #t) )
+ 		  (when (and emit-unsafe-marker (or (not file-partition) (zero? file-partition)))
+  		    (gen "C_externexport void C_dynamic_and_unsafe(void) {}" #t) )
+		  (cond [in-partition (gen "C_externexport void ")]
+			[else (gen "C_externimport void ")])
+		  (gen "C_" (if unit-name (string-append unit-name "_toplevel") "toplevel") ) ] )
 	   (gen #\()
 	   (unless customizable (gen "C_word c,"))
 	   (when (and direct (not (zero? allocated)))
@@ -488,12 +568,12 @@
 		  (gen ",...) C_noret;")
 		  (if (not (eq? rest-mode 'none))
 		      (begin
-			(gen #t "static void " id "r(")
+			(gen #t "static void " (prefix-id) id "r(")
 			(apply gen varlist)
 			(gen ",C_word t" (+ n 1) ") C_noret;") ) ) ]
 		 [else 
 		  (gen #\))
-		  (when customizable (gen " C_c_regparm"))
+		  ;(when customizable (gen " C_c_regparm"))
 		  (unless direct (gen " C_noret"))
 		  (gen #\;) ] ) ) )
        lambdas) 
@@ -521,7 +601,7 @@
 	  (gen #t #t "static void C_fcall tr" n #\r)
 	  (when vflag (gen #\v))
 	  (gen "(C_proc" n " k) C_regparm C_noret;")
-	  (gen #t "static void C_fcall tr" n #\r)
+	  (gen #t "C_regparm static void C_fcall tr" n #\r)
 	  (when vflag (gen #\v))
 	  (gen "(C_proc" n " k){"
 	       #t "int n;"
@@ -550,9 +630,9 @@
 	   (unless (lambda-literal-direct ll)
 	     (cond [customizable
 		    (gen #t #t "static void C_fcall tr" id "(void *dummy) C_regparm C_noret;")
-		    (gen #t "static void C_fcall tr" id "(void *dummy){")
+		    (gen #t "C_regparm static void C_fcall tr" id "(void *dummy){")
 		    (restore argc)
-		    (gen #t id #\()
+		    (gen #t (prefix-id) id #\()
 		    (let ([al (make-argument-list argc "t")])
 		      (apply gen (intersperse al #\,)) )
 		    (gen ");}") ]
@@ -566,7 +646,7 @@
       (for-each
        (lambda (n)
 	 (gen #t #t "static void C_fcall tr" n "(C_proc" n " k) C_regparm C_noret;")
-	 (gen #t "static void C_fcall tr" n "(C_proc" n " k){")
+	 (gen #t "C_regparm static void C_fcall tr" n "(C_proc" n " k){")
 	 (restore n)
 	 (gen #t "(k)(" n #\,)
 	 (apply gen (intersperse (make-argument-list n "t") #\,))
@@ -579,7 +659,7 @@
     (do ([i 0 (+ i 1)]
 	 [lits literals (cdr lits)] )
 	((null? lits))
-      (gen-lit (car lits) (sprintf "lf[~s]" i) #t) ) )
+      (gen-lit (car lits) (sprintf "~Alf[~s]" (prefix-id) i) #t) ) )
 
   (define (bad-literal lit)
     (bomb "type of literal not supported" lit) )
@@ -616,7 +696,7 @@
 	  ((number? lit)
 	   (cond [(eq? 'fixnum number-type)
 		  (let ([flit (##sys#flo2fix lit)])
-		    (warning "coerced inexact literal number ~S to fixnum ~S" lit flit)
+		    (warning "coerced inexact literal number `~S' to fixnum `~S'" lit flit)
 		    (gen #t to "=C_fix(" flit ");") ) ]
 		 [else (gen #t to "=C_flonum(C_heaptop," lit ");")] ) )
           ((boolean? lit) 
@@ -646,11 +726,16 @@
           ((symbol? lit)
 	   (let* ([str (##sys#slot lit 1)]
 		  [cstr (c-ify-string str)]
-		  [len (##sys#size str)] )
+		  [len (##sys#size str)] 
+		  [nsi (namespace-lookup lit)] )
 	     (gen #t to "=")
-	     (if lf
-		 (gen "C_h_intern(&" to #\, len #\, cstr ");")
-		 (gen "C_intern(C_heaptop," len #\, cstr ");") ) ) )
+	     (if nsi
+		 (if lf
+		     (gen "C_h_intern_in(&" to #\, len #\, cstr ",stable" nsi ");")
+		     (gen "C_intern_in(C_heaptop," len #\, cstr ",stable" nsi ");") )
+		 (if lf
+		     (gen "C_h_intern(&" to #\, len #\, cstr ");")
+		     (gen "C_intern(C_heaptop," len #\, cstr ");") ) ) ) )
 	  ((##sys#immediate? lit) (bad-literal lit))
 	  ((##sys#bytevector? lit)
 	   (if (##sys#permanent? lit)
@@ -697,6 +782,10 @@
      (lambda (ll)
        (let* ([n (lambda-literal-argument-count ll)]
 	      [id (lambda-literal-id ll)]
+              [partition (lambda-literal-partition ll)]
+              [in-partition (or (not file-partition) (= file-partition partition))]
+              [in-partition* (and in-partition file-partition)]
+              [out-partition (and file-partition (not (= file-partition partition)))]
 	      [rname (real-name id db)]
 	      [demand (lambda-literal-allocated ll)]
 	      [rest (lambda-literal-rest-argument ll)]
@@ -719,15 +808,18 @@
 	 (gen #t #t)
 	 (gen "/* " (cleanup rname) " */" #t)
 	 (cond [(not (eq? 'toplevel id)) 
-		(gen "static " (if direct "C_word " "void "))
+                (cond [in-partition* (gen "C_externexport ")]
+                      [in-partition (gen "static ")]                        
+                      [else (gen "C_externimport ")])
+                (gen (if direct "C_word " "void "))
 		(when customizable (gen "C_fcall "))
-		(gen id) ]
+		(gen (prefix-id) id) ]
 	       [else
-		(gen "static int toplevel_initialized=0;")
+		(gen "static C_TLS int toplevel_initialized=0;")
 		(unless unit-name
 		  (gen #t "C_main_entry_point") )
 		(gen #t "static void C_fcall toplevel_trampoline(void *dummy) C_regparm C_noret;"
-		     #t "static void C_fcall toplevel_trampoline(void *dummy){"
+		     #t "C_regparm static void C_fcall toplevel_trampoline(void *dummy){"
 		     #t "C_" topname "(2,C_SCHEME_UNDEFINED,C_restore);}"
 		     #t #t "void C_" topname) ] )
 	 (gen #\()
@@ -747,11 +839,22 @@
 		 ((zero? j))
 	       (gen #t "C_word t" i #\;) ) )
 	 (cond [(eq? 'toplevel id) 
+		(do ([i 0 (add1 i)]
+		     [ns namespace-table (cdr ns)] )
+		    ((null? ns))
+		  (gen #t "C_SYMBOL_TABLE *stable" i #\;) )
 		(let ([ldemand (fold (lambda (lit n) (+ n (literal-size lit))) 0 literals)]
 		      [llen (length literals)] )
 		  (gen #t "C_word *a;"
 		       #t "if(toplevel_initialized) C_kontinue(t1,C_SCHEME_UNDEFINED);"
 		       #t "else C_toplevel_entry(C_text(\"" topname "\"));")
+		  (do ([i 0 (add1 i)]
+		       [ns namespace-table (cdr ns)] )
+		      ((null? ns))
+		    (gen #t "stable" i "=C_new_symbol_table(\""
+			 (caar ns) "\"," (compute-namespace-size (cdar ns)) ");") )
+		  (when disable-stack-overflow-checking
+		    (gen #t "C_disable_overflow_check = 1;") )
 		  (unless unit-name
 		    (cond [target-initial-heap-size
 			   (gen #t "C_set_or_change_heap_size(" target-initial-heap-size ",1);") ]
@@ -771,13 +874,13 @@
 		       #t "toplevel_initialized=1;")
 		  (gen #t "if(!C_demand_2(" ldemand ")){"
 		       #t "C_save(t1);"
-		       #t "C_rereclaim(" (- ldemand) "*sizeof(C_word));"
+		       #t "C_rereclaim2(" ldemand "*sizeof(C_word), 1);"
 		       #t "t1=C_restore;}")
 		  (gen #t "a=C_alloc(" demand ");")
 		  (when (not (zero? llen))
-		    (gen #t "C_initialize_lf(lf," llen ");")
+		    (gen #t "C_initialize_lf(" (prefix-id) "lf," llen ");")
 		    (literal-frame)
-		    (gen #t "C_register_lf(lf," llen ");") ) ) ]
+		    (gen #t "C_register_lf(" (prefix-id) "lf," llen ");") ) ) ]
 	       [rest
 		(gen #t "va_list v;")
 		(gen #t "C_word *a,c2=c;")
@@ -800,7 +903,8 @@
 		      [else
 		       (unless direct (gen #t "C_word *a;"))
 		       (when looping (gen #t "loop:")) 
-		       (when (and direct (not unsafe)) (gen #t "C_stack_check;")) ] )
+		       (when (and direct (not unsafe) (not disable-stack-overflow-checking))
+			 (gen #t "C_stack_check;") ) ] )
 		(when (and external (not unsafe) (not no-argc-checks) (not customizable))
 		  (if (eq? rest-mode 'none)
 		      (when (> n 2) (gen #t "if(c<" n ") C_bad_min_argc(c," n ");"))
@@ -824,7 +928,7 @@
 	   (cond [rest
 		  (gen #t (if (> nec 0) "C_save_and_reclaim" "C_reclaim") "((void*)tr" n #\r)
 		  (when (eq? rest-mode 'vector) (gen #\v))
-		  (gen ",(void*)" id "r")
+		  (gen ",(void*)" (prefix-id) id "r")
 		  (when (> nec 0)
 		    (gen #\, nec #\,)
 		    (apply gen arglist) )
@@ -834,11 +938,11 @@
 		  (case rest-mode
 		    [(list #f) (gen #t "t" n "=C_restore_rest(a,C_rest_count(0));")]
 		    [(vector) (gen #t "t" n "=C_restore_rest_vector(a,C_rest_count(0));")] )
-		  (gen #t id "r(")
+		  (gen #t (prefix-id) id "r(")
 		  (apply gen (intersperse (make-argument-list n "t") #\,))
 		  (gen ",t" n ");}}")
 		  ;; Create secondary routine (no demand-check or argument-count-parameter):
-		  (gen #t #t "static void " id "r(")
+		  (gen #t #t "static void " (prefix-id) id "r(")
 		  (apply gen varlist)
 		  (gen ",C_word t" n "){")
 		  (gen #t "C_word tmp;")
@@ -851,7 +955,7 @@
 		  (gen #t (if (> nec 0) "C_save_and_reclaim" "C_reclaim") "((void*)tr")
 		  (if customizable 
 		      (gen id ",NULL")
-		      (gen n ",(void*)" id) )
+		      (gen n ",(void*)" (prefix-id) id) )
 		  (when (> nec 0)
 		    (gen #\, nec #\,)
 		    (apply gen arglist) )
@@ -863,18 +967,24 @@
 	      n)
 	  ll)
 	 (gen #\}) ) )
-     lambdas) )
+     (filter
+       (lambda (ll)
+         (let* ([partition (lambda-literal-partition ll)]
+                [in-partition (or (not file-partition) (= file-partition partition))])
+           in-partition))
+       lambdas)) )
 
   (debugging 'p "code generation phase...")
   (set! output out)
   (header)
+  (generate-foreign-callback-stub-prototypes foreign-callback-stubs)
   (declarations)
   (generate-external-variables external-variables)
-  (generate-foreign-callback-stub-prototypes foreign-callback-stubs)
   (generate-foreign-stubs foreign-lambda-stubs db)
   (prototypes)
   (generate-foreign-callback-stubs foreign-callback-stubs db)
   (trampolines)
+  (setup-quick-namespace-list)
   (procedures)
   (trailer) )
 
@@ -956,7 +1066,7 @@
        (apply gen varlist)
        (if callback
 	   (gen ") C_noret;" #t "static void " id "(C_word C_c,C_word C_self,C_word C_k,")
-	   (gen ") C_regparm;" #t "static C_word C_fcall " id #\() )
+	   (gen ") C_regparm;" #t "C_regparm static C_word C_fcall " id #\() )
        (apply gen varlist)
        (gen "){")
        (gen #t "C_word C_r=C_SCHEME_UNDEFINED,*C_a=(C_word*)C_buf;")
@@ -1005,7 +1115,7 @@
 
        (define (compute-size type var ns)
 	 (case type
-	   [(char int short bool void unsigned-short scheme-object unsigned-char unsigned-int) ns]
+	   [(char int short bool void unsigned-short scheme-object unsigned-char unsigned-int byte unsigned-byte) ns]
 	   [(float double c-pointer unsigned-integer long integer unsigned-long nonnull-c-pointer)
 	    (string-append ns "+3") ]
 	   [(c-string c-string*)
@@ -1018,7 +1128,8 @@
 			(compute-size (if (vector? t) (vector-ref t 0) t) var ns) ) ]
 		  [(pair? type)
 		   (case (car type)
-		     [(ref pointer c-pointer nonnull-pointer function) (string-append ns "+3")]
+		     [(ref pointer c-pointer nonnull-pointer function instance nonnull-instance)
+		      (string-append ns "+3") ]
 		     [(const) (compute-size (cadr type) var ns)]
 		     [else ns] ) ]
 		  [else ns] ) ] ) )
@@ -1058,12 +1169,12 @@
     (gen #\)) ) )
 
 (define (foreign-type-declaration type target)
-  (let ([err (lambda () (quit "illegal foreign type ~s" type))]
+  (let ([err (lambda () (quit "illegal foreign type `~S'" type))]
 	[str (lambda (ts) (string-append ts " " target))] )
     (case type
       [(scheme-object) (str "C_word")]
-      [(char) (str "char")]
-      [(unsigned-char) (str "unsigned char")]
+      [(char byte) (str "char")]
+      [(unsigned-char unsigned-byte) (str "unsigned char")]
       [(unsigned-int) (str "unsigned int")]
       [(int integer bool) (str "int")]
       [(short) (str "short")]
@@ -1106,6 +1217,7 @@
 		[`(struct ,sname) (string-append "struct " sname " " target)]
 		[`(union ,uname) (string-append "union " uname " " target)]
 		[`(enum ,ename) (string-append "enum " ename " " target)]
+		[((or 'instance 'nonnull-instance) cname sname) (string-append (->string cname) "*" target)]
 		[`(function ,rtype ,argtypes . ,callconv)
 		 (string-append
 		  (foreign-type-declaration rtype "")
@@ -1127,11 +1239,11 @@
 	     [else (err)] ) ] ) ) )
 
 (define (foreign-argument-conversion type)
-  (let ([err (lambda () (quit "illegal foreign argument type ~s" type))])
+  (let ([err (lambda () (quit "illegal foreign argument type `~S'" type))])
     (case type
       ((scheme-object) "(")
       ((char unsigned-char) "C_character_code((C_word)")
-      ((int unsigned-int) "C_unfix(")
+      ((byte int unsigned-int unsigned-byte) "C_unfix(")
       ((short) "C_unfix(")
       ((unsigned-short) "(unsigned short)C_unfix(")
       ((unsigned-long) "C_num_to_unsigned_long(")
@@ -1173,6 +1285,8 @@
 	      (match type
 		[((or 'pointer 'c-pointer) ptype) "C_c_pointer_or_null("]
 		[((or 'nonnull-pointer 'nonnull-c-pointer) ptype) "C_c_pointer_nn("]
+		[`(instance ,cname ,sname) "C_c_pointer_or_null("]
+		[`(nonnull-instance ,cname ,sname) "C_c_pointer_nn("]
 		[`(function ,rtype ,@argtypes) "C_c_pointer_or_null("]
 		[`(const ,ctype) (foreign-argument-conversion ctype)]
 		[`(enum ,etype) "C_num_to_int("]
@@ -1182,13 +1296,15 @@
 	     [else (err)] ) ) ) ) )
 	    
 (define (foreign-result-conversion type dest)
-  (let ([err (lambda () (quit "illegal foreign return type ~s" type))])
+  (let ([err (lambda () (quit "illegal foreign return type `~S'" type))])
     (case type
       ((char unsigned-char) "C_make_character((C_word)")
       ((int) "C_fix((C_word)")
-      ((unsigned-int) "C_fix(0x3fffffff&(C_word)")
+      ((unsigned-int) "C_fix(C_MOST_POSITIVE_FIXNUM&(C_word)")
       ((short) "C_fix((short)")
       ((unsigned-short) "C_fix(0xffff&(C_word)")
+      ((byte) "C_fix((char)")
+      ((unsigned-byte) "C_fix(0xff&(C_word)")
       ((float double) (sprintf "C_flonum(&~a," dest))
       ((nonnull-c-string c-string nonnull-c-pointer c-string* nonnull-c-string*) 
        (sprintf "C_mpointer(&~a,(void*)" dest) )
@@ -1209,6 +1325,10 @@
 		 (sprintf "C_mpointer(&~A,(void*)" dest) ]
 		[('ref rtype) 
 		 (sprintf "C_mpointer(&~A,(void*)&" dest) ]
+		[('instance cname sname)
+		 (sprintf "C_mpointer_or_false(&~A,(void*)" dest) ]
+		[('nonnull-instance cname sname)
+		 (sprintf "C_mpointer(&~A,(void*)" dest) ]
 		[('const ctype) (foreign-result-conversion ctype dest)]
 		[((or 'pointer 'c-pointer) ptype) 
 		 (sprintf "C_mpointer_or_false(&~a,(void*)" dest) ]
