@@ -1,6 +1,6 @@
 ;;;; c-platform.scm - Platform specific parameters and definitions
 ;
-; Copyright (c) 2000-2003, Felix L. Winkelmann
+; Copyright (c) 2000-2004, Felix L. Winkelmann
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -28,8 +28,8 @@
 ; felix@call-with-current-continuation.org
 ;
 ; Felix L. Winkelmann
-; Steinweg 1A
-; 37130 Gleichen, OT Weissenborn
+; Unter den Gleichen 1
+; 37130 Gleichen
 ; Germany
 
 
@@ -42,7 +42,7 @@
   non-foldable-standard-bindings foldable-standard-bindings non-foldable-extended-bindings foldable-extended-bindings
   standard-bindings-that-never-return-false side-effect-free-standard-bindings-that-never-return-false
   installation-home debugging
-  foreign-type-table-size
+  foreign-type-table-size dump-nodes
   unit-name insert-timer-checks used-units inlining
   foreign-declarations block-compilation analysis-database-size line-number-database-size
   target-heap-size target-stack-size try-harder default-installation-home 
@@ -55,7 +55,7 @@
   perform-cps-conversion analyze-expression simplifications perform-high-level-optimizations perform-pre-optimization!
   reorganize-recursive-bindings substitution-table simplify-named-call find-inlining-candidates perform-inlining!
   perform-closure-conversion prepare-for-code-generation compiler-source-file create-foreign-stub expand-foreign-lambda*
-  transform-direct-lambdas!
+  transform-direct-lambdas! decompose-lambda-list rewrite
   debugging-chicken warnings-enabled bomb check-signature posq stringify symbolify flonum? build-lambda-list
   string->c-identifier c-ify-string words check-and-open-input-file close-checked-input-file fold-inner constant?
   collapsable-literal? immediate? canonicalize-begin-body extract-mutable-constants string->expr get get-all
@@ -64,7 +64,7 @@
   simple-lambda-node? compute-database-statistics print-program-statistics output gen gen-list 
   pprint-expressions-to-file foreign-type-check estimate-foreign-result-size scan-used-variables scan-free-variables
   topological-sort print-version print-usage initialize-analysis-database
-  product copyright compiler-features default-declarations default-debugging-declarations units-used-by-default words-per-flonum
+  copyright compiler-features default-declarations default-debugging-declarations units-used-by-default words-per-flonum
   parameter-limit eq-inline-operator optimizable-rest-argument-operators
   membership-test-operators membership-unfold-limit valid-compiler-options valid-compiler-options-with-argument
   target-include-file default-profiling-declarations
@@ -78,8 +78,7 @@
 
 ;;; Parameters:
 
-(define product "the CHICKEN compiler")
-(define copyright "(c)2000-2003 Felix L. Winkelmann")
+(define copyright "(c)2000-2004 Felix L. Winkelmann")
 
 (define default-optimization-passes 3)
 
@@ -93,7 +92,7 @@
      ##sys#for-each ##sys#map ##sys#print
      ##sys#setslot ##sys#dynamic-wind ##sys#call-with-values ##sys#match-error
      ##sys#start-timer ##sys#stop-timer ##sys#gcd ##sys#lcm ##sys#make-promise ##sys#structure? ##sys#slot 
-     ##sys#allocate-vector ##sys#list->vector ##sys#record?
+     ##sys#allocate-vector ##sys#list->vector
      ##sys#list ##sys#cons ##sys#append ##sys#vector ##sys#foreign-char-argument ##sys#foreign-fixnum-argument
      ##sys#foreign-flonum-argument ##sys#error ##sys#peek-c-string ##sys#peek-nonnull-c-string 
      ##sys#peek-and-free-c-string ##sys#peek-and-free-nonnull-c-string
@@ -113,7 +112,7 @@
      '(bound-to-procedure
        ##sys#profile-entry ##sys#profile-exit) ) ) )
 
-(define units-used-by-default '(library eval))
+(define units-used-by-default '(library eval extras))
 (define words-per-flonum 4)
 (define parameter-limit 1024)
 (define small-parameter-limit 128)
@@ -127,16 +126,17 @@
 (define target-include-file "chicken.h")
 
 (define valid-compiler-options
-  '(help version verbose explicit-use quiet no-trace no-warnings unsafe block assume-self-calls
+  '(-help help version verbose explicit-use quiet no-trace no-warnings unsafe block assume-self-calls
     check-syntax hygienic to-stdout usual-integrations case-insensitive emit-debug-info profile
-    fixnum-arithmetic disable-interrupts optimize-leaf-routines debug-calls
-    debug-loops srfi-7 strict strict-srfi-0 strict-reader lambda-lift compile-time-macros-only
+    fixnum-arithmetic disable-interrupts optimize-leaf-routines debug-calls syntax syntax-at-run-time
+    debug-loops strict strict-srfi-0 strict-reader lambda-lift run-time-macros
+    disable-stack-overflow-checks disable-c-syntax-checks unsafe-libraries
     hygienic-at-run-time analyze-only strict-letrec dynamic ffi) )
 
 (define valid-compiler-options-with-argument
-  '(debug output-file include-path heap-size stack-size unit uses keyword-style visit require-for-syntax
-    prelude postlude prologue epilogue nursery extend feature no-feature compress-literals
-    heap-growth heap-shrinkage heap-initial-size ffi-define ffi-include-path) )
+  '(debug output-file include-path heap-size stack-size unit uses keyword-style require-for-syntax
+    prelude postlude prologue epilogue nursery extend feature no-feature compress-literals split-level
+    heap-growth heap-shrinkage heap-initial-size ffi-define ffi-include-path split) )
 
 
 ;;; Standard and extended bindings:
@@ -164,9 +164,10 @@
 (define default-extended-bindings
   '(bitwise-and bitwise-ior bitwise-xor bitwise-not add1 sub1 fx+ fx- fx* fx/ fxmod
     fx= fx> fx< fx>= fx<= fixnum? fxneg fxmax fxmin identity
-    arithmetic-shift signum void flush-output
-    not-pair? null-list? print print* error cpu-time proper-list?
-    u8vector->byte-vector s8vector->byte-vector u16vector->byte-vector s16vector->byte-vector u32vector->byte-vector 
+    arithmetic-shift signum void flush-output thread-specific thread-specific-set!
+    not-pair? null-list? print print* error cpu-time proper-list? call/cc
+    u8vector->byte-vector s8vector->byte-vector u16vector->byte-vector s16vector->byte-vector 
+    u32vector->byte-vector 
     s32vector->byte-vector byte-vector-length block-ref block-set! number-of-slots
     f32vector->byte-vector f64vector->byte-vector byte-vector-ref byte-vector-set!
     first second third fourth make-record-instance
@@ -187,7 +188,7 @@
     ##sys#foreign-block-argument ##sys#foreign-number-vector-argument
     ##sys#foreign-string-argument ##sys#foreign-pointer-argument 
     ##sys#foreign-integer-argument ##sys#foreign-unsigned-integer-argument
-    ##sys#peek-fixnum ##sys#setislot ##sys#poke-integer ##sys#permanent? ##sys#values
+    ##sys#peek-fixnum ##sys#setislot ##sys#poke-integer ##sys#permanent? ##sys#values ##sys#poke-double
     ##sys#intern-symbol ##sys#make-symbol ##sys#null-pointer? ##sys#peek-byte) )
 
 (define side-effecting-standard-bindings
@@ -349,7 +350,7 @@
 			 (make-node '##core#inline '("C_fixnum_divide") callargs) ) ) ) )
 	    (make-node
 	     '##core#call '(#t)
-	     (cons* (make-node '##core#proc '("C_quotient") '()) cont callargs) ) ) ) ) )
+	     (cons* (make-node '##core#proc '("C_quotient" #t) '()) cont callargs) ) ) ) ) )
 
 (rewrite
  'eqv? 8
@@ -411,7 +412,7 @@
 		  (append (cdr (butlast callargs)) (map qnode (first (node-parameters lastarg)))) ) )
 	  (make-node
 	   '##core#call '(#t)
-	   (cons* (make-node '##core#proc '("C_apply") '())
+	   (cons* (make-node '##core#proc '("C_apply" #t) '())
 		  cont callargs) ) ) ) )
   (rewrite 'apply 8 rewrite-apply)
   (rewrite '##sys#apply 8 rewrite-apply) )
@@ -570,7 +571,6 @@
 (rewrite 'char-downcase 2 1 "C_u_i_char_downcase" #t #f)
 (rewrite 'list-tail 2 2 "C_i_list_tail" #t #f)
 (rewrite '##sys#structure? 2 2 "C_i_structurep" #t #f)
-(rewrite '##sys#record? 2 2 "C_recordp" #t #f)
 (rewrite '##sys#bytevector? 2 2 "C_bytevectorp" #t #f)
 (rewrite 'block-ref 2 2 "C_slot" #f #f)
 (rewrite 'number-of-slots 2 1 "C_block_size" #f #f)
@@ -644,6 +644,8 @@
 (rewrite 'map 11 2 '##sys#map #t)
 (rewrite 'block-set! 11 3 '##sys#setslot #t)
 (rewrite 'make-record-instance 11 #f '##sys#make-structure #f)
+(rewrite 'substring 11 3 '##sys#substring #f)
+(rewrite 'string-append 11 2 '##sys#string-append #f)
 
 (rewrite 'vector-set! 11 3 '##sys#setslot #f)
 (rewrite 'vector-set! 2 3 "C_i_vector_set" #t #f)
@@ -773,6 +775,7 @@
 (rewrite 'pointer->object 17 2 "C_pointer_to_object")
 (rewrite '##sys#setislot 17 3 "C_i_set_i_slot")
 (rewrite '##sys#poke-integer 17 3 "C_poke_integer")
+(rewrite '##sys#poke-double 17 3 "C_poke_double")
 (rewrite 'string=? 17 2 "C_i_string_equal_p" "C_u_i_string_equal_p")
 (rewrite 'string-ci=? 17 2 "C_i_string_ci_equal_p")
 (rewrite '##sys#fudge 17 1 "C_fudge")
@@ -840,3 +843,29 @@
 					  (list-tabulate c (lambda (i) (varnode tmp)) ) ) ) ) ) ) ) ) ) ) ) ) ) )
   (rewrite 'make-vector 8 rewrite-make-vector)
   (rewrite '##sys#make-vector 8 rewrite-make-vector) )
+
+(rewrite 'thread-specific 7 1 "C_slot" 10 #f)
+(rewrite 'thread-specific-set! 20 2 "C_i_setslot" 10 #f)
+
+(let ()
+  (define (rewrite-call/cc db classargs cont callargs)
+    (and (= 1 (length callargs))
+	 (let ([val (first callargs)])
+	   (and (eq? '##core#variable (node-class val))
+		(and-let* ([proc (get db (first (node-parameters val)) 'value)]
+			   [(eq? '##core#lambda (node-class proc))] )
+		  (let ([llist (third (node-parameters proc))])
+		    (decompose-lambda-list 
+		     llist
+		     (lambda (vars argc rest)
+		       (and (<= 1 argc 2)
+			    (let ([var (or rest (second llist))])
+			      (and (not (get db var 'references))
+				   (not (get db var 'assigned)) 
+				   (begin
+				     (debugging 'x "removing unused `call/cc'")
+				     (make-node
+				      '##core#call '(#t)
+				      (list val cont (qnode #f)) ) ) ) ) ) ) ) ) ) ) ) ) )
+  (rewrite 'call-with-current-continuation 8 rewrite-call/cc)
+  (rewrite 'call/cc 8 rewrite-call/cc) )
