@@ -1,6 +1,6 @@
 ;;;; chicken.scm - The CHICKEN Scheme compiler (loader/main-module)
 ;
-; Copyright (c) 2000-2003, Felix L. Winkelmann
+; Copyright (c) 2000-2004, Felix L. Winkelmann
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -28,12 +28,15 @@
 ; felix@call-with-current-continuation.org
 ;
 ; Felix L. Winkelmann
-; Steinweg 1A
-; 37130 Gleichen, OT Weissenborn
+; Unter den Gleichen 1
+; 37130 Gleichen
 ; Germany
 
 
-(declare (uses extras srfi-1 match-support syntax-case srfi-4 script-utils support compiler optimizer driver platform backend easyffi))
+(declare
+  (uses extras srfi-1 match-support syntax-case srfi-4 utils support compiler optimizer driver platform backend 
+	easyffi partition lolevel)
+  (run-time-macros) )
 
 
 #{compiler
@@ -64,7 +67,7 @@
   simple-lambda-node? compute-database-statistics print-program-statistics output gen gen-list 
   pprint-expressions-to-file foreign-type-check estimate-foreign-result-size scan-used-variables scan-free-variables
   topological-sort print-version print-usage initialize-analysis-database
-  product copyright compiler-features default-declarations units-used-by-default words-per-flonum
+  copyright compiler-features default-declarations units-used-by-default words-per-flonum
   foreign-string-result-reserve parameter-limit default-output-filename eq-inline-operator optimizable-rest-argument-operators
   membership-test-operators membership-unfold-limit valid-compiler-options valid-compiler-options-with-argument
   generate-code make-variable-list make-argument-list generate-foreign-stubs foreign-type-declaration
@@ -74,8 +77,9 @@
 (include "tweaks")
 
 (eval-when (load) 
-  (include "moremacros")
-  (include "match") )
+  (include "chicken-more-macros")
+  (include "chicken-ffi-macros")
+  (include "chicken-match-macros") )
 
 (##sys#provide 'extras 'srfi-1 'syntax-case 'srfi-4)
 
@@ -104,8 +108,9 @@
 (define arguments (cdr (argv)))
 
 (eval-when (eval)
-  (unless (memq '#:match (features)) (include "match"))
+  (unless (memq '#:match (features)) (include "chicken-match-macros"))
   (include "support")
+  (include "partition")
   (include "compiler")
   (include "optimizer")
   (include "batch-driver")
@@ -156,13 +161,13 @@
 		 (case level
 		   [(0) #f]
 		   [(1)
-		    (set! options (cons* 'optimize-leaf-routines 'lambda-lift options)) ]
+		    (set! options (cons* 'optimize-leaf-routines options)) ]
 		   [(2)
 		    (set! options
-		      (cons* 'optimize-leaf-routines 'lambda-lift 'usual-integrations options) ) ] 
+		      (cons* 'optimize-leaf-routines 'usual-integrations options) ) ] 
 		   [(3)
 		    (set! options
-		      (cons* 'optimize-leaf-routines 'lambda-lift 'usual-integrations 'unsafe options) ) ]
+		      (cons* 'optimize-leaf-routines 'usual-integrations 'unsafe options) ) ]
 		   [else (warning "invalid optimization level ~S - ignored" (car rest))] )
 		 (loop (cdr rest)) ) ]
 	      [(eq? 'debug-level o)
