@@ -1,11 +1,11 @@
-;;;; compiler.scm - The CHICKEN Scheme compiler (front-end)
+;;;; compiler.scm - The CHICKEN Scheme compiler
 ;
 ;
 ; "This is insane. What we clearly want to do is not exactly clear, and is rooted in NCOMPLR."
 ;
 ;
-;------------------------------------------------------------------------------------------------------------------------
-; Copyright (c) 2000-2003, Felix L. Winkelmann
+;-----------------------------------------------------------------------------------------------------------
+; Copyright (c) 2000-2004, Felix L. Winkelmann
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -33,8 +33,8 @@
 ; felix@call-with-current-continuation.org
 ;
 ; Felix L. Winkelmann
-; Steinweg 1A
-; 37130 Gleichen, OT Weissenborn
+; Unter den Gleichen 1
+; 37130 Gleichen
 ; Germany
 ;
 ;
@@ -63,10 +63,13 @@
 ; (foreign-parse {<string>})
 ; (block)
 ; (separate)
-; (compile-time-macros-only)
+; (run-time-macros)
 ; (export {<name>})
 ; (compress-literals [<threshold>])
 ; (safe-globals)
+; (namespace <name> {<symbol})
+; (custom-declare (<tag> <name> <filename> <arg> ...) <string> ...)
+; (data <tag1> <exp1> ...)
 ;
 ;   <type> = fixnum | generic
 ;
@@ -74,15 +77,12 @@
 ;
 ; <variable>
 ; <constant>
-; (##core#include (quote <string>))
 ; (##core#declare {(quote <spec>)})
 ; (##core#immutable <exp>)
-; (##core#qualified (quote <exp>))
 ; (##core#global-ref <variable>)
 ; (quote <exp>)
 ; (if <exp> <exp> [<exp>])
 ; (let ({(<variable> <exp>)}) <body>)
-; (let-macro (<macrodef> ...) <body>)
 ; (##core#let-location (quote <symbol>) (quote <type>) [<init>] <exp>)
 ; (lambda <variable> <body>)
 ; (lambda ({<variable>}+ [. <variable>]) <body>)
@@ -113,7 +113,8 @@
 ; (##core#foreign-callback-wrapper (quote <name>) (quote <qualifiers>) (quote <type>) (quote {<type>}) <exp>)
 ; (##core#define-external-variable (quote <name>) (quote <type>) (quote <bool>))
 ; (##core#check <exp>)
-; (##core#require-for-syntax <symbol> ...)
+; (##core#require-for-syntax <exp> ...)
+; (##core#require-extension '<id> ...)
 ; (##core#app <exp> {<exp>})
 ; (<exp> {<exp>})
 ;
@@ -172,7 +173,7 @@
 ; [##core#setglobal_i {<literal> <block-mode>} <exp>]
 ; [##core#literal {<literal>}]
 ; [##core#immediate {<type> [<immediate>]}]     - type: bool/fix/nil/char
-; [##core#proc {<name>}]
+; [##core#proc {<name> [<non-internal>]}]
 ; [##core#recurse {<tail-flag> <call-id>} <exp1> ...]
 ; [##core#return <exp>]
 ; [##core#direct_call {<safe-flag> <debug-symbol> <call-id> <words>} <exp-f> <exp>...]
@@ -237,38 +238,36 @@
 #ifndef C_DEFAULT_TARGET_HEAP_SIZE
 # define C_DEFAULT_TARGET_HEAP_SIZE 0
 #endif
-
-#ifndef C_POSIX_IS_AVAILABLE
-# define C_POSIX_IS_AVAILABLE 0
-#endif
 EOF
 ) )
 
 
 #{compiler
-  build-information compiler-arguments process-command-line
+  build-information compiler-arguments process-command-line explicit-use-flag
   default-standard-bindings default-extended-bindings side-effecting-standard-bindings
   non-foldable-standard-bindings foldable-standard-bindings non-foldable-extended-bindings foldable-extended-bindings
   standard-bindings-that-never-return-false side-effect-free-standard-bindings-that-never-return-false
   installation-home decompose-lambda-list external-to-pointer defconstant-bindings
   foreign-type-table-size copy-node! error-is-extended-binding toplevel-scope toplevel-lambda-id
-  unit-name insert-timer-checks used-units external-variables
-  profile-info-vector-name finish-foreign-result pending-canonicalizations
+  unit-name insert-timer-checks used-units external-variables require-imports-flag custom-declare-alist
+  profile-info-vector-name finish-foreign-result pending-canonicalizations secondary-macroexpand
   foreign-declarations emit-trace-info block-compilation analysis-database-size line-number-database-size
   always-bound-to-procedure block-globals make-block-variable-literal block-variable-literal? block-variable-literal-name
-  target-heap-size target-stack-size try-harder default-installation-home 
-  target-initial-heap-size internal-bindings source-filename
-  default-default-target-heap-size default-default-target-stack-size verbose-mode original-program-size posix-is-available
+  target-heap-size target-stack-size try-harder default-installation-home valid-c-identifier?
+  target-initial-heap-size internal-bindings source-filename dump-nodes default-user-pass-2
+  default-default-target-heap-size default-default-target-stack-size verbose-mode original-program-size
   current-program-size line-number-database-2 foreign-lambda-stubs immutable-constants foreign-variables
   rest-parameters-promoted-to-vector inline-table inline-table-used constant-table constants-used mutable-constants
   broken-constant-nodes inline-substitutions-enabled loop-lambda-names expand-profile-lambda
-  profile-lambda-list profile-lambda-index emit-profile expand-profile-lambda
-  direct-call-ids foreign-type-table first-analysis callback-names
+  profile-lambda-list profile-lambda-index emit-profile expand-profile-lambda no-c-syntax-checks
+  direct-call-ids foreign-type-table first-analysis callback-names namespace-table
   initialize-compiler canonicalize-expression expand-foreign-lambda update-line-number-database! scan-toplevel-assignments
   perform-cps-conversion analyze-expression simplifications perform-high-level-optimizations perform-pre-optimization!
   reorganize-recursive-bindings substitution-table simplify-named-call
-  perform-closure-conversion prepare-for-code-generation compiler-source-file create-foreign-stub expand-foreign-lambda*
-  transform-direct-lambdas! expand-foreign-callback-lambda* debugging
+  perform-closure-conversion prepare-for-code-generation compiler-source-file create-foreign-stub 
+  expand-foreign-lambda* data-declarations
+  factor partition-fm module-export-list user-level-optimizers
+  transform-direct-lambdas! expand-foreign-callback-lambda* debugging emit-unsafe-marker
   debugging-chicken warnings-enabled bomb check-signature posq stringify symbolify flonum? build-lambda-list
   string->c-identifier c-ify-string words check-and-open-input-file close-checked-input-file fold-inner constant?
   collapsable-literal? immediate? canonicalize-begin-body extract-mutable-constants string->expr get get-all
@@ -276,15 +275,16 @@ EOF
   build-node-graph build-expression-tree fold-boolean inline-lambda-bindings match-node expression-has-side-effects?
   simple-lambda-node? compute-database-statistics print-program-statistics output gen gen-list 
   pprint-expressions-to-file foreign-type-check estimate-foreign-result-size scan-used-variables scan-free-variables
-  topological-sort print-version print-usage initialize-analysis-database export-list
+  topological-sort print-version print-usage initialize-analysis-database export-list csc-control-file
   estimate-foreign-result-location-size compressed-literals-initializer parse-easy-ffi
   expand-foreign-callback-lambda default-optimization-passes default-optimization-passes-when-trying-harder
-  product copyright compiler-features units-used-by-default words-per-flonum
-  parameter-limit eq-inline-operator optimizable-rest-argument-operators
+  copyright compiler-features units-used-by-default words-per-flonum disable-stack-overflow-checking
+  parameter-limit eq-inline-operator optimizable-rest-argument-operators check-c-syntax
   membership-test-operators membership-unfold-limit valid-compiler-options valid-compiler-options-with-argument
   make-random-name final-foreign-type real-name-table real-name set-real-name! safe-globals-flag
   location-pointer-map literal-compression-threshold compressed-literals compressable-literal
   generate-code make-variable-list make-argument-list generate-foreign-stubs foreign-type-declaration
+  process-custom-declaration
   foreign-argument-conversion foreign-result-conversion foreign-type-convert-argument foreign-type-convert-result}
 
 (eval-when (compile eval)
@@ -293,17 +293,17 @@ EOF
 (include "tweaks")
 (include "parameters")
 
+(define-inline (gensym-f-id) (gensym 'f_))
+
 (eval-when (eval)
   (define installation-home #f)
   (define default-target-heap-size #f)
-  (define default-target-stack-size #f) 
-  (define posix-is-available-flag (memq #:posix (features))) )
+  (define default-target-stack-size #f) )
 
 (eval-when (load)
   (define-foreign-variable installation-home c-string "C_INSTALL_HOME")
   (define-foreign-variable default-target-heap-size int "C_DEFAULT_TARGET_HEAP_SIZE")
-  (define-foreign-variable default-target-stack-size int "C_DEFAULT_TARGET_STACK_SIZE") 
-  (define-foreign-variable posix-is-available-flag bool "C_POSIX_IS_AVAILABLE") )
+  (define-foreign-variable default-target-stack-size int "C_DEFAULT_TARGET_STACK_SIZE") )
 
 (define user-options-pass (make-parameter #f))
 (define user-read-pass (make-parameter #f))
@@ -344,6 +344,12 @@ EOF
 (define literal-compression-threshold #f)
 (define compressed-literals-initializer #f)
 (define safe-globals-flag #f)
+(define explicit-use-flag #f)
+(define disable-stack-overflow-checking #f)
+(define namespace-table '())
+(define require-imports-flag #f)
+(define no-c-syntax-checks #f)
+(define emit-unsafe-marker #f)
 
 
 ;;; These are here so that the backend can access them:
@@ -351,7 +357,6 @@ EOF
 (define default-installation-home installation-home)
 (define default-default-target-heap-size default-target-heap-size)
 (define default-default-target-stack-size default-target-stack-size)
-(define posix-is-available posix-is-available-flag)
 
 
 ;;; Other global variables:
@@ -384,81 +389,34 @@ EOF
 (define error-is-extended-binding #f)
 (define real-name-table #f)
 (define location-pointer-map '())
-(define pending-canonicalizations #f)
+(define pending-canonicalizations '())
 (define defconstant-bindings '())
 (define callback-names '())
 (define toplevel-scope #t)
 (define toplevel-lambda-id #f)
+(define (secondary-macroexpand x) x)
+(define custom-declare-alist '())
+(define csc-control-file #f)
+(define data-declarations '())
+(define module-export-list '())
+(define user-level-optimizers #f)
 
 
 ;;; Initialize globals:
 
-(randomize)
-
 (define (initialize-compiler)
-  (set! unit-name #f)
-  (set! number-type 'generic)
-  (set! standard-bindings '())
-  (set! extended-bindings '())
-  (set! insert-timer-checks #t)
-  (set! used-units '())
-  (set! unsafe #f)
-  (set! export-list #f)
-  (set! always-bound '())
-  (set! foreign-declarations '())
-  (set! emit-trace-info #f)
-  (set! block-compilation #f)
-  (set! target-heap-size #f)
-  (set! target-initial-heap-size #f)
-  (set! target-stack-size #f)
-  (set! try-harder #t)
-  (set! verbose-mode #f)
-  (set! original-program-size #f)
-  (set! current-program-size 0)
   (if line-number-database-2
       (vector-fill! line-number-database-2 '())
       (set! line-number-database-2 (make-vector line-number-database-size '())) )
-  (set! foreign-lambda-stubs '())
-  (set! immutable-constants '())
-  (set! foreign-variables '())
-  (set! rest-parameters-promoted-to-vector '())
   (if inline-table
       (vector-fill! inline-table '())
       (set! inline-table (make-vector inline-table-size '())) )
-  (set! inline-table-used #f)
   (if constant-table
       (vector-fill! constant-table '())
       (set! constant-table (make-vector constant-table-size '())) )
-  (set! constants-used #f)
-  (set! mutable-constants '())
-  (set! inline-substitutions-enabled #f)
-  (set! direct-call-ids '())
-  (set! broken-constant-nodes '()) 
-  (set! first-analysis #t)
-  (set! optimize-leaf-routines #f)
-  (set! foreign-callback-stubs '())
-  (set! loop-lambda-names '())
-  (set! emit-profile #f)
-  (set! profile-lambda-list '())
-  (set! profile-lambda-index 0)
   (set! profile-info-vector-name (make-random-name 'profile-info))
-  (set! always-bound-to-procedure '())
-  (set! external-variables '())
-  (set! pending-canonicalizations '())
-  (set! no-bound-checks #f)
-  (set! no-argc-checks #f)
-  (set! no-procedure-checks #f)
-  (set! block-globals '())
-  (set! external-to-pointer '())
-  (set! location-pointer-map '())
-  (set! error-is-extended-binding #f)
   (set! real-name-table (make-vector real-name-table-size '()))
-  (set! compressed-literals '())
-  (set! literal-compression-threshold #f)
-  (set! defconstant-bindings '())
-  (set! safe-globals-flag #f)
-  (set! toplevel-lambda-id #f)
-  (set! callback-names '())
+  (set! user-level-optimizers (make-vector user-level-optimizer-table-size '()))
   (if foreign-type-table
       (vector-fill! foreign-type-table '())
       (set! foreign-type-table (make-vector foreign-type-table-size '())) ) )
@@ -524,7 +482,7 @@ EOF
 		   (quit "syntax error - malformed expression `~s'" x) ) )
 	     (set! ##sys#syntax-error-culprit x)
 	     (let* ([x2 (cons name rest)]
-		    [xexpanded (##sys#secondary-macroexpand (##sys#macroexpand-1-hook x2 me))] )
+		    [xexpanded (secondary-macroexpand (##sys#macroexpand-1-hook x2 me))] )
 	       (cond [(not (eq? x2 xexpanded))
 		      (when ln (update-line-number-database! xexpanded ln))
 		      (walk xexpanded ae me dest) ]
@@ -548,7 +506,7 @@ EOF
 					 (compressable-literal lit literal-compression-threshold) ) ] )
 			   (if cf
 			       (let ([var (gensym 'lf)])
-				 (debugging 'x "compressing literal of size" cf)
+				 (debugging 'o "compressing literal of size" cf)
 				 (set! compressed-literals
 				   (alist-cons var (write-to-string lit) compressed-literals) )
 				 (set! always-bound (cons var always-bound))
@@ -556,9 +514,6 @@ EOF
 				 var)
 			       x) ) )
 
-			((##core#qualified)
-			 (walk (cadr x) ae me dest) )
-
 			((##core#check)
 			 (if unsafe
 			     ''#t
@@ -586,6 +541,18 @@ EOF
 				  `(##sys#require ,@(map (lambda (x) `',x) rs)) )
 			      ae me dest) ) ) )
 
+			((##core#require-extension)
+			 (walk
+			  (let loop ([ids (cdr x)])
+			    (if (null? ids)
+				'(##core#undefined)
+				(let ([id (cadar ids)])
+				  (let-values ([(exp f) (##sys#do-the-right-thing id #t)])
+				    #;(unless f
+				      (warning "extension `~A' is currently not installed" id) )
+				    `(begin ,exp ,(loop (cdr ids))) ) ) ) )
+			  ae me dest) )
+
 			((let)
 			 (##sys#check-syntax 'let x '(let #((variable _) 0) . #(_ 1)))
 			 (let* ([bindings (cadr x)]
@@ -595,7 +562,7 @@ EOF
 			   `(let ,(map (lambda (alias b)
 					 (list alias (walk (cadr b) ae me (car b))) )
 				       aliases bindings)
-			      ,(walk (##sys#canonicalize-body (cddr x))
+			      ,(walk (##sys#canonicalize-body (cddr x) me)
 				     (append (map cons vars aliases) ae)
 				     me dest) ) ) )
 
@@ -603,13 +570,19 @@ EOF
 			 (##sys#check-syntax 'lambda x '(_ lambda-list . #(_ 1)))
 			 (let ([llist (cadr x)]
 			       [obody (cddr x)] )
+			   (when (##sys#extended-lambda-list? llist)
+			     (set!-values 
+			      (llist obody) 
+			      (##sys#expand-extended-lambda-list 
+			       llist obody
+			       ##sys#error) ) )
 			   (decompose-lambda-list
 			    llist
 			    (lambda (vars argc rest)
 			      (let* ([aliases (map gensym vars)]
 				     [body 
 				      (walk 
-				       (##sys#canonicalize-body (cddr x))
+				       (##sys#canonicalize-body obody me)
 				       (append (map cons vars aliases) ae)
 				       me #f) ]
 				     [llist2 
@@ -630,7 +603,7 @@ EOF
 				[aliases (map gensym vars)]
 				[body 
 				 (walk 
-				  (##sys#canonicalize-body obody)
+				  (##sys#canonicalize-body obody me)
 				  (append (map cons vars aliases) ae)
 				  me #f) ] )
 			   (set-real-names! aliases vars)
@@ -725,6 +698,7 @@ EOF
                                 [name (if (pair? (cdddr x))
                                           (cadr (fourth x))
                                           (symbol->string var) ) ] )
+			   (check-c-syntax name 'define-foreign-variable)
                            (set! foreign-variables
 			     (cons (list var type (if (string? name) name (symbol->string name)))
 				   foreign-variables))
@@ -818,21 +792,6 @@ EOF
                                     (set! always-bound (cons var always-bound))
                                     (walk `(##core#set! ,var ',val) ae me #f) ) ] ) ) )
 
-			((##core#include)
-			 (let* ([ln (get-line x)]
-				[file (##sys#resolve-include-filename (cadadr x))]
-				[in (check-and-open-input-file file ln)] )
-			   (when verbose-mode (printf "including ~s~%~!" file))
-			   (fluid-let ([##sys#read-line-counter 1])
-			     (walk 
-			      `(begin 
-				 ,@(do ((forms '() (cons (##sys#compiler-toplevel-macroexpand-hook x) forms))
-					(x (read in) (read in)) )
-				       ((eof-object? x) 
-					(reverse forms) )
-				     (when ln (update-line-number-database! x ln)) ) )
-			      ae me #f) ) ) )
-
 			((##core#declare)
 			 (walk
 			  `(begin
@@ -844,10 +803,13 @@ EOF
 			 (let-values ([(args lam) (split-at (cdr x) 4)])
 			   (let* ([lam (car lam)]
 				  [name (first args)]
-				  [rtype (cadadr (third args))]
-				  [atypes (cadadr (fourth args))]
+				  [rtype (cadr (third args))]
+				  [atypes (cadr (fourth args))]
 				  [vars (second lam)] )
-			     (set! callback-names (cons (cadr name) callback-names))
+			     (let ([name (cadr name)])
+			       (if (valid-c-identifier? name)
+				   (set! callback-names (cons name callback-names))
+				   (quit "name `~S' of external definition is not a valid C identifier" name) ) )
 			     (when (or (not (proper-list? vars)) 
 				       (not (proper-list? atypes))
 				       (not (= (length vars) (length atypes))) )
@@ -897,34 +859,22 @@ EOF
 
 			         [##sys#strict-mode (handle-call)]
 
-				 [(eq? name 'let-macro)
-				  (##sys#check-syntax 'let-macro x '(let-macro #(list 0) . #(_ 1)))
-				  (let* ([me2 (##sys#expand-local-macrodefs (cadr x))]
-					 [names (map car me2)]
-					 [aliases (map gensym names)] )
-				    (walk
-				     (##sys#canonicalize-body (cddr x))
-				     (append (map cons names aliases) ae)
-				     (append (map (lambda (alias mdef) (cons alias (cdr mdef))) aliases me2) me)
-				     dest) ) ]
-
 				 [(eq? 'location name)
-				  (##sys#check-syntax 'location x '(location symbol))
+				  (##sys#check-syntax 'location x '(location _))
 				  (let ([sym (cadr x)])
-				    (cond [(assq (resolve sym ae) location-pointer-map)
-					   => (lambda (a)
-						(walk
-						 `(##core#inline_allocate ("C_a_i_data_mpointer" 3) ,(second a)) 
-						 ae me #f) ) ]
-					  [(assq sym external-to-pointer) 
-					   => (lambda (a) (walk (cdr a) ae me #f)) ]
-					  [(memq sym callback-names)
-					   `(##core#inline_ref (,(symbol->string sym) c-pointer)) ]
-					  [else
-					   (let ([ln (get-line x)])
-					     (if ln
-						 (quit "syntax error in line ~S - undefined location ~S" ln sym)
-						 (quit "syntax error - undefined location ~S" sym) ) ) ] ) ) ]
+				    (if (symbol? sym)
+					(cond [(assq (resolve sym ae) location-pointer-map)
+					       => (lambda (a)
+						    (walk
+						     `(##sys#make-locative ,(second a) 0 #f 'location)
+						     ae me #f) ) ]
+					      [(assq sym external-to-pointer) 
+					       => (lambda (a) (walk (cdr a) ae me #f)) ]
+					      [(memq sym callback-names)
+					       `(##core#inline_ref (,(symbol->string sym) c-pointer)) ]
+					      [else 
+					       (walk `(##sys#make-locative ,sym 0 #f 'location) ae me #f) ] )
+					(walk `(##sys#make-locative ,sym 0 #f 'location) ae me #f) ) ) ]
 				 
 				 [else (handle-call)] ) ) ) ) ] ) ) ) )
 
@@ -1014,12 +964,19 @@ EOF
 	  (let ([fds (cdr spec)])
 	    (if (every string? fds)
 		(set! foreign-declarations (append foreign-declarations fds))
-		(quit "syntax error in declaration: `~S'" spec) ) ) )
+		(quit "syntax error in declaration: `~S'" spec) )
+	    (check-c-syntax (apply string-append fds) 'foreign-declare) ) )
 	 ((foreign-parse)
 	  (let ([fds (cdr spec)])
 	    (if (every string? fds)
-		(return `(begin ,@(parse-easy-ffi (string-intersperse fds ""))))
+		(return 
+		 (##sys#compiler-toplevel-macroexpand-hook 
+		  `(begin ,@(parse-easy-ffi (string-intersperse fds "")))) )
 		(quit "syntax error in declaration: `~S'" spec) ) ) )
+	 ((custom-declare)
+	  (if (or (not (list? spec)) (not (list? (cadr spec))) (< (length (cadr spec)) 3))
+	      (quit "syntax error in declaration: `~S'" spec)
+	      (process-custom-declaration (cadr spec) (cddr spec)) ) )
 	 ((block) (set! block-compilation #t))
 	 ((separate) (set! block-compilation #f))
 	 ((not)
@@ -1049,22 +1006,60 @@ EOF
 		(set! unsafe #t)
 		(##match#set-error-control #:unspecified) ]
 	       [else (warning "illegal declaration specifier `~s'" spec)] ) ] ) )
-	 ((compile-time-macros-only) (set! ##sys#enable-runtime-macros #f))
+	 ((run-time-macros) (set! ##sys#enable-runtime-macros #t))
 	 ((block-global hide) 
 	  (let ([syms (cdr spec)])
 	    (when export-list 
 	      (set! export-list (lset-difference eq? export-list syms)) )
 	    (set! block-globals (lset-union eq? syms block-globals)) ) )
 	 ((export) 
-	  (let ([syms (cdr spec)])
+	  (let-values ([(syms mods) (partition symbol? (cdr spec))])
 	    (set! block-globals (lset-difference eq? block-globals syms))
-	    (set! export-list (lset-union eq? syms (or export-list '()))) ) )
+	    (set! export-list (lset-union eq? syms (or export-list '())))
+	    (for-each
+	      (match-lambda
+		[((? symbol?) (? symbol?) ...) #t]
+		[m (quit "invalid syntax in `export' declaration: `~S'" m)] )
+	      mods)
+	    (set! module-export-list
+	      (let loop ([all (append module-export-list mods)]) 
+		(if (null? all)
+		    '()
+		    (match (car all)
+		      [(mod exps ...)
+		       (let-values ([(same others)
+				     (partition (lambda (m) (eq? (car m) mod)) all) ] )
+			 (let ([syms (apply lset-union eq? exps (map cdr same))])
+			   ;; We should check here for multiple exports with same name
+			   (set! export-list (lset-union eq? syms export-list))
+			   (alist-cons mod syms (loop others) ) ) ) ] ) ) ) ) ) )
 	 ((compress-literals)
 	  (set! literal-compression-threshold 
 	    (or (and (pair? (cdr spec)) (number? (cadr spec)) (cadr spec))
 		default-literal-compression-threshold) ) 
 	  (when (and (list? spec) (= 3 (length spec)))
 	    (set! compressed-literals-initializer (third spec)) ) )
+	 ((data)
+	  (let loop ([ddefs (cdr spec)])
+	    (match ddefs
+	      [() #f]
+	      [(tag data . more)
+	       (set! data-declarations
+		 (alist-cons 
+		  (->string tag)
+		  (->string data) 
+		  data-declarations) )
+	       (loop more) ]
+	      [_ (quit "invalid arguments to `data' declaration: ~S" spec)] ) ) )
+	 ((namespace)
+	  (check-decl spec 2)
+	  (let* ([syms (cdr spec)]
+		 [ns (car syms)] )
+	    (if (every symbol? syms)
+		(let ([oldsyms (or (and-let* ([a (assq ns namespace-table)]) (cdr a)) '())])
+		  (set! namespace-table 
+		    (alist-update! ns (lset-union eq? oldsyms (cdr syms)) namespace-table eq?) ) )
+		(quit "invalid arguments to `namespace' declaration: ~S" spec) ) ) )
 	 (else (warning "illegal declaration specifier `~s'" spec)) )
        '(##core#undefined) ) ) )
 
@@ -1142,6 +1137,7 @@ EOF
 	 [body (apply string-append (map cadr (cdddr exp)))]
  	 [argtypes (map car args)]
 	 [argnames (map cadr args)] )
+    (check-c-syntax body 'foreign-lambda*)
     (create-foreign-stub rtype #f argtypes argnames body #f) ) )
 
 (define (expand-foreign-callback-lambda* exp)
@@ -1150,6 +1146,7 @@ EOF
 	 [body (apply string-append (map cadr (cdddr exp)))]
  	 [argtypes (map car args)]
 	 [argnames (map cadr args)] )
+    (check-c-syntax body 'foreign-callback-lambda*)
     (create-foreign-stub rtype #f argtypes argnames body #t) ) )
 
 
@@ -1200,7 +1197,7 @@ EOF
 		     (k1 (lambda (r) (make-node '##core#call '(#t) (list (varnode t1) r)))) )
 		(make-node 'let
 			   (list t1)
-			   (list (make-node '##core#lambda (list (gensym 'f) #f (list t2) 0) 
+			   (list (make-node '##core#lambda (list (gensym-f-id) #f (list t2) 0) 
 					    (list (k (varnode t2))) )
 				 (walk (car subs)
 				       (lambda (v)
@@ -1216,7 +1213,7 @@ EOF
 			     (make-node 'let
 					(list (car vars))
 					(list r (loop (cdr vars) (cdr vals))) ) ) ) ) ) )
-	((lambda) (cps-lambda (gensym 'f) (first params) subs k))
+	((lambda) (cps-lambda (gensym-f-id) (first params) subs k))
 	((set!) (let ((t1 (gensym 't)))
 		  (walk (car subs)
 			(lambda (r)
@@ -1224,7 +1221,7 @@ EOF
 				     (list (make-node 'set! (list (first params)) (list r))
 					   (k (varnode t1)) ) ) ) ) ) )
 	((##core#foreign-callback-wrapper)
-	 (let ([id (gensym 'f)]
+	 (let ([id (gensym-f-id)]
 	       [lam (first subs)] )
 	   (set! foreign-callback-stubs
 	     (cons (apply make-foreign-callback-stub id params) foreign-callback-stubs) )
@@ -1241,7 +1238,7 @@ EOF
           (t3 (gensym 'r)) )
       (make-node
        'let (list t0)
-       (list (make-node '##core#lambda (list (gensym 'f) #f (list t3) 0) 
+       (list (make-node '##core#lambda (list (gensym-f-id) #f (list t3) 0) 
 			(list (k (varnode t3))) )
 	     (walk-arguments
 	      args
@@ -1255,7 +1252,7 @@ EOF
 	  (t3 (gensym 'r)) )
       (make-node
        'let (list t0)
-       (list (make-node '##core#lambda (list (gensym 'f) #f (list t3) 0) 
+       (list (make-node '##core#lambda (list (gensym-f-id) #f (list t3) 0) 
 			(list (k (varnode t3))) )
 	     (make-node '##core#callunit (list unitname)
 			(list (varnode t0)) ) ) ) ) )
@@ -1688,8 +1685,8 @@ EOF
       (set! direct-call-ids (lset-adjoin eq? direct-call-ids id)) )
 
     ;; Gather free-variable information:
-    ;;   - register direct calls.
-    ;;   - update (by mutation) call information in "##core#call" nodes.
+    ;; (and: - register direct calls
+    ;;       - update (by mutation) call information in "##core#call" nodes)
     (define (gather n here env)
       (let ((subs (node-subexpressions n))
 	    (params (node-parameters n)) )
@@ -1717,13 +1714,16 @@ EOF
 					 [val (and (not (test varname 'unknown)) (test varname 'value))] )
 				    (if (and val (eq? '##core#lambda (node-class val)))
 					(let* ([params (node-parameters val)]
+					       [llist (third params)]
 					       [id (first params)]
 					       [refs (test varname 'references)]
 					       [sites (test varname 'call-sites)] 
 					       [custom
 						(and refs sites
 						     (= (length refs) (length sites)) 
-						     (proper-list? (third params)) ) ] )
+						     (proper-list? llist) ) ] )
+					  (when (and name custom (not (= (length llist) (length (cdr subs)))))
+					    (warning "known procedure called with wrong number of arguments: `~A'" name) )
 					  (register-direct-call! id)
 					  (when custom (register-customizable! varname id)) 
 					  (list id custom) )
@@ -1862,7 +1862,7 @@ EOF
 
 	  ((##core#primitive) 
 	   (make-node '##core#closure '(1) 
-		      (list (make-node '##core#proc params '())) ) )
+		      (list (make-node '##core#proc (list (car params) #t) '())) ) )
 
 	  (else (bomb "bad node (closure2)")) ) ) )
 
@@ -1922,9 +1922,10 @@ EOF
   customizable				; boolean
   rest-argument-mode			; #f | LIST | VECTOR | UNUSED
   body					; expression
-  direct)				; boolean
+  direct				; boolean
+  partition)                            ; integer
   
-(define (prepare-for-code-generation node db)
+(define (prepare-for-code-generation node db partitions)
   (let ([literals '()]
         [lambdas '()]
         [temporaries 0]
@@ -2061,7 +2062,8 @@ EOF
 			   (or direct (get db id 'customizable))
 			   rest-mode
 			   body
-			   direct)
+			   direct
+                           0)
 			  lambdas) )
 		  (set! looping lping)
 		  (set! temporaries temps)
@@ -2161,12 +2163,312 @@ EOF
 		       ((boolean? x) `(bool ,x))
 		       ((char? x) `(char ,x))
 		       ((null? x) '(nil))
+		       ((eof-object? x) '(eof))
 		       (else (bomb "bad immediate (prepare)")) )
 		 '() ) )
+
+    (define (partition-phase! partitions)
+      (define (set-inregion! cells partition)
+        (for-each
+          (lambda (c)
+            (graph-cell-info$$-inregion-set!
+              (graph-cell$$-info c)
+              (eq? (graph*partition c) partition)))
+          cells))
+      (let [(callgraph (make-graph$))]
+        (debugging 'p "partitioning phase...")
+        ;; init call graph
+        (for-each
+          (lambda (l)
+            (graph$-add-cell! callgraph (make-graph-cell$ (lambda-literal-id l))))
+          lambdas)
+        (##sys#hash-table-for-each
+        (lambda (sym plist)
+          (let loop [(es plist)]
+            (if (and (pair? es)
+                     (eq? (caar es) 'contains))
+                (for-each
+                  (lambda (o)
+                    (let [(symcell (graph$-get callgraph sym))]
+                      (and symcell ;; toplevel et al aren't in lambdas
+                           (graph-cell$-add-undirected-edge!
+                             symcell
+                             (graph$-get callgraph o)))))
+                  (cdar es)))
+            (and (pair? es) (loop (cdr es)))))
+        db)
+        ;; partition recursion.  i threw together something with
+        ;; primes to decompose any partition into bi-partitions.  i am
+        ;; sure there is a better way.
+        (randomize 20040619) ;; make partition repeatable
+	(define partition-level 0)
+	(define partition-index 0)
+	(let pway ([factors (sort (factor partitions) <)]
+		   [cells (graph$->list callgraph)])
+	  (cond-expand [testing 
+			(let loop ([n partition-level]) 
+			  (and (> n 0) (display "  ") (loop (sub1 n)))) ]
+		       [else] )
+	  (set! partition-level (add1 partition-level))
+	  (cond
+	   ;; terminate and update
+	   [(null? factors)
+	    (for-each
+	     (lambda (cell)
+	       (let* [(id (graph*id cell))
+		      (ll (member id lambdas
+				  (lambda (x ei)
+				    (eq? x (lambda-literal-id ei)))))]
+		 (and ll
+		      (lambda-literal-partition-set!
+		       (car ll)
+		       partition-index))))
+	     cells)
+	    (set! partition-index (add1 partition-index))]
+	   ;; just continue if 1
+	   [(= (car factors) 1)
+	    (pway (cdr factors) cells)]
+	   ;; bi-partition, balanced 50/50
+	   [(= (car factors) 2)
+	    (partition-fm cells
+			  graph*id graph*color graph*color-set!
+			  graph*partition graph*partition-move!
+			  graph*neighbours
+			  graph*gain
+			  (lambda () (graph*cost callgraph))
+			  (lambda (w n#f n#t) 
+			    (graph*balance cells
+					   (length cells)
+					   w n#f n#t))
+			  ;; 50/50
+			  '(1 . 1)
+			  10)
+
+	    (let ([bpart#f (map (complement graph*partition) cells)]
+		  [bpart#t (map graph*partition cells)])
+	      ;; descend binary tree
+              
+              (set-inregion! cells #f)
+	      (pway (cdr factors) (compress bpart#f cells))
+              
+              (set-inregion! cells #t)
+	      (pway (cdr factors) (compress bpart#t cells)))
+	    ]
+	   ;; bi-partition, but unbalanced
+	   [else
+	    (partition-fm cells
+			  graph*id graph*color graph*color-set!
+			  graph*partition graph*partition-move!
+			  graph*neighbours
+			  graph*gain
+			  (lambda () (graph*cost callgraph))
+			  (lambda (w n#f n#t) 
+			    (graph*balance cells
+					   (length cells)
+					   w n#f n#t))			  
+			  ;; split into 1:(prime-1) ratio
+			  (cons 1 (sub1 (car factors))) 
+			  10)
+	    (let ([bpart#f (map (complement graph*partition) cells)]
+		  [bpart#t (map graph*partition cells)])
+	      ;;; modify binary tree
+              
+              (set-inregion! cells #f)
+	      (pway (cons 1 (cdr factors)) 
+		    (compress bpart#f cells))
+              
+              (set-inregion! cells #t)
+	      (pway (append (factor (sub1 (car factors))) (cdr factors)) 
+		    (compress bpart#t cells)))
+	    ]
+	   )
+	  (set! partition-level (sub1 partition-level))
+	  )
+	(and #f ;; debugging
+	     (for-each
+	      (lambda (l)
+		(display (lambda-literal-id l))
+		(display " in ")
+		(display (lambda-literal-partition l))
+		(newline))
+	      lambdas))
+        (randomize);; reset random
+        ))
     
     (debugging 'p "preparation phase...")
-    (let ((node2 (walk node '() #f '())))
+    (let* ((node2 (walk node '() #f '()))
+	   (t0 (##sys#fudge 6)) )
+      (partition-phase! partitions)
+      (debugging 'b (sprintf "  time needed for partitioning: ~A ms" (- (##sys#fudge 6) t0)))
       (debugging 'o "fast box initializations" fastinits)
       (debugging 'o "fast global references" fastrefs)
       (debugging 'o "fast global assignments" fastsets)
       (values node2 literals lambdas) ) ) )
+
+
+;; factor - Jonah Beckford - slimmed down from SLIB. -*- Hen -*-
+
+;;;; "factor.scm" factorization, prime test and generation
+;;; Copyright (C) 1991, 1992, 1993, 1998 Aubrey Jaffer
+;
+;Permission to copy this software, to modify it, to redistribute it,
+;to distribute modified versions, and to use it for any purpose is
+;granted, subject to the following restrictions and understandings.
+;
+;1.  Any copy made of this software must include this copyright notice
+;in full.
+;
+;2.  I have made no warranty or representation that the operation of
+;this software will be error-free, and I am under no obligation to
+;provide any services, by way of maintenance, update, or otherwise.
+;
+;3.  In conjunction with products arising from the use of this
+;material, there shall be no use of my name in any advertising,
+;promotional, or sales literature without prior written consent in
+;each case.
+
+;;(declare (unit factor)
+;;	 (uses lolevel)
+;;	 (export factor))
+
+;;; prime:products are products of small primes.
+;;; was (comlist:notevery (lambda (prd) (= 1 (gcd n prd))) comps))
+(declare
+  (hide primes-gcd? prime:prime-sqr prime:products prime:sieve prime:f prime:fo prime:fe) )
+
+(define (primes-gcd? n comps)
+  (not (let mapf ((lst comps))
+	 (or (null? lst) (and (= 1 (gcd n (car lst))) (mapf (cdr lst)))))))
+(define prime:prime-sqr 121)
+(define prime:products '(105))
+(define prime:sieve (vector 0 0 1 1 0 1 0 1 0 0 0))
+(letrec ((lp (lambda (comp comps primes nexp)
+	       (cond ((< comp (quotient (##sys#fudge 21) nexp))
+		      (let ((ncomp (* nexp comp)))
+			(lp ncomp comps
+			    (cons nexp primes)
+			    (next-prime nexp (cons ncomp comps)))))
+		     ((< (quotient comp nexp) (* nexp nexp))
+		      (set! prime:prime-sqr (* nexp nexp))
+		      (set! prime:sieve (make-vector nexp 0))
+		      (for-each (lambda (prime)
+				  (vector-set! prime:sieve prime 1))
+				primes)
+		      (set! prime:products (reverse (cons comp comps))))
+		     (else
+		      (lp nexp (cons comp comps)
+			  (cons nexp primes)
+			  (next-prime nexp (cons comp comps)))))))
+	 (next-prime (lambda (nexp comps)
+		       (set! comps (reverse comps))
+		       (do ((nexp (+ 2 nexp) (+ 2 nexp)))
+			   ((not (primes-gcd? nexp comps)) nexp)))))
+  (lp 3 '() '(2 3) 5))
+;;;;Lankinen's recursive factoring algorithm:
+;From: ld231782@longs.LANCE.ColoState.EDU (L. Detweiler)
+
+;                  |  undefined if n<0,
+;                  |  (u,v) if n=0,
+;Let f(u,v,b,n) := | [otherwise]
+;                  |  f(u+b,v,2b,(n-v)/2) or f(u,v+b,2b,(n-u)/2) if n odd
+;                  |  f(u,v,2b,n/2) or f(u+b,v+b,2b,(n-u-v-b)/2) if n even
+
+;Thm: f(1,1,2,(m-1)/2) = (p,q) iff pq=m for odd m.
+
+;It may be illuminating to consider the relation of the Lankinen function in
+;a `computational hierarchy' of other factoring functions.*  Assumptions are
+;made herein on the basis of conventional digital (binary) computers.  Also,
+;complexity orders are given for the worst case scenarios (when the number to
+;be factored is prime).  However, all algorithms would probably perform to
+;the same constant multiple of the given orders for complete composite
+;factorizations.
+
+;Thm: Eratosthenes' Sieve is very roughtly O(ln(n)/n) in time and
+;     O(n*log2(n)) in space.
+;Pf: It works with all prime factors less than n (about ln(n)/n by the prime
+;    number thm), requiring an array of size proportional to n with log2(n)
+;    space for each entry.
+
+;Thm: `Odd factors' is O((sqrt(n)/2)*log2(n)) in time and O(log2(n)) in
+;     space.
+;Pf: It tests all odd factors less than the square root of n (about
+;    sqrt(n)/2), with log2(n) time for each division.  It requires only
+;    log2(n) space for the number and divisors.
+
+;Thm: Lankinen's algorithm is O(sqrt(n)/2) in time and O((sqrt(n)/2)*log2(n))
+;     in space.
+;Pf: The algorithm is easily modified to seach only for factors p<q for all
+;    pq=m.  Then the recursive call tree forms a geometric progression
+;    starting at one, and doubling until reaching sqrt(n)/2, or a length of
+;    log2(sqrt(n)/2).  From the formula for a geometric progression, there is
+;    a total of about 2^log2(sqrt(n)/2) = sqrt(n)/2 calls.  Assuming that
+;    addition, subtraction, comparison, and multiplication/division by two
+;    occur in constant time, this implies O(sqrt(n)/2) time and a
+;    O((sqrt(n)/2)*log2(n)) requirement of stack space.
+
+(define (prime:f u v b n)
+  (if (<= n 0)
+      (cond ((negative? n) #f)
+	    ((= u 1) #f)
+	    ((= v 1) #f)
+	    ; Do both of these factors need to be factored?
+	    (else (append (or (prime:f 1 1 2 (quotient (- u 1) 2))
+			      (list u))
+			  (or (prime:f 1 1 2 (quotient (- v 1) 2))
+			      (list v)))))
+      (if (even? n)
+	  (or (prime:f u v (+ b b) (quotient n 2))
+	      (prime:f (+ u b) (+ v b) (+ b b) (quotient (- n (+ u v b)) 2)))
+	  (or (prime:f (+ u b) v (+ b b) (quotient (- n v) 2))
+	      (prime:f u (+ v b) (+ b b) (quotient (- n u) 2))))))
+
+(define (prime:fo m)
+  (let* ((s (gcd m (car prime:products)))
+	 (r (quotient m s)))
+    (if (= 1 s)
+	(or (prime:f 1 1 2 (quotient (- m 1) 2)) (list m))
+	(append
+	 (if (= 1 r) '()
+	     (or (prime:f 1 1 2 (quotient (- r 1) 2)) (list r)))
+	 (or (prime:f 1 1 2 (quotient (- s 1) 2)) (list s))))))
+
+(define (prime:fe m)
+  (if (even? m)
+      (cons 2 (prime:fe (quotient m 2)))
+      (if (eqv? 1 m)
+	  '()
+	  (prime:fo m))))
+
+;;@body
+;;Returns a list of the prime factors of @1.  The order of the
+;;factors is unspecified.  In order to obtain a sorted list do
+;;@code{(sort! (factor @var{k}) <)}.
+(define (factor k)
+  (case k
+    ((-1 0 1) (list k))
+    (else (if (negative? k)
+	      (cons -1 (prime:fe (- k)))
+	      (prime:fe k)))))
+
+
+;;; Default user-pass (2):
+
+(define (default-user-pass-2 node get put)
+  (define (walkrest n)
+    (for-each (cut walk <>) (node-subexpressions n)) )
+  (define (walk n)
+    (let ([n2 (and (eq? '##core#call (node-class n))
+		   (let ([arg (first (node-subexpressions n))])
+		     (and (eq? '##core#variable (node-class arg))
+			  (let* ([sym (first (node-parameters arg))]
+				 [opt (##sys#hash-table-ref user-level-optimizers sym)] )
+			    (and opt
+				 (let ([result (opt n db-ref: get db-set: put)])
+				   (debugging 'o "user-level optimizer triggered" sym (if result #t #f))
+				   result) ) ) ) ) ) ] )
+      (if n2
+	  (begin
+	    (copy-node! n2 n)
+	    (walk n2) )
+	  (walkrest n) ) ) ) 
+  (walk node) )
