@@ -1,6 +1,6 @@
 ;;;; csi.scm - Simple interpreter stub for CHICKEN
 ;
-; Copyright (c) 2000-2003, Felix L. Winkelmann
+; Copyright (c) 2000-2004, Felix L. Winkelmann
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -28,15 +28,17 @@
 ; felix@call-with-current-continuation.org
 ;
 ; Felix L. Winkelmann
-; Steinweg 1A
-; 37130 Gleichen, OT Weissenborn
+; Unter den Gleichen 1
+; 37130 Gleichen
 ; Germany
 
 
+(declare (uses syntax-case match-support))
+
 (declare
-  (uses extras srfi-1 syntax-case srfi-4 srfi-18 srfi-37 format match-support tinyclos regex script-utils)
   (usual-integrations)
   (disable-interrupts)
+  (run-time-macros)
   (foreign-declare #<<EOF
 #ifdef C_USE_C_DEFAULTS
 # include "c_defaults.h"
@@ -56,20 +58,21 @@
 EOF
 ) )
 
-(include "moremacros")
-(include "match")
+(include "chicken-more-macros")
+(include "chicken-match-macros")
+(include "banner")
 
 #{csi 
   max-history-count print-usage print-banner write-to-echo-port
-  run hexdump delq transcript-port
+  run hexdump del transcript-port
   trace-hl-macros parse-option-string chop-separator lookup-script-file
   report describe dump hexdump bytevector-data get-config
-  advised-procedures do-advise do-unadvise
+  advised-procedures do-advise do-unadvise deldups
   trace-indent trace-indent-level traced-procedures traced-procedure-entry traced-procedure-exit}
 
 (declare
-  (hide parse-option-string trace-hl-macros bytevector-data transcript-port slibpath
-	write-to-echo-port trace-indent setup-usage) )
+  (hide parse-option-string trace-hl-macros bytevector-data transcript-port banner
+	write-to-echo-port trace-indent) )
 
 
 ;;; Parameters:
@@ -81,18 +84,7 @@ EOF
 
 (define-foreign-variable installation-home c-string "C_INSTALL_HOME")
 
-(define slibpath (getenv "SCHEME_LIBRARY_PATH"))
-
-
-;;; Declare used units as provided:
-
-(provide
- 'extras 'srfi-1 'syntax-case 'srfi-4
- 'srfi-18 'format 'tinyclos 'regex
- 'script-utils)
-
-(when (memq #:posix (features)) (provide 'posix))
-(when (memq #:tcp (features)) (provide 'tcp))
+(set! ##sys#features (cons #:csi ##sys#features))
 
 
 ;;; Print all sorts of information:
@@ -112,38 +104,21 @@ Usage: csi {FILENAME | OPTION}
     -no-init                    do not load initialization file `~/#{init-file}'
     -no-feature SYMBOL          unregister feature identifier
     -batch                      terminate after command-line processing
-    -hygienic                   install SYNTAX-CASE macro package
+    -syntax  -hygienic          install SYNTAX-CASE macro package
     -no-warnings                disable all warnings
-    -keyword-style STYLE        enable alternative keyword-syntax (prefix or suffix)
+    -keyword-style STYLE        enable alternative keyword-syntax (none, prefix or suffix)
     -r5rs                       equivalent to `-hygienic -strict'
     -script PATHNAME            use interpreter for shell scripts
     -script-meta PATHNAME       `-script' with extra options
-    -srfi-7                     process source files as SRFI-7 config. language
     -strict                     disable non-standard macros and symbol syntax
     -strict-srfi-0              disable non-standard macros except `cond-expand'
     -strict-reader              disable non-standard read syntax
     -strict-letrec              enable fully R5RS conformant `letrec'
-    -setup [OPTION | EXTENSION] unpack extension ("egg") and run setup script
-    -setup-help                 show list of available setup options
-    -visit FILENAME             visit (macroexpand) source file
-    -slib                       enable SLIB support (needs additional files)
     --                          ignore all following options
 
 EOF
 ) )
 
-(define (setup-usage)
-  (print #<<EOF
-usage: csi -setup [EXTENSION] [MODE]
-
-  -i  --install        install extension, build if needed
-  -b  --build          build extension, extract if needed
-  -w  --wrap           create wrapped package
-  -u  --uninstall      uninstall extension
-EOF
-  )
-  (exit 0) )
-
 (define print-banner
   (let ([config 
 	 (let ([bp (build-platform)]
@@ -156,14 +131,13 @@ EOF
 		 (string-append (##sys#slot x 1) "-") ) )
 	   (string-append (str sv) (str st) (str bp) (##sys#slot mt 1)) ) ] )
     (lambda ()
-      (printf "; This is the CHICKEN interpreter.~%~
-               ; Version ~S, Build ~S - ~A~%~
-               ; (c)2000-2003 Felix L. Winkelmann~%"
-	      build-version build-number config) ) ) )
+      (print banner "\nVersion " build-version ", Build " build-number " - " config "\n(c)2000-2004 Felix L. Winkelmann") ) ) )
 
 
 ;;; Reader for she-bang comment:
 
+(set! ##sys#she-bang-checked #f)
+
 (set! ##sys#user-read-hook
   (let ([read-char read-char]
 	[read read]
@@ -171,7 +145,9 @@ EOF
     (lambda (char port)
       (if (char=? char #\!)
 	  (do ((c (read-char port) (read-char port)))
-	      ((char=? c #\newline) (read port))
+	      ((char=? c #\newline) 
+	       (set! ##sys#she-bang-checked #t)
+	       (read port) )
 	    (when (eof-object? c) (##sys#error "unexpected end of file")) )
 	  (old-hook char port) ) ) ) )
 
@@ -234,9 +210,7 @@ EOF
 	(write write)
 	(printf printf)
 	(integer? integer?)
-	(second second)
 	(butlast butlast)
-	(take take)
 	(list-ref list-ref)
 	(values values)
 	(history-list '())
@@ -251,9 +225,9 @@ EOF
 
       (define (add-h form results)
 	(cond ((fx>= history-count max-history-count)
-	       (set! history-list 
-		 (cons (cons form results) 
-		       (take history-list (fx- history-count 1)) ) ) )
+	       (let ([tl (list-tail history-list (fx- history-count 1))])
+		 (set-cdr! tl '())
+		 (set! history-list (cons (cons form results) history-list)) ) )
 	      (else
 	       (set! history-list (cons (cons form results) history-list))
 	       (set! history-count (fx+ history-count 1)) ) ) )
@@ -279,11 +253,11 @@ EOF
       (cond ((eof-object? form) (exit))
 	    ((and (pair? form)
 		  (eq? 'unquote (##sys#slot form 0)) )
-	     (let ((cmd (second form)))
+	     (let ((cmd (cadr form)))
 	       (case cmd
 		 ((x)
 		  (let ([x (read)])
-		    (add-h `(pretty-print (##sys#secondary-macroexpand (macroexpand ',x))) (list (void)))
+		    (add-h `(pretty-print (macroexpand ',x)) (list (void)))
 		    (pretty-print (macroexpand x))
 		    (void) ) )
 		 ((p)
@@ -313,11 +287,6 @@ EOF
 		    (add-h `(load ,fn) (list (void)))
 		    (load fn)
 		    (void) ) )
-		 ((l7)
-		  (let ((fn (filename (read))))
-		    (add-h `(load-srfi-7-program ,fn) (list (void)))
-		    (load-srfi-7-program fn)
-		    (void) ) )
 		 ((ln) 
 		  (let ((fn (filename (read))))
 		    (add-h `(load-noisily ,fn) (list (void)))
@@ -367,7 +336,6 @@ EOF
  ,q            Quit interpreter
  ,l FILENAME   Load file with given filename (may be symbol or string)
  ,ln FILENAME  Load file and print result of each top-level expression
- ,l7 FILENAME  Process file as SRFI-7 configuration language
  ,r            Show system information
  ,s STRING     Execute shell-command
  ,t EXP        Evaluate form and print elapsed time
@@ -419,18 +387,18 @@ EOF
   (let ([display display])
     (lambda ()
       (when (or (##sys#fudge 12) (##sys#tty-port? (current-input-port)))
-	(display ">>> ") 
+	(display (repl-prompt))
 	(flush-output) ) ) ) )
 
 
 ;;; Tracing/advise:
 
-(define (delq x lst)
+(define (del x lst tst)
   (let loop ([lst lst])
     (if (null? lst)
 	'()
 	(let ([y (car lst)])
-	  (if (eq? x y)
+	  (if (tst x y)
 	      (cdr lst)
 	      (cons y (loop (cdr lst))) ) ) ) ) )
 
@@ -477,7 +445,7 @@ EOF
   (let ([a (assq name advised-procedures)])
     (if a
 	(begin
-	  (set! advised-procedures (delq a advised-procedures))
+	  (set! advised-procedures (del a advised-procedures eq?))
 	  (##sys#slot a 1) )
 	(begin
 	  (printf "Procedure `~S' was never advised.~%" name)
@@ -510,7 +478,7 @@ EOF
 		   (if ,old
 		       (set! ,s (cdr ,old))
 		       (printf "Procedure `~s' was not traced.~%" ,var) )
-		   (set! traced-procedures (delq ,old traced-procedures) ) ) ) )
+		   (set! traced-procedures (del ,old traced-procedures eq?) ) ) ) )
 	    names) ) )
 
 (define-macro (advise name mode proc)
@@ -573,7 +541,7 @@ EOF
 	 (syntax
 	  (let ([tmp (assq 'name ##csi#traced-procedures)] ...)
 	    (if tmp (set! name (cdr tmp))) ...
-	    (set! ##csi#traced-procedures (##csi#delq tmp ##csi#traced-procedures)) ...) ) ) ] ) ) )
+	    (set! ##csi#traced-procedures (delq tmp ##csi#traced-procedures eq?)) ...) ) ) ] ) ) )
 
 (define-syntax advise
   (syntax-rules (before after around)
@@ -608,79 +576,78 @@ EOF
 ;;; Parse options from string:
 
 (define (parse-option-string str)
-  (with-input-from-string str
-    (lambda ()
-      (map (lambda (o)
-	     (if (string? o)
-		 o
-		 (sprintf "~S" o) ) )
-	   (handle-exceptions ex (##sys#error "invalid option syntax" str)
-	     (read-file) ) ) ) ) )
+  (let ([ins (open-input-string str)])
+    (map (lambda (o)
+	   (if (string? o)
+	       o
+	       (let ([os (open-output-string)])
+		 (write o os)
+		 (get-output-string os) ) ) )
+	 (handle-exceptions ex (##sys#error "invalid option syntax" str)
+	   (do ([x (read ins) (read ins)]
+		[xs '() (cons x xs)] )
+	       ((eof-object? x) (reverse xs)) ) ) ) ) )
 
 
 ;;; Print status information:
 
 (define report
-  (let ([format format]
+  (let ([printf printf]
 	[with-output-to-port with-output-to-port] 
 	[current-output-port current-output-port] )
     (lambda port
       (with-output-to-port (if (pair? port) (car port) (current-output-port))
 	(lambda ()
+	  (gc)
 	  (let ([sinfo (##sys#symbol-table-info)]
-		[minfo (##sys#memory-info)] )
+		[minfo (memory-statistics)] )
 	    (define (shorten n) (/ (truncate (* n 100)) 100))
-	    (let-values ([(sysname nodename release version machine) 
-			  (cond-expand [posix (system-information)]
-				       [else (values "<unknown>" "" "" "" "")] ) ] )
-	      (format #t "Features:~%~{  ~A~%~}~%~
-                   Host system information:~%  ~A ~A ~A ~A ~A~%~%~
+	    (printf "Features:~%")
+	    (for-each
+	     (cut print #\tab <>) 
+	     (map keyword->string ##sys#features) )
+	    (printf "~%~
                    Machine type:    \t~A ~A~%~
                    Software type:   \t~A~%~
                    Software version:\t~A~%~
-                   Build platform:  \t~A~%~%~
-                   Current working dir:\t~A~%~
-                   Include path:~%~{  ~A~%~}~%~
+                   Build platform:  \t~A~%~
+                   Include path:    \t~A~%~
                    Symbol-table load:\t~S~%  ~
                      Avg bucket length:\t~S~%  ~
                      Total symbols:\t~S~%~%~
-                   Memory:\t~S heap~A, ~S stack, stack grows ~A~%~
-                   Bytes free in heap:\t~S~%"
-		      (sort (features) (lambda (x y) (string<? (symbol->string x) (symbol->string y))))
-		      sysname nodename release version machine
-		      (machine-type)
-		      (if (##sys#fudge 3) "(64-bit)" "")
-		      (software-type)
-		      (software-version)
-		      (build-platform)
-		      (cond-expand [posix (current-directory)]
-				   [else "<unknown>"] )
-		      ##sys#include-pathnames
-		      (shorten (vector-ref sinfo 0))
-		      (shorten (vector-ref sinfo 1))
-		      (vector-ref sinfo 2)
-		      (vector-ref minfo 0)
-		      (if (##sys#fudge 17) " (fixed)" "")
-		      (vector-ref minfo 1)
-		      (if (= 1 (##sys#fudge 18)) "downward" "upward")
-		      (gc) )
-	      (newline)
-	      (when (##sys#fudge 14) (display "interrupts are enabled\n"))
-	      (when (##sys#fudge 15) (display "symbol gc is enabled\n")) 
-	      (##core#undefined) ) ) ) ) ) ) )
+                   Memory:\theap size is ~S bytes~A with ~S bytes currently in use~%~  
+                     nursery size is ~S bytes, stack grows ~A~%"
+		    (machine-type)
+		    (if (##sys#fudge 3) "(64-bit)" "")
+		    (software-type)
+		    (software-version)
+		    (build-platform)
+		    ##sys#include-pathnames
+		    (shorten (vector-ref sinfo 0))
+		    (shorten (vector-ref sinfo 1))
+		    (vector-ref sinfo 2)
+		    (vector-ref minfo 0)
+		    (if (##sys#fudge 17) " (fixed)" "")
+		    (vector-ref minfo 1)
+		    (vector-ref minfo 2)
+		    (if (= 1 (##sys#fudge 18)) "downward" "upward") )
+	    (newline)
+	    (when (##sys#fudge 14) (display "interrupts are enabled\n"))
+	    (when (##sys#fudge 15) (display "symbol gc is enabled\n")) 
+	    (##core#undefined) ) ) ) ) ) )
 
 
 ;;; Describe & dump:
 
 (define bytevector-data
-  `((u8vector "vector of unsigned bytes" ,u8vector-length ,u8vector-ref)
-    (s8vector "vector of signed bytes" ,s8vector-length ,s8vector-ref)
-    (u16vector "vector of unsigned 16-bit words" ,u16vector-length ,u16vector-ref)
-    (s16vector "vector of signed 16-bit words" ,s16vector-length ,s16vector-ref)
-    (u32vector "vector of unsigned 32-bit words" ,u32vector-length ,u32vector-ref)
-    (s32vector "vector of signed 32-bit words" ,s32vector-length ,s32vector-ref)
-    (f32vector "vector of 32-bit floats" ,f32vector-length ,f32vector-ref)
-    (f64vector "vector of 64-bit floats" ,f64vector-length ,f64vector-ref) ) )
+  '((u8vector "vector of unsigned bytes" u8vector-length u8vector-ref)
+    (s8vector "vector of signed bytes" s8vector-length s8vector-ref)
+    (u16vector "vector of unsigned 16-bit words" u16vector-length u16vector-ref)
+    (s16vector "vector of signed 16-bit words" s16vector-length s16vector-ref)
+    (u32vector "vector of unsigned 32-bit words" u32vector-length u32vector-ref)
+    (s32vector "vector of signed 32-bit words" s32vector-length s32vector-ref)
+    (f32vector "vector of 32-bit floats" f32vector-length f32vector-ref)
+    (f64vector "vector of 64-bit floats" f64vector-length f64vector-ref) ) )
 
 (define-constant max-describe-lines 40)
 
@@ -691,8 +658,6 @@ EOF
 	[length length]
 	[list-ref list-ref]
 	[string-ref string-ref]
-	[describe-object describe-object]
-	[instance? instance?]
 	[newline newline]
 	[hash-table-for-each hash-table-for-each] 
 	[void void] )
@@ -752,7 +717,7 @@ EOF
 			(##sys#slot x 7)
 			(##sys#slot x 3)
 			(##sys#peek-unsigned-integer x 0) ) ]
-	      [(instance? x)
+	      [(and (memq #:tinyclos ##sys#features) (instance? x))
 	       (describe-object x out) ]
 	      [(##sys#locative? x)
 	       (fprintf out "locative~%  pointer ~X~%  index ~A~%  type ~A~%"
@@ -803,7 +768,7 @@ EOF
 	       (let ([st (##sys#slot x 0)])
 		 (let ([data (assq st bytevector-data)] )
 		   (if data
-		       (apply descseq (append (cdr data) (list 0)))
+		       (apply descseq (append (map eval (cdr data)) (list 0)))
 		       (begin
 			 (fprintf out "structure of type ~S:~%" (##sys#slot x 0))
 			 (descseq #f ##sys#size ##sys#slot 1) ) ) ) ) ]
@@ -869,13 +834,68 @@ EOF
 	(write-char #\newline out) ) ) ) )
 
 
+;;; emacs interface (suggested by Linh Dang)
+
+(define ##csi#find-symbol-table (foreign-lambda c-pointer "C_find_symbol_table" c-string))
+(define ##csi#enum-symbols! (foreign-lambda scheme-object "C_enumerate_symbols" c-pointer scheme-object))
+
+(define (##csi#oblist)
+  (let ([it (cons -1 '())]
+	[ns (##csi#find-symbol-table ".")] )
+    (let loop ([lst '()])
+      (let ([s (##csi#enum-symbols! ns it)])
+	(if s
+	    (loop (cons s lst))
+	    lst) ) ) ) )
+
+(define (##csi#oblist->strings)
+  (let ([it (cons -1 '())]
+	[ns (##csi#find-symbol-table ".")] )
+    (let loop ([lst '()])
+      (let ([s (##csi#enum-symbols! ns it)])
+	(if s
+	    (loop (cons (->string s) lst))
+	    lst) ) ) ) )
+
+(define (##csi#name-of-symbols-starting-with prefix)
+  (let ([it (cons -1 '())]
+	[ns (##csi#find-symbol-table ".")]
+	[re (string-append "^" prefix)] )
+    (let loop ([lst '()])
+      (let ([s (##csi#enum-symbols! ns it)])
+	(if s 
+	    (loop (if (string-search re (symbol->string s))
+		      (cons (symbol->string s) lst)
+		      lst))
+	    lst) ) ) ) )
+
+(define (##csi#symbols-matching re-string)
+  (let ([it (cons -1 '())]
+	[ns (##csi#find-symbol-table ".")] )
+    (let loop ([lst '()])
+      (let ([s (##csi#enum-symbols! ns it)])
+	(if s 
+	    (loop (if (string-search re-string (symbol->string s))
+		      (cons s lst)
+		      lst))
+	    lst) ) ) ) )
+
+
 ;;; Start interpreting:
 
+(define (deldups lis . maybe-=)
+  (let ((elt= (:optional maybe-= equal?)))
+    (let recur ((lis lis))
+      (if (null? lis) lis
+	  (let* ((x (car lis))
+		 (tail (cdr lis))
+		 (new-tail (recur (del x tail elt=))))
+	    (if (eq? tail new-tail) lis (cons x new-tail)))))))
+
 (define (run)
   (let* ([extraopts (parse-option-string (or (getenv "CSI_OPTIONS") ""))]
 	 [args (cdr (argv))]
 	 [loadlater #f]
-	 [setup (member "-setup" args)]
 	 [kwstyle (member "-keyword-style" args)]
 	 [script (or (member "-script" args)
 		     (member "-script-meta" args) ) ] )
@@ -896,15 +916,14 @@ EOF
 	   (set! args (append extraopts args))
 	   (and-let* ([p (member "--" args)])
 	     (set-cdr! p '()) ) ] )
-    (let* ([batch (or (member "-batch" args) script setup)]
-	   [quiet (or (member "-quiet" args) script setup)]
-	   [srfi7 (member "-srfi-7" args)]
+    (let* ([batch (or (member "-batch" args) script)]
+	   [quiet (or (member "-quiet" args) script)]
 	   [r5rs (member "-r5rs" args)]
 	   [strict (or r5rs (member "-strict" args))]
 	   [strictr (member "-strict-reader" args)]
 	   [strict-srfi0 (member "-strict-srfi-0" args)]
 	   [strict-letrec (or strict strict-srfi0 (member "-strict-letrec" args))]
-	   [hygienic (or r5rs (member "-hygienic" args))]
+	   [hygienic (or r5rs (member "-hygienic" args) (member "-syntax" args))]
 	   [home (chop-separator (or (getenv "CHICKEN_HOME") installation-home ""))]
 	   [ipath (map chop-separator (string-split (or (getenv "CHICKEN_INCLUDE_PATH") "") ";"))] )
       
@@ -923,110 +942,90 @@ EOF
 	  (when (file-exists? fn) 
 	    (load fn) ) ) )
 
-      (when (member "-help" args)
+      (when (or (member "-help" args) (member "--help" args))
 	(print-usage)
 	(exit 0) )
-      (when (or (member "-setup-help" args) (member "-help-setup" args))
-	(setup-usage) )
       (when (member "-version" args)
 	(print-banner)
 	(exit 0) )
       (when (member "-no-warnings" args)
-	(unless quiet (display "; Warnings are disabled\n"))
+	(unless quiet (display "Warnings are disabled\n"))
 	(set! ##sys#warnings-enabled #f) )
       (unless quiet
 	(load-verbose #t)
 	(print-banner) )
       (when (member "-case-insensitive" args)
-	(unless quiet (display "; Identifiers and symbols are case insensitive\n"))
+	(unless quiet (display "Identifiers and symbols are case insensitive\n"))
 	(register-feature! 'case-insensitive)
 	(case-sensitive #f) )
       (for-each register-feature! (collect-options "-feature"))
       (for-each unregister-feature! (collect-options "-no-feature"))
       (set! ##sys#include-pathnames 
-	(delete-duplicates
+	(deldups
 	 (append (map chop-separator (collect-options "-include-path"))
 		 ##sys#include-pathnames
 		 ipath
 		 (if home (list home) '()) )
 	 string=?) )
-      (unless hygienic (provide 'match))
+      (unless hygienic
+	(set! ##sys#features (cons #:match (cons #:match-support ##sys#features)))
+	(provide 'match) )
       (cond [hygienic
 	     (unless quiet
-	       (display "; Using hygienic macros")
+	       (display "Using hygienic macros")
 	       (cond [strict (display " - strict mode\n")]
 		     [strict-srfi0 (display " - strict mode with SRFI-0\n")]
 		     [else (newline)] ) )
 	     (##syncase#install-macro-package (or strict strict-srfi0) (not strict) strict-letrec)
 	     (##syncase#install-macro-defs trace-hl-macros) ]
 	    [strict
-	     (unless quiet (display "; Strict mode\n"))
+	     (unless quiet (display "Strict mode\n"))
 	     (set! ##sys#strict-mode #t)
 	     (unregister-feature! #:srfi-8 #:srfi-2 #:srfi-0 #:srfi-16 #:srfi-9 #:srfi-26)
 	     (register-feature! #:strict)
 	     (##sys#undefine-non-standard-macros '()) ]
 	    [strict-srfi0
-	     (unless quiet (display "; Strict mode with SRFI-0\n"))
+	     (unless quiet (display "Strict mode with SRFI-0\n"))
 	     (set! ##sys#strict-mode #t)
 	     (unregister-feature! #:srfi-8 #:srfi-2 #:srfi-16 #:srfi-9 #:srfi-26)
 	     (register-feature! #:strict)
 	     (##sys#undefine-non-standard-macros '(cond-expand)) ] )
-      (when (or strict strictr strict-srfi0)
-	(strict-reader #t)
-	(case-sensitive #f) )
-      (when (and strict-letrec (not hygienic))
-	(##sys#enable-strict-letrec) )
       (when kwstyle
-	(cond [(not-pair? (cdr kwstyle)) 
+	(cond [(not (pair? (cdr kwstyle)))
 	       (##sys#error "missing argument to `-keyword-style' option") ]
 	      [(string=? "prefix" (cadr kwstyle))
 	       (keyword-style #:prefix) ]
+	      [(string=? "none" (cadr kwstyle))
+	       (keyword-style #:none) ]
 	      [(string=? "suffix" (cadr kwstyle))
 	       (keyword-style #:suffix) ] ) )
-      (when (member "-slib" args)
-	(unless slibpath
-	  (##sys#error "you need to set the SCHEME_LIBRARY_PATH environment variable") )
-	(load (make-pathname slibpath "chicken-slib")) )
-      (unless (or (member "-no-init" args) script setup) (loadinit))
-      (if setup
-	  (if (##sys#provided? 'chicken-setup)
-	      (match (cdr setup)
-		[() (initialize-setup-stuff)]
-		[("-init") (initialize-registry!)]
-		[(ext . args) (run-setup-script ext args)]
-		[_ (##sys#error "invalid command-line format for `-setup' option")] )
-	      (begin
-		(print "Sorry, the `-setup' option is not available on this platform.")
-		(exit 1) ) )
-	  (do ([args args (cdr args)])
-	      ((null? args)
-	       (when loadlater (load loadlater))
-	       (unless batch (read-eval-print-loop)) )
-	    (let* ([arg (car args)]
-		   [len (string-length arg)] )
-	      (cond ((member 
-		      arg 
-		      '("--" "-batch" "-quiet" "-no-init" "-hygienic" "-no-warnings" "-script"
-			"-strict-reader" "-script-meta" "-strict-letrec" "-slib" "-r5rs"
-			"-case-insensitive" "-srfi-7" "-strict" "-strict-srfi-0") ) )
-		    ((member arg '("-feature" "-no-feature" "-include-path" "-keyword-style"))
-		     (set! args (cdr args)) )
-		    ((string=? "-visit" arg)
-		     (visit (cadr args))
-		     (set! args (cdr args)) )
-		    ((string=? "-eval" arg)
-		     (let ([in (open-input-string (cadr args))])
-		       (do ([x (read in) (read in)])
-			   ((eof-object? x))
-			 (eval x) )
-		       (set! args (cdr args)) ) )
-		    ((and (>= len 2) (string=? "-:" (substring arg 0 2))))
-		    (srfi7
-		     (eval 
-		      (##sys#expand-srfi-7-program
-		       (chop-separator arg)
-		       (with-input-from-file arg read)
-		       ##sys#error) ) )
-		    (else (load arg) ) ) ) ) ) ) ) )
+      (when (or strict strictr strict-srfi0)
+	(strict-reader #t)
+	(keyword-style #:none)
+	(case-sensitive #f) )
+      (when (and strict-letrec (not hygienic))
+	(##sys#enable-strict-letrec) )
+      (unless (or (member "-no-init" args) script) (loadinit))
+      (do ([args args (cdr args)])
+	  ((null? args)
+	   (when loadlater (load loadlater))
+	   (unless batch (repl)) )
+	(let* ([arg (car args)]
+	       [len (string-length arg)] )
+	  (cond ((member 
+		  arg 
+		  '("--" "-batch" "-quiet" "-no-init" "-hygienic" "-syntax" "-no-warnings" "-script"
+		    "-strict-reader" "-script-meta" "-strict-letrec" "-r5rs"
+		    "-case-insensitive" "-strict" "-strict-srfi-0") ) )
+		((member arg '("-feature" "-no-feature" "-include-path" "-keyword-style"))
+		 (set! args (cdr args)) )
+		((string=? "-eval" arg)
+		 (let ([in (open-input-string (cadr args))])
+		   (do ([x (read in) (read in)])
+		       ((eof-object? x))
+		     (eval x) )
+		   (set! args (cdr args)) ) )
+		((and (>= len 2) (string=? "-:" (substring arg 0 2))))
+		(else (load arg) ) ) ) ) ) ) )
 
 (run)
