@@ -1,6 +1,6 @@
 ;;;; easyffi.scm
 ;
-; Copyright (c) 2000-2003, Felix L. Winkelmann
+; Copyright (c) 2000-2004, Felix L. Winkelmann
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -28,48 +28,56 @@
 ; felix@call-with-current-continuation.org
 ;
 ; Felix L. Winkelmann
-; Steinweg 1A
-; 37130 Gleichen, OT Weissenborn
+; Unter den Gleichen 1
+; 37130 Gleichen
 ; Germany
 
-#{compiler debugging-chicken register-ffi-macro parse-easy-ffi ffi-include-path foreign-type-declaration
-	   foreign-declarations}
+#{compiler debugging-chicken used-units register-ffi-macro parse-easy-ffi ffi-include-path-list foreign-type-declaration 
+	   foreign-declarations explicit-use-flag check-c-syntax no-c-syntax-checks}
 
 (declare
   (unit easyffi)
   (compress-literals)
-  ;(no-bound-checks) (no-procedure-checks)
-  (export parse-easy-ffi register-ffi-macro) )
+  (no-bound-checks)
+  (no-procedure-checks)
+  (export parse-easy-ffi register-ffi-macro used-units foreign-declarations number-type ffi-include-path-list
+	  foreign-type-declaration explicit-use-flag debugging-chicken check-c-syntax no-c-syntax-checks) )
 
 (eval-when (compile) (match-error-control #:unspecified))
 
 (include "easyffi.l.silex")
 
+(define use-finalizers #f)
+(define exception-handler #f)
+(define destructor-name 'destroy)
 (define pp-mode #f)
 (define processed-output '())
 (define macro-table '((|CHICKEN| * ())))
-(define return-type #f)
 (define pp-conditional-stack '())
 (define pp-process #t)
 (define type-map '())
-(define ffi-include-path '("."))
+(define ffi-include-path-list '("."))
 (define export-constants #f)
 (define prefix #f)
 (define name-substitution-rxs '())
 (define name-substitution-repls '())
 (define declared-types '())
 (define rename-list '())
+(define abstract-classes '())
+(define full-specialization #f)
+(define defined-enums '())
+(define parsing-error quit)
 
 (define (pp-token tok)
   (if pp-mode
       tok
-      (quit "FFI lexer error in line ~A: preprocessor-command out of context: ~S" ##sys#read-line-counter tok) ) )
+      (quit "FFI lexer error: preprocessor-command out of context: ~S" tok) ) )
 
 (define (pp-token2 tok tok2)
   (if pp-mode tok tok2) )
 
 (define (lexer-error)
-  (quit "FFI lexer error in line ~A: illegal character" ##sys#read-line-counter) )
+  (parsing-error "FFI lexer error: illegal character") )
 
 (define (chunkify)
   (let rec ([scope 0])
@@ -84,7 +92,7 @@
 	    [(pp-define pp-include pp-if pp-ifdef pp-ifndef pp-else pp-endif pp-undef pp-pragma pp-error)
 	     (loop 'pp (list t)) ]
 	    [(close-curly)
-	     (cond [(not (positive? scope)) (quit "`}' out of context")]
+	     (cond [(not (positive? scope)) (parsing-error "`}' out of context")]
 		   [(null? tokens) (reverse chunks)]
 		   [else (cons (reverse tokens) chunks)] ) ]
 	    [(open-curly)
@@ -101,7 +109,7 @@
 	     (loop 'declare '(declare)) ]
 	    [(semicolon)
 	     (if mode
-		 (quit "unexpected semicolon")
+		 (parsing-error "unexpected semicolon")
 		 (begin
 		   (set! chunks (cons (reverse tokens) chunks))
 		   (loop #f '()) ) ) ]
@@ -112,11 +120,11 @@
     [() #f]
     [('pp-else)
      (when (null? pp-conditional-stack)
-       (quit "unbalanced preprocessor conditionals") )
+       (parsing-error "unbalanced preprocessor conditionals") )
      (set! pp-process (and (not (car pp-conditional-stack)) (every identity (cdr pp-conditional-stack)))) ]
     [('pp-endif)
      (when (null? pp-conditional-stack)
-       (quit "unbalanced preprocessor conditionals") )
+       (parsing-error "unbalanced preprocessor conditionals") )
      (set! pp-conditional-stack (cdr pp-conditional-stack))
      (set! pp-process (every identity pp-conditional-stack)) ]
     [('pp-ifdef ('id name))
@@ -146,39 +154,58 @@
 	   [('pp-undef ('id name))
 	    (set! macro-table (delete (assq (string->symbol name) macro-table) eq?)) ]
 	   [('pp-error msgs ...)
-	    (quit (string-intersperse (cons "(#error) " (map reparse-item msgs)) "")) ]
+	    (parsing-error (string-intersperse (cons "(#error) " (map reparse-item msgs)) "")) ]
 	   [('pp-include ((or 'string 'i-string) filename))
 	    (let ([fname (resolve-ffi-include-file filename)])
 	      (if fname
 		  (call-with-input-file fname parse-easy-ffi-rec)
-		  (quit "can not open include file `~A'" filename) ) ) ]
+		  (parsing-error "can not open include file `~A'" filename) ) ) ]
 	   [('pp-pragma . more) #f]
 	   [('declare 'open-paren ('id decl) 'comma val 'close-paren)
 	    (parse-declaration decl val) ]
 	   [('declare . _)
-	    (quit "invalid syntax in pseudo declaration: ~S" c) ]
-	   [_ (let loop ([c (subst-macros c)] [cb #f])
-		(match c
-		  [((or 'extern 'static 'volatile 'inline) . more)
-		   (loop more cb) ]
-		  [('callback . more)
-		   (loop more #t) ]
-		  [('const . more)
-		   (parse-constant more) ]
-		  [('enum ('scope more))
-		   (parse-enum-def #f (subst-macros more)) ]
-		  [('enum ('id name) ('scope more))
-		   (parse-enum-def name (subst-macros more)) ]
-		  [('class . more)
-		   (parse-class-def more) ]
-		  [('struct ('id name)) #f]
-		  [('using . more) #f]
-		  [('typedef . more)
-		   (parse-typedef more) ]
-		  [(and more (('id name) . _))
-		   (parse-prototype more cb) ]
-		  [more
-		   (parse-prototype more cb)] ) ) ] ) ) ] ) )
+	    (parsing-error "invalid syntax in pseudo declaration: ~S" c) ]
+	   [_ (let ([cb #f] 
+		    [ab #f]
+		    [sp #f]
+		    [dc #f]
+		    [ds #f] )
+		(let loop ([c (subst-macros c)])
+		  (match c
+		    [((or 'extern 'static 'volatile 'inline) . more)
+		     (loop more) ]
+		    [('abstract . more)
+		     (set! ab #t)
+		     (loop more) ]
+		    [('specialize . more)
+		     (set! sp #t)
+		     (loop more) ]
+		    [('callback . more)
+		     (set! cb #t) 
+		     (loop more) ]
+		    [('discard . more)
+		     (set! ds #t)
+		     (loop more) ]
+		    [('const . more)
+		     (if (memq 'open-paren more)
+			 (parse-prototype c cb sp dc ds)
+			 (begin
+			   (set! dc #t)
+			   (loop more) ) ) ]
+		    [('enum ('scope more))
+		     (parse-enum-def #f (subst-macros more)) ]
+		    [('enum ('id name) ('scope more))
+		     (parse-enum-def name (subst-macros more)) ]
+		    [('class . more)
+		     (parse-class-def more ab) ]
+		    [('struct ('id name)) #f]
+		    [('using . more) #f]
+		    [('typedef . more)
+		     (parse-typedef more) ]
+		    [(and more (('id name) . _))
+		     (parse-prototype more cb sp dc ds) ]
+		    [more
+		     (parse-prototype more cb sp dc ds)] ) ) ) ] ) ) ] ) )
 
 (define parse-again parse)
 
@@ -199,14 +226,17 @@
        (values `(template ,t0 ,@ts) more) ) ]
     [('signed . more) (parse-type-rec more)]
     [`(unsigned fixnum . ,more) (values 'unsigned-int more)]
-    [`(unsigned int . ,more) (values 'unsigned-integer more)]
+    [`(unsigned int . ,more) (values (if (eq? number-type 'fixnum) 'unsigned-int 'unsigned-integer) more)]
     [`(unsigned char . ,more) (values 'unsigned-char more)]
     [`(unsigned short . ,more) (values 'unsigned-short more)]
     [`(unsigned long . ,more) (values 'unsigned-long more)]
     [`(void . ,more) (values 'void more)]
     [`(bool . ,more) (values 'bool more)]
+    [`(unsigned byte . ,more) (values 'unsigned-byte more)]
+    [`(byte . ,more) (values 'byte more)]
+    [`(scheme-value . ,more) (values 'scheme-object more)]
     [`(fixnum . ,more) (values 'int more)]
-    [`(int . ,more) (values 'integer more)]
+    [`(int . ,more) (values (if (eq? number-type 'fixnum) 'int 'integer) more)]
     [`(char . ,more) (values 'char more)]
     [`(short . ,more) (values 'short more)]
     [`(long . ,more) (values 'long more)]
@@ -218,49 +248,68 @@
     [(('id t) . more)
      (let ([st (string->symbol t)])
        (cond [(assq st type-map) => (lambda (a) (values (cdr a) more))]
+	     [(memq st defined-enums) (values `(enum ,t) more)]
 	     [(memq st declared-types) (values st more)]
 	     [else (values t more)] ) ) ]
-    [x (quit "invalid type specifier: ~S" x)] ) )
+    [x (parsing-error "invalid type specifier: ~S" x)] ) )
 
-(define (parse-type ts . rt)
-  (fluid-let ([return-type (:optional rt #f)])
-    (let-values ([(t0 more) (parse-type-rec ts)])
-      (let loop ([t0 t0] [more more])
-	(match more
-	  [('star . more)
-	   (loop `(pointer ,t0) more) ]
-	  [(('op "&") . more)
-	   (loop `(ref ,t0) more) ]
-	  [_ (values
-	      (if return-type
-		  (match t0
-		    ['(pointer char) 'c-string]
-		    ['(pointer (const char)) 'c-string]
-		    [_ t0] )
-		  (let loop ([t1 t0])
-		    (match t1
-		      [`(pointer (const ,t2)) (loop `(pointer ,t2))]
-		      ['(pointer unsigned-fixnum) 'u32vector]
-		      ['(pointer unsigned-integer) 'u32vector]
-		      ['(pointer unsigned-short) 'u16vector]
-		      ['(pointer unsigned-char) 'u8vector]
-		      ['(pointer unsigned-long) 'u32vector]
-		      ['(pointer fixnum) 's32vector]
-		      ['(pointer integer) 's32vector]
-		      ['(pointer int) 's32vector]
-		      ['(pointer short) 's16vector]
-		      ['(pointer char) 'c-string]
-		      ['(pointer long) 's32vector]
-		      ['(pointer float) 'f32vector]
-		      ['(pointer double) 'f64vector]
-		      [_ t0] ) ) )
-	      more) ] ) ) ) ) )
+(define (parse-type ts #!optional return-type discard)
+  (let-values ([(t0 more) (parse-type-rec ts)])
+    (let loop ([t0 t0] [more more])
+      (match more
+	[('star . more)
+	 (loop `(pointer ,t0) more) ]
+	[(('op "&") . more)
+	 (loop `(ref ,t0) more) ]
+	[('open-paren 'star . (or (('id _) 'close-paren 'open-paren . more) ('close-paren 'open-paren . more)))
+	 (let-values ([(ts more) (parse-arglist more)])
+	   (values `(function ,t0 ,ts) more) ) ]
+	[_ (values (simplify-type t0 return-type discard) more)] ) ) ) )
+
+(define (simplify-type t0 return-type discard)
+  (define (strtype) (if discard 'c-string* 'c-string))
+  (if return-type
+      (match t0
+	['(pointer char) (strtype)]
+	['(pointer (const char)) (strtype)]
+	[`(pointer ,(? string? t))
+	 (let ([st (string->symbol t)])
+	   (if (memq st defined-classes) 
+	       `(instance ,t ,(fix-cname t))
+	       t0) ) ] 
+	[_ t0] )
+      (let loop ([t1 t0])
+	(match t1
+	  [`(pointer (const ,t2)) (loop `(pointer ,t2))]
+	  ['(pointer unsigned-fixnum) 'u32vector]
+	  ['(pointer unsigned-integer) 'u32vector]
+	  ['(pointer unsigned-short) 'u16vector]
+	  ['(pointer unsigned-char) 'u8vector]
+	  ['(pointer unsigned-byte) 'u8vector]
+	  ['(pointer byte) 's8vector]
+	  ['(pointer unsigned-long) 'u32vector]
+	  ['(pointer fixnum) 's32vector]
+	  ['(pointer integer) 's32vector]
+	  ['(pointer int) 's32vector]
+	  ['(pointer short) 's16vector]
+	  ['(pointer char) (strtype)]
+	  ['(pointer long) 's32vector]
+	  ['(pointer float) 'f32vector]
+	  ['(pointer double) 'f64vector]
+	  [`(pointer ,(? string? t))
+	   (let ([st (string->symbol t)])
+	     (if (memq st defined-classes) 
+		 `(instance ,t ,(fix-cname t))
+		 t1) ) ] 
+	  [_ t1] ) ) ) )
 
 (define (parse-arglist ts)
   (let rec ([more ts] [args '()])
     (match more
       [('close-paren . more)
        (values (reverse args) more) ]
+      [('dots . _)
+       (parsing-error "varargs are not supported") ]
       [_ (let-values ([(type more) (parse-type more #f)])
 	   (match more
 	     [(('id str) 'comma . more)
@@ -271,7 +320,7 @@
 	      (rec more (cons type args)) ]
 	     [('close-paren . more)
 	      (values (reverse (cons type args)) more) ]
-	     [_ (quit "bad argument list syntax: ~S" more)] ) ) ] ) ) )
+	     [_ (parsing-error "bad argument list syntax: ~S" more)] ) ) ] ) ) )
 
 (define (parse-typelist ts)
   (let rec ([more ts] [ts '()])
@@ -284,7 +333,7 @@
 	      (rec more (cons type ts)) ]
 	     [(('op ">") . more)
 	      (values (reverse (cons type ts)) more) ]
-	     [_ (quit "bad template type list syntax: ~S" more)] ) ) ] ) ) )
+	     [_ (parsing-error "bad template type list syntax: ~S" more)] ) ) ] ) ) )
 
 (define (subst-macros chunk)
   (let loop ([c chunk])
@@ -297,36 +346,37 @@
 	      (cons x (loop more)) ) ) ]
       [(x . y) (cons x (loop y))] ) ) )
 
-(define (parse-prototype ts cb)
-  (let-values ([(rtype more) (parse-type ts #t)])
-    (match more
-      [(('id str) 'open-paren 'void 'close-paren . more)
-       (process-prototype-def rtype (string->symbol str) '() cb)
-       (match more
-	 [(('scope . _) . more) (parse-again more)]
-	 [() #f]
-	 [_ (quit "unexpected tokens: ~S" more)] ) ]
-      [(('id str) 'open-paren . more)
-       (let-values ([(args more) (parse-arglist more)])
-	 (process-prototype-def rtype (string->symbol str) args cb)
-	 (match more
-	   [(('scope . _) . more) (parse-again more)]
-	   [() #f]
-	   [_ (quit "unexpected tokens: ~S" more)] ) ) ]
-      [(('id str) . (or (('op "=") . _) ()))
-       (process-variable-def rtype (string->symbol str)) ]
-      [else (quit "bad prototype syntax `~A'" more)] ) ) )
-
-(define (parse-constant ts)
-  (let-values ([(type more) (parse-type ts #t)])
-    (match more
-      [(('id name) ('op "=") (num n))
-       (process-constant-def (string->symbol name) n) ]
-      [(('id name) . _)
-       (process-macro-def (string->symbol name) type) ] 
-      [_ (quit "syntax error in constant form: ~S" ts)] ) ) )
+(define (parse-prototype ts cb sp const discard)
+  (fluid-let ([full-specialization (or sp full-specialization)])
+    (let-values ([(rtype more) (parse-type ts #t discard)])
+      (let loop ([more more])
+	(match more
+	  [() #f]
+	  [(('id str) ('op "::") . more) #f]
+	  [(('id str) 'open-paren 'void 'close-paren . more)
+	   (process-prototype-def rtype (string->symbol str) '() cb)
+	   (match more
+	     [(('scope . _) . more) (parse-again more)]
+	     [() #f]
+	     [_ (parsing-error "unexpected tokens: ~S" more)] ) ]
+	  [(('id str) 'open-paren . more)
+	   (let-values ([(args more) (parse-arglist more)])
+	     (process-prototype-def rtype (string->symbol str) args cb)
+	     (match more
+	       [(('scope . _) . more) (parse-again more)]
+	       [() #f]
+	       [_ (parsing-error "unexpected tokens: ~S" more)] ) ) ]
+	  [(('id str) 'comma . more)
+	   (process-variable-def rtype (string->symbol str) const)
+	   (loop more) ]
+	  [(('id str))
+	   (process-variable-def rtype (string->symbol str) const) ]
+	  [(('id str) . (or (('op "=") . _) ()))
+	   (process-variable-def rtype (string->symbol str) const) ]
+	  [else (parsing-error "bad prototype syntax `~A'" more)] ) ) ) ) )
 
 (define (parse-enum-def ename ts)
+  (when ename (set! defined-enums (cons (string->symbol ename) defined-enums)))
   (let loop ([ts ts] [i 0] [items '()])
     (match ts
       [('close-curly) #f]
@@ -336,7 +386,7 @@
 	     (match more
 	       [() (process-enum-def ename items)]
 	       [('comma . more) (loop more i items)]
-	       [_ (quit "syntax error in enum form `~A'" more)] ) ) ) ] ) ) )
+	       [_ (parsing-error "syntax error in enum form `~A'" more)] ) ) ) ] ) ) )
 
 (define (parse-enum-item ts i)
   (match ts
@@ -346,24 +396,29 @@
 	 (error "non-exact enum value for `~A'" name) ) ]
     [(('id name) . more)
      (values (string->symbol name) i more) ] 
-    [_ (quit "invalid enum syntax `~A'" ts)] ) )
+    [_ (parsing-error "invalid enum syntax `~A'" ts)] ) )
 
 (define (parse-typedef ts)
   (let-values ([(type more) (parse-type ts #t)])
-    (match more
-      [(('id tname))
-       (set! type-map (alist-cons (string->symbol tname) type type-map)) ]
-      [_ (quit "invalid typedef syntax `~A'" more)] ) ) )
+    (let loop ([more more] [type type])
+      (match more
+	[('star . more)
+	 (loop more `(pointer ,type)) ]
+	[(('id tname))
+	 (set! type-map (alist-cons (string->symbol tname) (simplify-type type #t #f) type-map)) ]
+	[_ (parsing-error "invalid typedef syntax `~A'" more)] ) ) ) )
 
 (define has-constructor #f)
 (define defined-classes '())
 
-(define (parse-class-def ts)
+(define (parse-class-def ts ab)
   (match ts
     [(('id name)) 
      (set! defined-classes (cons (string->symbol name) defined-classes)) ]
     [(('id name) . more)
-     (set! defined-classes (cons (string->symbol name) defined-classes)) 
+     (let ([sym (string->symbol name)])
+       (set! defined-classes (cons sym defined-classes))
+       (when ab (set! abstract-classes (cons sym abstract-classes))) )
      (let loop ([more more] [t '(op ":")] [bases '()])
        (if (and (pair? more) (equal? t (car more)))
 	   (match more
@@ -377,10 +432,11 @@
 		    (if (memq (string->symbol bname) defined-classes)
 			(cons bname bases)
 			bases) ) ]
-	     [_ (quit "invalid class definition for `~A': ~S" name more)] ) 
+	     [_ (parsing-error "invalid class definition for `~A': ~S" name more)] ) 
 	   (match more
 	     [(('scope . chunks))
-	      (let ([cname (fix-cname name)])
+	      (let ([cname (fix-cname name)]
+		    [csname (string->symbol name)] )
 		(process-class-def name cname bases)
 		(fluid-let ([has-constructor #f])
 		  (let ([exp #f])
@@ -389,28 +445,33 @@
 		       (let loop ([more (subst-macros chunk)])
 			 (match more
 			   [() #f]
-			   [((or 'public 'protected) ('op ":") . more) 
+			   [('public ('op ":") . more) 
 			    (set! exp #t)
 			    (loop more) ]
-			   [('private ('op ":") . more) 
+			   [((or 'private 'protected) ('op ":") . more) 
 			    (set! exp #f)
 			    (loop more) ]
 			   [more 
 			    (when exp 
 			      (fluid-let ([parse-again loop])
-				(parse-member-prototype name cname more #f) ) ) ] ) ) )
+				(parse-member-prototype name cname more #f #f) ) ) ] ) ) )
 		     chunks)
-		    (unless has-constructor
+		    (when (and (not has-constructor) (not (memq csname abstract-classes)))
 		      (process-constructor-def name cname '()) ) ) ) ) ]
-	     [_ (quit "invalid class definition for `~A': ~S" name more)] ) ) ) ]
-    [_ (quit "invalid class definition: ~S" ts)] ) )
+	     [_ (parsing-error "invalid class definition for `~A': ~S" name more)] ) ) ) ]
+    [_ (parsing-error "invalid class definition: ~S" ts)] ) )
 
-(define (parse-member-prototype name cname ts cb)
+(define (parse-member-prototype name cname ts cb discard)
   (match ts
+    [('specialize . more)
+     (fluid-let ([full-specialization #t])
+       (parse-member-prototype name cname more #t discard) ) ]
     [('callback . more) 
-     (parse-member-prototype name cname more #t) ]
+     (parse-member-prototype name cname more #t discard) ]
+    [('discard . more)
+     (parse-member-prototype name cname more cb #t) ]
     [((or 'explicit 'virtual) . more)
-     (parse-member-prototype name cname more cb) ]
+     (parse-member-prototype name cname more cb discard) ]
     [(('id name2) 'open-paren 'void 'close-paren . more)
      (if (string=? name2 name)
 	 (begin
@@ -419,8 +480,8 @@
 	   (match more
 	     [(('scope . _) . more) (parse-again more)]
 	     [() #f]
-	     [_ (quit "unexpected tokens: ~S" more)] ) )
-	 (quit "invalid constructor for `~A': ~S" name ts) ) ]
+	     [_ (parsing-error "unexpected tokens: ~S" more)] ) )
+	 (parsing-error "invalid constructor for `~A': ~S" name ts) ) ]
     [(('id name2) 'open-paren . more)
      (if (string=? name2 name)
 	 (let-values ([(args more) (parse-arglist more)])
@@ -429,15 +490,15 @@
 	   (match more
 	     [(('scope . _) . more) (parse-again more)]
 	     [() #f]
-	     [_ (quit "unexpected tokens: ~S" more)] ) )
-	 (quit "invalid constructor for `~A': ~S" name ts) ) ]
+	     [_ (parsing-error "unexpected tokens: ~S" more)] ) )
+	 (parsing-error "invalid constructor for `~A': ~S" name ts) ) ]
     [(('op "~") ('id name2) 'open-paren . (or ('void 'close-paren . more) ('close-paren . more)))
      (if (string=? name2 name)
 	 (match more
 	   [(('scope . _) . more) (parse-again more)]
 	   [() #f]
-	   [_ (quit "unexpected tokens: ~S" more)] )
-	 (quit "invalid destructor for `~A': ~S" name ts) ) ]
+	   [_ (parsing-error "unexpected tokens: ~S" more)] )
+	 (parsing-error "invalid destructor for `~A': ~S" name ts) ) ]
     [('static . more)
      (let-values ([(rtype more) (parse-type more #t)])
        (match more
@@ -446,16 +507,16 @@
 	    (match more
 	      [(('scope . _) . more) (parse-again more)]
 	      [() #f]
-	      [_ (quit "unexpected tokens: ~S" more)] ) ]
+	      [_ (parsing-error "unexpected tokens: ~S" more)] ) ]
 	 [(('id str) 'open-paren . more)
 	  (let-values ([(args more) (parse-arglist more)])
 	    (process-prototype-def rtype (string->symbol (string-append name "::" str)) args cb)
 	    (match more
 	      [(('scope . _) . more) (parse-again more)]
 	      [() #f]
-	      [_ (quit "unexpected tokens: ~S" more)] ) ) ]
-	 [_ (quit "bad static member prototype syntax: ~S" more)] ) ) ]
-    [_ (let-values ([(rtype more) (parse-type ts #t)])
+	      [_ (parsing-error "unexpected tokens: ~S" more)] ) ) ]
+	 [_ (parsing-error "bad static member prototype syntax: ~S" more)] ) ) ]
+    [_ (let-values ([(rtype more) (parse-type ts #t discard)])
 	 (match more
 	   [(('id str) 'open-paren 'void 'close-paren . more)
 	    (process-member-prototype-def name cname rtype (string->symbol str) '() cb)
@@ -466,7 +527,7 @@
 	      (parse-member-body more) ) ]
 	   [(('id str) . (or (('op "=") . _) ()))
 	    #f]				; member variables are ignored
-	   [_ (quit "bad member prototype syntax: ~S" more)] ) ) ] ) )
+	   [_ (parsing-error "bad member prototype syntax: ~S" more)] ) ) ] ) )
 
 (define (parse-member-body ts)
   (let loop ([more ts])
@@ -477,7 +538,7 @@
        (loop more) ]
       [(('scope . _) . more) (parse-again more)]
       [() #f]
-      [_ (quit "unexpected tokens: ~S" more)] ) ) )
+      [_ (parsing-error "unexpected tokens: ~S" more)] ) ) )
 
 (define reparse-item 
   (match-lambda 
@@ -553,20 +614,32 @@
 
 (define (process-prototype-def rtype name args cb)
   (let ([name2 (fix-name name)])
-    (emit `(define ,name2
-	     (,(if cb 'foreign-callback-lambda 'foreign-lambda)
-	      ,rtype ,(->string name) ,@args))) ) )
+    (emit
+     (if (and full-specialization (pair? args))
+	 (let ([slist (gen-spec-list args)]
+	       [tmp (gensym)] )
+	   `(begin
+	      (declare (hide ,tmp))
+	      (define ,tmp
+		(,(if cb 'foreign-callback-lambda 'foreign-lambda)
+		 ,rtype ,(->string name) ,@args) )
+	      (define-method (,name2 ,@slist)
+		(,tmp ,@(unzip1 slist) ) ) ) )
+	 `(define ,name2
+	    (,(if cb 'foreign-callback-lambda 'foreign-lambda)
+	     ,rtype ,(->string name) ,@args))) ) ) )
 
-(define (process-variable-def rtype name)
+(define (process-variable-def rtype name const)
   (let ([tmp (gensym)]
 	[var (gensym)] 
 	[name2 (fix-name name)] )
-    (emit 
-     `(define-foreign-variable ,tmp ,rtype ,(->string name))
-     `(define (,name2 . ,var)
-	(if (pair? ,var)
-	    (set! ,tmp (car ,var))
-	    ,tmp) ) ) ) )
+    (emit `(define-foreign-variable ,tmp ,rtype ,(->string name)))
+    (if const
+	(emit `(define ,name2 ,tmp))
+	(emit `(define (,name2 . ,var)
+		 (if (pair? ,var)
+		     (set! ,tmp (car ,var))
+		     ,tmp) ) ) ) ) )
 
 (define (process-enum-def ename items)
   (for-each
@@ -579,17 +652,20 @@
 
 (define (process-class-def name cname basenames)
   (let ([destr (gensym)]
-	[name2 (fix-name name)] )
+	[name2 (fix-name name)] 
+	[csname (string->symbol name)] )
     (emit
      `(declare (hide ,destr))
-     `(define ,destr (foreign-lambda void "delete " (pointer ,name)))
      `(define-class ,cname
 	,(if (null? basenames)
 	     `(<c++-object>)
 	     (map (lambda (b) (fix-cname b)) (reverse basenames) ) )
-	() )
-     `(define-method (destroy (this ,cname))
-	(,destr (slot-ref this 'this)) ) ) ) )
+	() ) )
+    (unless (memq csname abstract-classes)
+      (emit
+       `(define ,destr (foreign-lambda void "delete " (pointer ,name)))
+       `(define-method (,destructor-name (this ,cname))
+	  (,destr (slot-ref this 'this)) ) ) ) ) )
 
 (define (process-constructor-def name cname args)
   (let ([constr (gensym)])
@@ -598,7 +674,14 @@
      `(define ,constr (foreign-lambda (pointer ,name) ,(string-append "new " name) ,@args))
      `(define-method (initialize (this ,cname) initargs) 
 	;; no CALL-NEXT-METHOD here: we don't want to invoke the base-class constructor.
-	(slot-set! this 'this (apply ,constr initargs)) ) ) ) )
+	,@(if (and use-finalizers (not (memq (string->symbol name) abstract-classes)))
+	      `((set-finalizer! this ,destructor-name))
+	      '() )
+	(slot-set! 
+	 this 'this
+	 (if (and (pair? initargs) (eq? 'this (##sys#slot initargs 0)))
+	     (cadr initargs)
+	     (apply ,constr initargs) ) ) ) ) ) )
 
 (define (process-member-prototype-def name cname rtype mname args cb)
   (let* ([stub (gensym)]
@@ -610,21 +693,42 @@
      `(define ,stub 
 	(,(if cb 'foreign-callback-lambda* 'foreign-lambda*)
 	 ,rtype (((pointer ,name) ,this) ,@fvars)
-	 ,(sprintf (if (eq? 'void rtype) "~A->~A(~A);" "return(~A->~A(~A));")
+	 ,(sprintf (let ([code (if (eq? 'void rtype) 
+				   "~A->~A(~A);"
+				   "return(~A->~A(~A));") ] )
+		     (if exception-handler
+			 (sprintf "try { ~A } ~A;" code exception-handler)
+			 code) )
 		   this mname
 		   (string-intersperse (map ->string vars) ",")) ) )
-     `(define-method (,mname (this ,cname) . args)
-	(apply ,stub (slot-ref this 'this) args) ) ) ) )
+     (if (and full-specialization (pair? args))
+	 (let ([slist (gen-spec-list args)])
+	   `(define-method (,(fix-name mname) (this ,cname) ,@slist)
+	      (,stub (slot-ref this 'this) ,@(unzip1 slist) ) ) )
+	 `(define-method (,(fix-name mname) (this ,cname) . args)
+	    (apply ,stub (slot-ref this 'this) args) ) ) ) ) )
 
 (define parse-declaration
   (match-lambda*
-    [("export_constants" (or ('id "yes") ('num 1)))
+    [("export_constants" ('id "yes"))
      (set! export-constants #t) ]
     [("export_constants" _)
      (set! export-constants #f) ]
+    [("abstract" ('id cls))
+     (set! abstract-classes (cons (string->symbol cls) abstract-classes)) ]
+    [("class_finalizers" ('id "yes"))
+     (set! use-finalizers #t) ]
+    [("class_finalizers" _)
+     (set! use-finalizers #f) ]
+    [("destructor_name")
+     (set! destructor-name 'destroy) ]
+    [("destructor_name" ('string name))
+     (set! destructor-name (string->symbol name)) ]
+    [("exception_handler" ('string code))
+     (set! exception-handler code) ]
     [("prefix" ('string str))
      (set! prefix str) ]
-    [("prefix" (or ('id "no") ('num 0)))
+    [("prefix" ('id "no"))
      (set! prefix #f) ]
     [("scheme" ('string str))
      (let ([exp (with-input-from-string str read)])
@@ -635,44 +739,69 @@
      (match (string-split str ";")
        [(from to) 
 	(set! rename-list (alist-cons (string->symbol from) (string->symbol to) rename-list)) ]
-       [_ (quit "invalid rename declaration: ~S" decl)] ) ]
+       [_ (parsing-error "invalid rename declaration: ~S" str)] ) ]
     [("substitute" ('string str))
      (match (string-split str ";")
        [(from to) 
-	(set! name-substitution-rxs (cons from name-substitution-rxs))
-	(set! name-substitution-repls (cons to name-substitution-repls)) ]
-       [_ (quit "invalid name substitution string: ~S" str)] ) ]
+	(set! name-substitution-rxs (append name-substitution-rxs (list from)))
+	(set! name-substitution-repls (append name-substitution-repls (list to))) ]
+       [_ (parsing-error "invalid name substitution string: ~S" str)] ) ]
+    [("transform" ('string str))
+     (match (string-split str ";")
+       [(from to)
+	(let ([tr (handle-exceptions ex (parsing-error "error in transformation expression: ~S" to)
+		    (eval (safe-read-from-string to)) ) ] )
+	  (unless (procedure? tr)
+	    (parsing-error "transformation expression does not evaluate to procedure: ~S" to) )
+	  (set! name-substitution-rxs (append name-substitution-rxs (list from)))
+	  (set! name-substitution-repls (append name-substitution-repls (list tr))) ) ]
+       [_ (parsing-error "invalid transformation: ~S" str)] ) ]
+    [("full_specialization" ('id "yes"))
+     (set! full-specialization #t) ]
+    [("full_specialization" _)
+     (set! full-specialization #f) ]
     [(decl _)
-     (quit "invalid pseudo declaration: ~S" decl) ] ) )
+     (parsing-error "invalid pseudo declaration: ~S" decl) ] ) )
+
+(define (safe-read-from-string str)
+  (handle-exceptions ex (parsing-error "can not parse expression: ~S" str)
+    (with-input-from-string str read) ) )
 
 (define (parse-type-declaration vals)
-  (define (reads str)
-    (handle-exceptions ex (quit "can not parse expression: ~S" str)
-      (with-input-from-string str read) ) )
   (let rec ([vals vals])
     (match vals
       [(tname stype arg ret)
-       (let ([stype (reads stype)]
-	     [arg (and arg (reads arg))]
-	     [ret (and ret (reads ret))] 
+       (let ([stype (safe-read-from-string stype)]
+	     [arg (and arg (safe-read-from-string arg))]
+	     [ret (and ret (safe-read-from-string ret))] 
 	     [stname (string->symbol tname)] )
-	 (set! foreign-declarations (cons (sprintf "#define ~A ~A~%" tname (foreign-type-declaration stype "")) foreign-declarations))
+	 (set! foreign-declarations 
+	   (cons (sprintf "#define ~A ~A~%" tname (foreign-type-declaration stype "")) foreign-declarations))
 	 (emit 
 	  `(define-foreign-type ,stname ,stype ,@(if arg (list arg) '()) ,@(if ret (list ret) '())) )
 	 (set! declared-types (cons stname declared-types)) ) ]
       [(tname stype arg) (rec (list tname stype arg #f))]
       [(tname stype) (rec (list tname stype #f #f))]
-      [_ (quit "invalid value-syntax in type declaration: ~S" vals)] ) ) )
+      [_ (parsing-error "invalid value-syntax in type declaration: ~S" vals)] ) ) )
 
 (define (fix-name str)
   (let ([a (assq (->symbol str) rename-list)])
     (if a 
 	(cdr a)
-	(let ([n1 (fold string-substitute (->string str) name-substitution-rxs name-substitution-repls)])
+	(let ([n1 (fold 
+		   (lambda (rx repl str)
+		     (if (procedure? repl)
+			 (let ([m (string-match rx str)])
+			   (if m (repl m) str) )
+			 (string-substitute rx repl str #t) ) )
+		   (->string str)
+		   name-substitution-rxs
+		   name-substitution-repls) ] )
 	  (string->symbol
-	   (if prefix
-	       (string-append prefix n1)
-	       n1) ) ) ) ) )
+	   (strdowncase
+	    (if prefix
+		(string-append prefix n1)
+		n1) ) ) ) ) ) )
 
 (define (fix-cname str)
   (let ([a (assq (->symbol str) rename-list)])
@@ -693,6 +822,10 @@
   (let ([chunks (chunkify)])
     ;;(pretty-print chunks)		
     (for-each parse chunks)
+    (when (and (pair? defined-classes) 
+	       (not (member "tinyclos" used-units))
+	       (not explicit-use-flag) )
+      (set! used-units (append used-units '("tinyclos"))) )
     (reverse processed-output) ) )
 
 (define (parse-easy-ffi-rec port)
@@ -707,4 +840,110 @@
   (set! macro-table (cons (list (string->symbol name) '* '()) macro-table)) )
 
 (define (resolve-ffi-include-file fname)
-  (find file-exists? (map (cut make-pathname <> fname) ffi-include-path)) )
+  (find file-exists? (map (cut make-pathname <> fname) ffi-include-path-list)) )
+
+(define (foreign-type->class ftype)
+  (let rec ([ftype ftype])
+    (match ftype
+      ['char '<char>]
+      ['bool '<boolean>]
+      ['c-string '<string>]
+      [(or 'unsigned-char 'int 'unsigned-int 'short 'unsigned-short) '<exact>]
+      [(or 'long 'unsigned-long 'integer 'unsigned-integer) '<integer>]
+      [(or 'float 'double) '<inexact>]
+      [('enum _) '<exact>]
+      [('const t) (rec t)]
+      [('function . _) '<pointer>]
+      [('instance _ c) c]
+      [((or 'pointer 'c-pointer 'ref) _) '<pointer>]
+      ['u8vector '<u8vector>]
+      ['s8vector '<s8vector>]
+      ['u16vector '<u16vector>]
+      ['s16vector '<s16vector>]
+      ['u32vector '<u32vector>]
+      ['s32vector '<s32vector>]
+      ['f32vector '<f32vector>]
+      ['f64vector '<f64vector>]
+      [_ '<top>] ) ) )
+
+(define (gen-spec-list args)
+  (map (lambda (t) (list (gensym) (foreign-type->class t))) args) )
+
+(define strdowncase
+  (let ([cs case-sensitive])
+    (lambda (str)
+      (if (cs)
+	  str
+	  (let ([s2 (string-copy str)]
+		[len (string-length str)] )
+	    (do ([i (sub1 len) (sub1 i)])
+		((negative? i) s2)
+	      (string-set! s2 i (char-downcase (string-ref str i))) ) ) ) ) ) )
+
+
+;;; C syntax checker:
+
+(define syntax-check-location #f)
+
+(define (check-syntax-error text)
+  (lambda (fstr . args)
+    (quit #<<EOF
+suspicious code fragment~A:
+------------------------------------------------------------
+~A
+------------------------------------------------------------
+~?
+EOF
+	  (if syntax-check-location
+	      (sprintf " in `~A' form" syntax-check-location)
+	      "")
+	  text
+	  fstr
+	  args) ) )
+
+(define (check-c-syntax text . loc)
+  (unless no-c-syntax-checks
+    (fluid-let ([parsing-error (check-syntax-error text)]
+		[syntax-check-location (:optional loc #f)] )
+      (define (checkp p s)
+	(cond [(null? s) (parsing-error "unbalanced parantheses - missing match to `~A'" p)]
+	      [(not (eq? p (car s)))
+	       (parsing-error "unbalanced parantheses - expected `~A', but found `~A'" p (car s)) ] ) )
+      (define (checkpp p s)
+	(cond [(null? s) (parsing-error "unbalanced parantheses - missing match to `~A'" p)]
+	      [(not (equal? p (car s)))
+	       (parsing-error "unbalanced preprocessor conditional - expected `~A', but found `~A'" p (car s)) ] ) )
+      (lexer-init 'string text)
+      (set! pp-process #t)
+      (let loop ([pstack '()] [ppstack '()])
+	(let ([t (lexer)])
+	  (case t
+	    [(stop)
+	     (when (pair? pstack)
+	       (parsing-error "unbalanced parentheses - missing `~A'" (car pstack)) )
+	     (when (pair? ppstack)
+	       (parsing-error "unbalanced preprocessor command - missing `~A'" (car ppstack)) ) ]
+	    [(pp-else)
+	     (checkpp "#endif" ppstack) 
+	     (loop pstack ppstack) ]
+	    [(pp-endif)
+	     (checkpp "#endif" ppstack) 
+	     (loop pstack (cdr ppstack)) ]
+	    [(pp-if pp-ifdef pp-ifndef)
+	     (loop pstack (cons "#endif" ppstack)) ]
+	    [(open-curly)
+	     (loop (cons #\} pstack) ppstack) ]
+	    [(close-curly)
+	     (checkp #\} pstack) 
+	     (loop (cdr pstack) ppstack) ]
+	    [(open-paren)
+	     (loop (cons #\) pstack) ppstack) ]
+	    [(close-paren)
+	     (checkp #\) pstack)
+	     (loop (cdr pstack) ppstack) ] 
+	    [(open-bracket)
+	     (loop (cons #\] pstack) ppstack) ]
+	    [(close-bracket)
+	     (checkp #\] pstack)
+	     (loop (cdr pstack) ppstack) ] 
+	    [else (loop pstack ppstack)] ) ) ) ) ) )
\ No newline at end of file
