@@ -3,11 +3,10 @@
 
 (declare
   (unit prolog)
-  (uses srfi-1 match extras)
+  (uses srfi-1 extras)
   (usual-integrations)
-  (interrupts-disabled)
-  (fixnum)
-  (block) )
+  (disable-interrupts)
+  (fixnum) )
 
 
 ;----------------------------------------------------------------------------------------------------
