@@ -1,6 +1,6 @@
 ;;; extras.scm - Optional non-standard extensions
 ;
-; Copyright (c) 2000-2003, Felix L. Winkelmann
+; Copyright (c) 2000-2004, Felix L. Winkelmann
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -28,18 +28,41 @@
 ; felix@call-with-current-continuation.org
 ;
 ; Felix L. Winkelmann
-; Steinweg 1A
-; 37130 Gleichen, OT Weissenborn
+; Unter den Gleichen 1
+; 37130 Gleichen
 ; Germany
 
 
 (declare
  (unit extras)
  (usual-integrations)
- (foreign-declare "
+ (foreign-declare #<<EOF
 #define C_hashptr(x)   C_fix(x & C_MOST_POSITIVE_FIXNUM)
 #define C_mem_compare(to, from, n)   C_fix(C_memcmp(C_c_string(to), C_c_string(from), C_unfix(n)))
-") )
+
+#ifdef _WIN32
+# define FGETS_INTO_BUFFER  0
+static C_word fgets_into_buffer(C_word str, C_word port, C_word size) { return 0; }
+#else
+# define FGETS_INTO_BUFFER  1
+
+static C_word fgets_into_buffer(C_word str, C_word port, C_word size)
+{
+  int len, n = C_unfix(size);
+  char *buf = C_c_string(str);
+  FILE *fp = C_port_file(port);
+
+  if(fgets(buf, n, fp) == NULL) return C_fix(0);
+
+  len = C_strlen(buf);
+
+  if(len >= n - 1 && buf[ len - 1 ] != '\n') return C_SCHEME_FALSE;
+
+  return C_fix(len);
+}
+#endif
+EOF
+) )
 
 (cond-expand
  [paranoia]
@@ -86,12 +109,17 @@
 
 (define read-file
   (let ([read read]
-	[reverse reverse] )
+	[reverse reverse] 
+	[call-with-input-file call-with-input-file] )
     (lambda port
       (let ([port (:optional port ##sys#standard-input)])
-	(do ([x (read port) (read port)]
-	     [xs '() (cons x xs)] )
-	    ((eof-object? x) (reverse xs)) ) ) ) ) )
+	(define (slurp port)
+	  (do ([x (read port) (read port)]
+	       [xs '() (cons x xs)] )
+	      ((eof-object? x) (reverse xs)) ) )
+	(if (port? port)
+	    (slurp port)
+	    (call-with-input-file port slurp) ) ) ) ) )
 
 
 ;;; Combinators:
@@ -111,9 +139,9 @@
 (define (disjoin . preds)
   (lambda (x)
     (let loop ([preds preds])
-      (or (null? preds)
-	  ((##sys#slot preds 0) x)
-	  (loop (##sys#slot preds 1)) ) ) ) )
+      (and (not (null? preds))
+	   (or ((##sys#slot preds 0) x)
+	       (loop (##sys#slot preds 1)) ) ) ) ) )
 
 (define (constantly . xs)
   (if (eq? 1 (length xs))
@@ -231,6 +259,50 @@
 	      [(##sys#slot blst 0) (cons (##sys#slot lst 0) (loop (##sys#slot blst 1) (##sys#slot lst 1)))]
 	      [else (loop (##sys#slot blst 1) (##sys#slot lst 1))] ) ) ) ) )
 
+(define shuffle
+  ;; this should really shadow SORT! and RANDOM...
+  (lambda (l)
+    (map cdr
+	 (sort! (map (lambda (x) (cons (random 10000) x)) l)
+		(lambda (x y) (< (car x) (car y)))) ) ) )
+
+(define (alist-update! x y lst . cmp)
+  (let* ([cmp (if (pair? cmp) (car cmp) eqv?)]
+	 [aq (cond [(eq? eq? cmp) assq]
+		   [(eq? eqv? cmp) assv]
+		   [(eq? equal? cmp) assoc]
+		   [else 
+		    (lambda (x lst)
+		      (let loop ([lst lst])
+			(and (pair? lst)
+			     (let ([a (##sys#slot lst 0)])
+			       (if (and (pair? a) (cmp (##sys#slot a 0) x))
+				   a
+				   (loop (##sys#slot lst 1)) ) ) ) ) ) ] ) ] 
+	 [item (aq x lst)] )
+    (if item
+	(begin
+	  (##sys#setslot item 1 y)
+	  lst)
+	(cons (cons x y) lst) ) ) )
+
+(define (alist-ref x lst #!optional (cmp eqv?) (default #f))
+  (let* ([aq (cond [(eq? eq? cmp) assq]
+		   [(eq? eqv? cmp) assv]
+		   [(eq? equal? cmp) assoc]
+		   [else 
+		    (lambda (x lst)
+		      (let loop ([lst lst])
+			(and (pair? lst)
+			     (let ([a (##sys#slot lst 0)])
+			       (if (and (pair? a) (cmp (##sys#slot a 0) x))
+				   a
+				   (loop (##sys#slot lst 1)) ) ) ) ) ) ] ) ] 
+	 [item (aq x lst)] )
+    (if item
+	(##sys#slot item 0)
+	default) ) )
+
 
 ;;; Random numbers:
 
@@ -252,41 +324,58 @@
 
 ;;; Line I/O:
 
+
+(define-foreign-variable FGETS_INTO_BUFFER bool)
+
 (define read-line
-  (let ([read-char read-char]
-	[string-append string-append]
-	[make-string make-string] )
+  (let ([make-string make-string])
+    (define (fixup str len)
+      (##sys#substring
+       str 0
+       (if (and (fx>= len 1) (char=? #\return (##core#inline "C_subchar" str (fx- len 1))))
+	   (fx- len 1)
+	   len) ) )
     (lambda args
       (let* ([parg (pair? args)]
 	     [p (if parg (car args) ##sys#standard-input)]
 	     [limit (and parg (pair? (cdr args)) (cadr args))]
 	     [buffer (make-string 256)]
 	     [len 256] )
-	(##sys#call-with-current-continuation
-	 (lambda (return)
-	   (let loop ([i 0])
-	     (if (and limit (fx>= i limit))
-		 (return (##sys#substring buffer 0 i))
-		 (let ([c (read-char p)])
-		   (if (eof-object? c)
-		       (if (fx= i 0)
-			   c
-			   (##sys#substring buffer 0 i) ) 
-		       (case c
-			 [(#\newline) (return (##sys#substring buffer 0 i))]
-			 [(#\return)
-			  (let ([c (read-char p)])
-			    (if (char=? c #\newline)
-				(return (##sys#substring buffer 0 i))
-				(begin
-				  (##core#inline "C_setsubchar" buffer i c)
-				  (loop (fx+ i 1)) ) ) ) ]
-			 [else
-			  (when (fx>= i len)
-			    (set! buffer (string-append buffer (make-string len)))
-			    (set! len (fx+ len len)) )
-			  (##core#inline "C_setsubchar" buffer i c)
-			  (loop (fx+ i 1)) ] ) ) ) ) ) ) ) ) ) ) )
+	(if (and FGETS_INTO_BUFFER (eq? 'stream (##sys#slot p 7)))
+	    (let loop ([len 256] [buffer buffer] [result ""] [f #f])
+	      (let ([n (##core#inline "fgets_into_buffer" buffer p len)])
+		(cond [(eq? 0 n) (if f result (##sys#fudge 1))]
+		      [(not n)
+		       (loop (fx* len 2) (make-string (fx* len 2))
+			     (##sys#string-append result (##sys#substring buffer 0 (fx- len 1)))
+			     #t) ]
+		      [f (##sys#string-append result (fixup buffer (fx- n 1)))]
+		      [else (fixup buffer (fx- n 1))] ) ) )
+	    (##sys#call-with-current-continuation
+	     (lambda (return)
+	       (let loop ([i 0])
+		 (if (and limit (fx>= i limit))
+		     (return (##sys#substring buffer 0 i))
+		     (let ([c (##sys#read-char-0 p)])
+		       (if (eof-object? c)
+			   (if (fx= i 0)
+			       c
+			       (##sys#substring buffer 0 i) ) 
+			   (case c
+			     [(#\newline) (return (##sys#substring buffer 0 i))]
+			     [(#\return)
+			      (let ([c (##sys#read-char-0 p)])
+				(if (char=? c #\newline)
+				    (return (##sys#substring buffer 0 i))
+				    (begin
+				      (##core#inline "C_setsubchar" buffer i c)
+				      (loop (fx+ i 1)) ) ) ) ]
+			     [else
+			      (when (fx>= i len)
+				(set! buffer (##sys#string-append buffer (make-string len)))
+				(set! len (fx+ len len)) )
+			      (##core#inline "C_setsubchar" buffer i c)
+			      (loop (fx+ i 1)) ] ) ) ) ) ) ) ) ) ) ) ) ) 
 
 (define read-lines
   (let ([read-line read-line]
@@ -301,9 +390,7 @@
 	    ((or (eof-object? ln) (eq? n 0)) (reverse lns)) ) ) ) ) )
 
 (define read-string
-  (let ([read-char read-char]
-	[write-char write-char]
-	[open-output-string open-output-string]
+  (let ([open-output-string open-output-string]
 	[get-output-string get-output-string] )
     (lambda n-and-port
       (let-optionals n-and-port ([n #f] [p ##sys#standard-input])
@@ -311,18 +398,34 @@
 	  (when n (##sys#check-exact n 'read-string))
 	  (let loop ([n n])
 	    (or (and (eq? n 0) (get-output-string str))
-		(let ([c (read-char p)])
+		(let ([c (##sys#read-char-0 p)])
 		  (if (eof-object? c)
 		      (get-output-string str)
 		      (begin
-			(write-char c str) 
+			(##sys#write-char c str) 
 			(loop (and n (fx- n 1))) ) ) ) ) ) ) ) ) ) )
 
+(define read-token
+  (let ([open-output-string open-output-string]
+	[get-output-string get-output-string] )
+    (lambda (pred . port)
+      (let ([port (:optional port ##sys#standard-input)])
+	(##sys#check-port port 'read-token)
+	(let ([out (open-output-string)])
+	  (let loop ()
+	    (let ([c (##sys#peek-char-0 port)])
+	      (if (and (not (eof-object? c)) (pred c))
+		  (begin
+		    (##sys#write-char-0 (##sys#read-char-0 port) out)
+		    (loop) )
+		  (get-output-string out) ) ) ) ) ) ) ) )
+
 (define write-string 
   (let ([display display])
     (lambda (s . more)
       (##sys#check-string s 'write-string)
       (let-optionals more ([n #f] [port ##sys#standard-output])
+	(when n (##sys#check-exact n 'write-string))
 	(display 
 	 (if (and n (fx< n (##sys#size s)))
 	     (##sys#substring s 0 n)
@@ -542,7 +645,9 @@
 					   (out "#\\" col))))
 	      ((##core#inline "C_undefinedp" obj) (out "#<unspecified>" col))
 	      ((##sys#generic-structure? obj)
-	       (out (string-append "#<" (##sys#slot (##sys#slot obj 0) 1) ">") col) )
+	       (let ([o (open-output-string)])
+		 (##sys#user-print-hook obj #t o)
+		 (out (get-output-string o) col) ) )
 	      ((port? obj) (out (string-append "#<port " (##sys#slot obj 3) ">") col))
 	      ((eof-object? obj)  (out "#<eof>" col))
 	      (else               (out "#<unprintable object>" col)) ) )
@@ -742,6 +847,8 @@
     (generic-write obj #f (pretty-print-width) (lambda (s) (display s port) #t))
     (##core#undefined) ) )
 
+(define pp pretty-print)
+
 
 ;;; Anything->string conversion:
 
@@ -758,6 +865,11 @@
 	       (display x o)
 	       (get-output-string o) ) ] ) ) ) )
 
+(define conc
+  (let ([string-append string-append])
+    (lambda args
+      (apply string-append (map ->string args)) ) ) )
+
 
 ;;; Search one string inside another:
 
@@ -992,11 +1104,22 @@
     (collect 0 0 0 '()) ) )
 
 
+;;; Chop string into substrings:
+
+(define (string-chop str len)
+  (##sys#check-string str 'string-chop)
+  (##sys#check-exact len 'string-chop)
+  (let ([total (##sys#size str)])
+    (let loop ([total total] [pos 0])
+      (cond [(fx<= total 0) '()]
+	    [(fx<= total len) (list (##sys#substring str pos (fx+ pos total)))]
+	    [else (cons (##sys#substring str pos (fx+ pos len)) (loop (fx- total len) (fx+ pos len)))] ) ) ) )
+	   
+
 ;;; Write simple formatted output:
 
 (define fprintf
   (let ([write write]
-	[write-char write-char]
 	[newline newline]
 	[display display] )
     (lambda (port msg . args)
@@ -1024,7 +1147,7 @@
 		  (case (char-upcase dchar)
 		    ((#\S) (write (next) port))
 		    ((#\A) (display (next) port))
-		    ((#\C) (write-char (next) port))
+		    ((#\C) (##sys#write-char-0 (next) port))
 		    ((#\B) (display (number->string (next) 2) port))
 		    ((#\O) (display (number->string (next) 8) port))
 		    ((#\X) (display (number->string (next) 16) port))
@@ -1034,7 +1157,7 @@
 			    [lst (next)] )
 		       (##sys#check-list lst 'fprintf)
 		       (rec fstr lst) ) )
-		    ((#\~) (write-char #\~ port))
+		    ((#\~) (##sys#write-char-0 #\~ port))
 		    ((#\%) (newline port))
 		    (else
 		     (if (char-whitespace? dchar)
@@ -1043,7 +1166,7 @@
 			       (skip (fetch))
 			       (set! index (##core#inline "C_fixnum_difference" index 1)) ) )
 			 (##sys#error 'fprintf "illegal format-string character" dchar) ) ) ) )
-		(write-char c port) ) ) ) ) ) ) )
+		(##sys#write-char-0 c port) ) ) ) ) ) ) )
 
 
 (define printf
@@ -1309,6 +1432,10 @@
   (##sys#check-structure ht 'hash-table 'hash-table-count)
   (##sys#slot ht 2) )
 
+(define (hash-table-size ht)
+  (##sys#check-structure ht 'hash-table 'hash-table-count)
+  (##sys#size (##sys#slot ht 1)) )
+
 (define hash-table-ref
   (let ([hash hash]
 	[eq0 eq?] )
