@@ -170,7 +170,7 @@
 (define (##format#out-char ch)
   (if ##format#case-conversion
       (display (##format#case-conversion (string ch)) ##format#port)
-      (write-char ch ##format#port))
+      (##sys#write-char-0 ch ##format#port))
   (set! ##format#output-col
 	(if (char=? ch #\newline)
 	    0
@@ -183,7 +183,7 @@
 (define (##format#out-substr str i n)
   (do ((k i (+ k 1)))
       ((= k n))
-    (write-char (string-ref str k) ##format#port))
+    (##sys#write-char-0 (string-ref str k) ##format#port))
   (set! ##format#output-col (+ ##format#output-col n)))
 
 ;(define (##format#out-fill n ch)       ; this allocates a new string
@@ -192,7 +192,7 @@
 (define (##format#out-fill n ch)
   (do ((i 0 (+ i 1)))
       ((= i n))
-    (write-char ch ##format#port))
+    (##sys#write-char-0 ch ##format#port))
   (set! ##format#output-col (+ ##format#output-col n)))
 
 ; format's user error handler
@@ -871,69 +871,81 @@
 ;; It uses ##format#iobj->str to print out internal objects as
 ;; quoted strings so that the output can always be processed by (read)
 
-(define (##format#obj->str obj slashify)
-  (cond
-   ((string? obj)
-    (if slashify
-	(let ((obj-len (string-length obj)))
-	  (string-append
-	   "\""
-	   (let loop ((i 0) (j 0))	; taken from Marc Feeley's pp.scm
-	     (if (= j obj-len)
-		 (string-append (substring obj i j) "\"")
-		 (let ((c (string-ref obj j)))
-		   (if (or (char=? c #\\)
-			   (char=? c #\"))
-		       (string-append (substring obj i j) "\\"
-				      (loop j (+ j 1)))
-		       (loop i (+ j 1))))))))
-	obj))
+(define ##format#obj->str
+  (let ([ksp keyword-style])
+    (lambda (obj slashify)
 
-   ((boolean? obj) (if obj "#t" "#f"))
+      (define (symname x)
+	(cond [(fx= 0 (##sys#byte (##sys#slot x 1) 0))
+	       (let ([str (##sys#symbol->string x)])
+		 (case (ksp)
+		   [(#:prefix) (##sys#string-append ":" str)]
+		   [(#:suffix) (##sys#string-append str ":")]
+		   [else (##sys#string-append "#:" str)] ) ) ]
+	      [else (##sys#symbol->qualified-string x)] ) )
 
-   ((number? obj) (number->string obj))
+      (cond
+       ((string? obj)
+	(if slashify
+	    (let ((obj-len (string-length obj)))
+	      (##sys#string-append
+	       "\""
+	       (let loop ((i 0) (j 0)) ; taken from Marc Feeley's pp.scm
+		 (if (= j obj-len)
+		     (##sys#string-append (substring obj i j) "\"")
+		     (let ((c (string-ref obj j)))
+		       (if (or (char=? c #\\)
+			       (char=? c #\"))
+			   (string-append (substring obj i j) "\\"
+					  (loop j (+ j 1)))
+			   (loop i (+ j 1))))))))
+	    obj))
 
-   ((symbol? obj)
-    (if ##format#symbol-case-conv
-	(##format#symbol-case-conv (symbol->string obj))
-	(symbol->string obj)))
+       ((boolean? obj) (if obj "#t" "#f"))
 
-   ((char? obj)
-    (if slashify
-	(##format#char->str obj)
-	(string obj)))
+       ((number? obj) (number->string obj))
 
-   ((null? obj) "()")
+       ((symbol? obj)
+	(if ##format#symbol-case-conv
+	    (##format#symbol-case-conv (symname obj))
+	    (symname obj)))
 
-   ((input-port? obj)
-    (##format#iobj->str obj))
+       ((char? obj)
+	(if slashify
+	    (##format#char->str obj)
+	    (string obj)))
 
-   ((output-port? obj)
-    (##format#iobj->str obj))
+       ((null? obj) "()")
 
-   ((list? obj)
-    (string-append "("
-		   (let loop ((obj-list obj))
-		     (if (null? (cdr obj-list))
-			 (##format#obj->str (car obj-list) #t)
-			 (string-append
-			  (##format#obj->str (car obj-list) #t)
-			  " "
-			  (loop (cdr obj-list)))))
-		   ")"))
+       ((input-port? obj)
+	(##format#iobj->str obj))
 
-   ((pair? obj)
-    (string-append "("
-		   (##format#obj->str (car obj) #t)
-		   " . "
-		   (##format#obj->str (cdr obj) #t)
-		   ")"))
+       ((output-port? obj)
+	(##format#iobj->str obj))
 
-   ((vector? obj)
-    (string-append "#" (##format#obj->str (vector->list obj) #t)))
+       ((list? obj)
+	(string-append "("
+		       (let loop ((obj-list obj))
+			 (if (null? (cdr obj-list))
+			     (##format#obj->str (car obj-list) #t)
+			     (string-append
+			      (##format#obj->str (car obj-list) #t)
+			      " "
+			      (loop (cdr obj-list)))))
+		       ")"))
 
-   (else				; only objects with an #<...>
-    (##format#iobj->str obj))))		; representation should fall in here
+       ((pair? obj)
+	(string-append "("
+		       (##format#obj->str (car obj) #t)
+		       " . "
+		       (##format#obj->str (cdr obj) #t)
+		       ")"))
+
+       ((vector? obj)
+	(string-append "#" (##format#obj->str (vector->list obj) #t)))
+
+       (else				; only objects with an #<...>
+	(##format#iobj->str obj)))))) ; representation should fall in here
 
 ;; ##format#iobj->str reveals the implementation dependent representation of
 ;; #<...> objects with the use of display and ##format#call-with-output-string.
@@ -1099,7 +1111,7 @@
 	(let loop ((n n)
 		   (romans ##format#roman-alist)
 		   (s '()))
-	  (if (null? romans) (list->string (reverse s))
+	  (if (null? romans) (##sys#reverse-list->string s)
 	      (let ((roman-val (caar romans))
 		    (roman-dgt (cadar romans)))
 		(do ((q (quotient n roman-val) (- q 1))
@@ -1117,7 +1129,7 @@
 		   (boundaries ##format#roman-boundary-values)
 		   (s '()))
 	  (if (null? romans)
-	      (list->string (reverse s))
+	      (##sys#reverse-list->string s)
 	      (let ((roman-val (caar romans))
 		    (roman-dgt (cadar romans))
 		    (bdry (car boundaries)))
@@ -1193,7 +1205,7 @@
 	   (##format#error
 	     "only integers can be converted to English cardinals"))
 	  ((= n 0) "zero")
-	  ((< n 0) (string-append "minus " (##format#num->cardinal (- n))))
+	  ((< n 0) (##sys#string-append "minus " (##format#num->cardinal (- n))))
 	  (else
 	    (let ((power3-word-limit
 		    (length ##format#cardinal-thousand-block-list)))
@@ -1241,13 +1253,13 @@
 	   (##format#error
 	     "only integers can be converted to English ordinals"))
 	  ((= n 0) "zeroth")
-	  ((< n 0) (string-append "minus " (##format#num->ordinal (- n))))
+	  ((< n 0) (##sys#string-append "minus " (##format#num->ordinal (- n))))
 	  (else
 	    (let ((hundreds (quotient n 100))
 		  (tens+ones (remainder n 100)))
-	      (string-append
+	      (##sys#string-append
 		(if (> hundreds 0)
-		    (string-append
+		    (##sys#string-append
 		      (##format#num->cardinal (* hundreds 100))
 		      (if (= tens+ones 0) "th" " "))
 		    "")
