@@ -1,6 +1,6 @@
 ;;;; lolevel.scm - Low-level routines for CHICKEN
 ;
-; Copyright (c) 2000-2003, Felix L. Winkelmann
+; Copyright (c) 2000-2004, Felix L. Winkelmann
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -28,8 +28,8 @@
 ; felix@call-with-current-continuation.org
 ;
 ; Felix L. Winkelmann
-; Steinweg 1A
-; 37130 Gleichen, OT Weissenborn
+; Unter den Gleichen 1
+; 37130 Gleichen
 ; Germany
 
 
@@ -135,7 +135,9 @@ EOF
 (define null-pointer ##sys#null-pointer)
 
 (define (pointer? x)
-  (and (##core#inline "C_blockp" x) (##core#inline "C_pointerp" x)) )
+  (and (##core#inline "C_blockp" x)
+       (or (##core#inline "C_pointerp" x)
+	   (##core#inline "C_taggedpointerp" x) ) ) )
 
 (define address->pointer
     (lambda (addr)
@@ -199,8 +201,8 @@ EOF
 (define pointer-s16-set! (foreign-lambda* void ([c-pointer p] [int n]) "*((short *)p) = n;"))
 (define pointer-u32-set! (foreign-lambda* void ([c-pointer p] [int n]) "*((C_u32 *)p) = n;"))
 (define pointer-s32-set! (foreign-lambda* void ([c-pointer p] [int n]) "*((C_s32 *)p) = n;"))
-(define pointer-f32-set! (foreign-lambda* void ([c-pointer p] [int n]) "*((float *)p) = n;"))
-(define pointer-f64-set! (foreign-lambda* void ([c-pointer p] [int n]) "*((double *)p) = n;"))
+(define pointer-f32-set! (foreign-lambda* void ([c-pointer p] [double n]) "*((float *)p) = n;"))
+(define pointer-f64-set! (foreign-lambda* void ([c-pointer p] [float n]) "*((double *)p) = n;"))
 
 (define pointer-u8-ref (foreign-lambda* int ([c-pointer p]) "return(*((unsigned char *)p));"))
 (define pointer-s8-ref (foreign-lambda* int ([c-pointer p]) "return(*((char *)p));"))
@@ -211,6 +213,24 @@ EOF
 (define pointer-f32-ref (foreign-lambda* float ([c-pointer p]) "return(*((float *)p));"))
 (define pointer-f64-ref (foreign-lambda* double ([c-pointer p]) "return(*((double *)p));"))
 
+(define (tag-pointer ptr tag)
+  (let ([tp (##sys#make-tagged-pointer tag)])
+    (if (and (##core#inline "C_blockp" ptr) (##core#inline "C_specialp" ptr))
+	(##core#inline "C_copy_pointer" ptr tp)
+	(##sys#signal-hook #:type-error 'tag-pointer "bad argument type - not a pointer" ptr) )
+    tp) )
+
+(define (tagged-pointer? x tag)
+  (and (##core#inline "C_blockp" x) 
+       (##core#inline "C_taggedpointerp" x)
+       (eq? tag (##sys#slot x 1)) ) )
+
+(define (pointer-tag x)
+  (if (and (##core#inline "C_blockp" x) (##core#inline "C_specialp" x))
+      (and (##core#inline "C_taggedpointerp" x)
+	   (##sys#slot x 1) )
+      (##sys#signal-hook #:type-error 'tag-pointer "bad argument type - not a pointer" x) ) )
+
 
 ;;; Procedures extended with data:
 
@@ -383,7 +403,7 @@ EOF
   (define (make size init alloc loc)
     (##sys#check-exact size loc)
     (if (fx> size _c_header_size_mask)
-	(##sys#error loc "out of range" size _c_header_size_mask)
+	(##sys#signal-hook #:bounds-error loc "out of range" size _c_header_size_mask)
 	(let ([bv (alloc size)])
 	  (cond [bv
 		 (when (pair? init) (byte-vector-fill! bv (##sys#slot init 0)))
@@ -448,6 +468,15 @@ EOF
   (and (##core#inline "C_blockp" x)
        (##core#inline "C_structurep" x) ) )
 
+(define (record->vector x)
+  (if (and (not (##sys#immediate? x)) (##sys#generic-structure? x))
+      (let* ([n (##sys#size x)]
+	     [v (##sys#make-vector n)] )
+	(do ([i 0 (fx+ i 1)])
+	    ((fx>= i n) v)
+	  (##sys#setslot v i (##sys#slot x i)) ) )
+      (##sys#signal-hook #:type-error 'record->vector "bad argument type - not a record structure" x) ) )
+
 
 ;;; Convert executable bytevector into a procedure:
 
@@ -474,7 +503,7 @@ EOF
 
 ;;; Copy arbitrary object:
 
-(define copy
+(define object-copy
   (let ([make-vector make-vector])
     (lambda (x)
       (let copy ([x x])
@@ -493,9 +522,9 @@ EOF
 
 ;;; Evict objects into static memory:
 
-(define (evicted? x) (##core#inline "C_permanentp" x))
+(define (object-evicted? x) (##core#inline "C_permanentp" x))
 
-(define evict
+(define object-evict
   (let ([make-hash-table make-hash-table]
 	[hash-table-ref hash-table-ref]
 	[hash-table-set! hash-table-set!] )
@@ -521,7 +550,7 @@ EOF
 		       (##sys#setislot y i (evict (##sys#slot x i))) ) )
 		   y) ] ) ) ) ) ) )
 
-(define release
+(define object-release
   (lambda (x . releaser)
     (let ([free (if (pair? releaser) 
 		    (car releaser) 
@@ -537,7 +566,7 @@ EOF
 		     (release (##sys#slot x i))) )
 		 (free (##sys#address->pointer (##core#inline "C_block_address" x))) ) ] ) ) ) ) )
 
-(define evict-to-location
+(define object-evict-to-location
   (let ([make-hash-table make-hash-table]
 	[hash-table-ref hash-table-ref]
 	[align-to-word align-to-word]
@@ -546,12 +575,12 @@ EOF
       (cond-expand
        [(not unsafe)
 	(when (not (and (##core#inline "C_blockp" ptr) (##core#inline "C_specialp" ptr)))
-	  (##sys#signal-hook #:type-error 'evict-to-location "bad argument type - not a pointer" ptr) ) ]
+	  (##sys#signal-hook #:type-error 'object-evict-to-location "bad argument type - not a pointer" ptr) ) ]
        [else] )
       (let* ([limit
 	      (if (pair? limit)
 		  (let ([limit (car limit)])
-		    (##sys#check-exact limit 'evict-to-location)
+		    (##sys#check-exact limit 'object-evict-to-location)
 		    limit)
 		  #f) ]
 	     [ptr2 (##sys#address->pointer (##sys#pointer->address ptr))]
@@ -567,7 +596,7 @@ EOF
 				    (##core#inline "C_bytes" 1) ) ] )
 			 (when limit
 			   (set! limit (fx- limit bytes))
-			   (when (fx< limit 0) (##sys#error 'evict-to-location "can not evict object - limit exceeded" x)) )
+			   (when (fx< limit 0) (##sys#error 'object-evict-to-location "can not evict object - limit exceeded" x)) )
 			 (let ([y (##core#inline "C_evict_block" x ptr2)])
 			   (when (symbol? x) (##sys#setislot y 0 (##core#undefined)))
 			   (##sys#set-pointer-address! ptr2 (+ (##sys#pointer->address ptr2) bytes))
@@ -607,14 +636,14 @@ EOF
 		       (set! bytes (fx+ (evict (##sys#slot x i)) bytes)) ) )
 		   bytes) ] ) ) ) ) ) )
 
-(define unevict
+(define object-unevict
   (let ([make-vector make-vector]
 	[make-hash-table make-hash-table]
 	[hash-table-set! hash-table-set!]
 	[hash-table-ref hash-table-ref] )
     (lambda (x)
       (define (err x)
-	(##sys#signal-hook #:type-error 'unevict "can not copy object" x) )
+	(##sys#signal-hook #:type-error 'object-unevict "can not copy object" x) )
       (let ([tab (make-hash-table)])
 	(let copy ([x x])
 	  (cond [(not (##core#inline "C_blockp" x)) x]
@@ -635,96 +664,35 @@ EOF
 		   y) ] ) ) ) ) ) )
 
 
-;;; `become!':
+;;; `become':
 
-(define become! 
+(define object-become! 
   (cond-expand
    [unsafe ##sys#become!]
    [else
     (lambda (lst)
-      (##sys#check-list lst 'become!)
+      (##sys#check-list lst 'object-become!)
       (let loop ([lst lst])
 	(cond [(null? lst)]
 	      [(pair? lst)
 	       (let ([a (##sys#slot lst 0)])
-		 (##sys#check-pair a 'become!)
+		 (##sys#check-pair a 'object-become!)
 		 (unless (##core#inline "C_blockp" (##sys#slot a 0))
-		   (##sys#signal-hook #:type-error 'become! "bad argument type - old item is immediate" a) )
+		   (##sys#signal-hook #:type-error 'object-become! "bad argument type - old item is immediate" a) )
 		 (unless (##core#inline "C_blockp" (##sys#slot a 1))
-		   (##sys#signal-hook #:type-error 'become! "bad argument type - new item is immediate" a) )
+		   (##sys#signal-hook #:type-error 'object-become! "bad argument type - new item is immediate" a) )
 		 (loop (##sys#slot lst 1)) ) ]
-	      [else (##sys#signal-hook #:type-error 'become! "bad argument type - not an a-list")] ) )
+	      [else (##sys#signal-hook #:type-error 'object-become! "bad argument type - not an a-list")] ) )
       (##sys#become! lst) ) ] ) )
 
 
 ;;; locatives:
- 
+
 (define (make-locative obj . index)
-  (let ([index (:optional index #f)])
-    (cond [(##sys#immediate? obj)
-	   (##sys#signal-hook #:type-error 'make-locative "locative can not refer to immediate object" obj) ]
-	  [(or (vector? obj) (pair? obj))
-	   (when index (##sys#check-range index 0 (##sys#size obj) 'make-locative))
-	   (##core#inline_allocate ("C_a_i_make_locative" 4) 0 obj (or index 0)) ]
-	  [(byte-vector? obj)
-	   (when index (##sys#check-range index 0 (##sys#size obj) 'make-locative))
-	   (##core#inline_allocate ("C_a_i_make_locative" 4) 2 obj (or index 0)) ]
-	  [(##sys#generic-structure? obj)
-	   (unless index (##sys#error "index required for record-locative" obj))
-	   (case (##sys#slot obj 0)
-	     [(record)
-	      (let loop ([i 2] [slots (##sys#slot (##sys#slot obj 1) 2)])
-		(cond [(null? slots)
-		       (##sys#error 'make-locative "record-type does not contain slot" obj index) ]
-		      [(eq? index (##sys#slot slots 0))
-		       (##core#inline_allocate ("C_a_i_make_locative" 4) 0 obj i) ]
-		      [else (loop (fx+ i 1) (##sys#slot slots 1))] ) ) ]
-	     [(array) 
-	      (unless index (##sys#error "index required for array-locative" obj))
-	      (array:make-locative obj index) ]
-	     [(u8vector)
-	      (let ([v (##sys#slot obj 1)])
-		(when index (##sys#check-range index 0 (##sys#size v) 'make-locative))
-		(##core#inline_allocate ("C_a_i_make_locative" 4) 2 v (or index 0)) ) ]
-	     [(s8vector)
-	      (let ([v (##sys#slot obj 1)])
-		(when index (##sys#check-range index 0 (##sys#size v) 'make-locative))
-		(##core#inline_allocate ("C_a_i_make_locative" 4) 3 v (or index 0)) ) ]
-	     [(u16vector)
-	      (let ([v (##sys#slot obj 1)])
-		(when index (##sys#check-range index 0 (##sys#size v) 'make-locative))
-		(##core#inline_allocate ("C_a_i_make_locative" 4) 4 v (or index 0)) ) ]
-	     [(s16vector)
-	      (let ([v (##sys#slot obj 1)])
-		(when index (##sys#check-range index 0 (##sys#size v) 'make-locative))
-		(##core#inline_allocate ("C_a_i_make_locative" 4) 5 v (or index 0)) ) ]
-	     [(u32vector)
-	      (let ([v (##sys#slot obj 1)])
-		(when index (##sys#check-range index 0 (##sys#size v) 'make-locative))
-		(##core#inline_allocate ("C_a_i_make_locative" 4) 6 v (or index 0)) ) ]
-	     [(s32vector)
-	      (let ([v (##sys#slot obj 1)])
-		(when index (##sys#check-range index 0 (##sys#size v) 'make-locative))
-		(##core#inline_allocate ("C_a_i_make_locative" 4) 7 v (or index 0)) ) ]
-	     [(f32vector)
-	      (let ([v (##sys#slot obj 1)])
-		(when index (##sys#check-range index 0 (##sys#size v) 'make-locative))
-		(##core#inline_allocate ("C_a_i_make_locative" 4) 8 v (or index 0)) ) ]
-	     [(f64vector)
-	      (let ([v (##sys#slot obj 1)])
-		(when index (##sys#check-range index 0 (##sys#size v) 'make-locative))
-		(##core#inline_allocate ("C_a_i_make_locative" 4) 9 v (or index 0)) ) ]
-	     [else 
-	      (when index (##sys#check-range index 0 (fx- (##sys#size obj) 1) 'make-locative))
-	      (##core#inline_allocate ("C_a_i_make_locative" 4) 0 obj (fx+ (or index 0) 1)) ] ) ]
-	  [(string? obj)
-	   (when index (##sys#check-range index 0 (##sys#size obj) 'make-locative))
-	   (##core#inline_allocate ("C_a_i_make_locative" 4) 1 obj (or index 0)) ] 
-	  [else
-	   (##sys#signal-hook
-	    #:type-error 'make-locative 
-	    "bad argument type - locative can not refer to objects of this type" 
-	    obj) ] ) ) )
+  (##sys#make-locative obj (:optional index 0) #f 'make-locative) )
+
+(define (make-weak-locative obj . index)
+  (##sys#make-locative obj (:optional index 0) #t 'make-weak-locative) )
 
 (define locative-ref (##core#primitive "C_locative_ref"))
 (define (locative-set! x y) (##core#inline "C_i_locative_set" x y))
@@ -739,10 +707,10 @@ EOF
 (define (invalid-procedure-call-handler proc)
   (unless (procedure? proc)
     (##sys#signal-hook #:type-error 'invalid-procedure-call-handler "bad argument type - not a procedure" proc) )
-  (set! icp-hook proc)
+  (set! ipc-hook proc)
   (set! ##sys#invalid-procedure-call-hook 
     (lambda args
-      (icp-hook ##sys#last-invalid-procedure args) ) ) )
+      (ipc-hook ##sys#last-invalid-procedure args) ) ) )
 
 (define (unbound-variable-value . val)
   (set! ##sys#unbound-variable-value-hook 
@@ -763,3 +731,8 @@ EOF
 (define (global-bound? sym)
   (##sys#check-symbol sym 'global-bound?)
   (##sys#symbol-has-toplevel-binding? sym) )
+
+(define (global-make-unbound! sym)
+  (##sys#check-symbol sym 'global-make-unbound!)
+  (##sys#setslot sym 0 (##sys#slot '##sys#arbitrary-unbound-symbol 0))
+  sym)
