@@ -1,6 +1,6 @@
 ;;;; optimizer.scm - The CHICKEN Scheme compiler (optimizations)
 ;
-; Copyright (c) 2000-2003, Felix L. Winkelmann
+; Copyright (c) 2000-2004, Felix L. Winkelmann
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -28,8 +28,8 @@
 ; felix@call-with-current-continuation.org
 ;
 ; Felix L. Winkelmann
-; Steinweg 1A
-; 37130 Gleichen, OT Weissenborn
+; Unter den Gleichen 1
+; 37130 Gleichen
 ; Germany
 
 
@@ -67,7 +67,7 @@
   pprint-expressions-to-file foreign-type-check estimate-foreign-result-size scan-used-variables scan-free-variables
   topological-sort print-version print-usage initialize-analysis-database
   expand-foreign-callback-lambda default-optimization-passes default-optimization-passes-when-trying-harder
-  product copyright compiler-features units-used-by-default words-per-flonum
+  copyright compiler-features units-used-by-default words-per-flonum rewrite
   parameter-limit eq-inline-operator optimizable-rest-argument-operators
   membership-test-operators membership-unfold-limit valid-compiler-options valid-compiler-options-with-argument
   make-random-name final-foreign-type
@@ -346,10 +346,12 @@
 							     (append 
 							      args
 							      (list
-							       (make-node
-								'##core#inline_allocate 
-								(list "C_a_i_list" (* 3 (length rargs)))
-								rargs) ) ) ) ) ) ] )
+							       (if (null? rargs)
+								   (qnode '())
+								   (make-node
+								    '##core#inline_allocate 
+								    (list "C_a_i_list" (* 3 (length rargs)))
+								    rargs) ) ) ) ) ) ) ] )
 					 (set! rest-consers (cons n2 rest-consers))
 					 n2) ) ) ]
 				  [else (walk-generic n class params subs)] ) ) ) ]
@@ -974,7 +976,7 @@
 	  (or (second classargs) unsafe)
 	  (let ((pname (first classargs)))
 	    (make-node '##core#call (if (pair? params) (cons #t (cdr params)) params)
-		       (cons* (make-node '##core#proc (list pname) '())
+		       (cons* (make-node '##core#proc (list pname #t) '())
 			      cont callargs) ) ) ) )
 
     ;; (<op> <x> ...) -> (##core#inline <iop-safe>/<iop-unsafe> <x> ...)
@@ -1084,6 +1086,23 @@
 		      callargs) ) ) ]
 		  [else #f] ) ) ) )
 
+    ;; (<op> ...) -> (##core#inline <iop> <arg1> ... (quote <x>) <argN>)
+    ((20) ; classargs = (<argc> <iop> <x> <safe>)
+     (let ([n (length callargs)])
+       (and (or (fourth classargs) unsafe)
+	    inline-substitutions-enabled
+	    (= n (first classargs))
+	    (or (test name 'standard-binding) (test name 'extended-binding))
+	    (make-node
+	     '##core#call '(#t)
+	     (list cont
+		   (make-node 
+		    '##core#inline (list (second classargs))
+		    (let-values ([(head tail) (split-at callargs (sub1 n))])
+		      (append head
+			      (list (qnode (third classargs)))
+			      tail) ) ) ) ) ) ) )
+
     (else (bomb "bad type (optimize)")) ) )
 
 
@@ -1115,7 +1134,7 @@
 				 [refs (get db d 'references)]
 				 [sites (get db d 'call-sites)] )
 			(and (eq? n val)
-			     (= (length refs) (length sites)) 
+			     (= (length refs) (length sites))
 			     (scan (first subs) (first llist) d dn (cons d llist)) ) ) )
 		 (transform n d inner-ks hoistable dn allocated) 
 		 (walk #f (first subs) #f) ) ) ]
@@ -1210,10 +1229,11 @@
 	  (debugging 'o "direct leaf routine with hoistable closures/allocation" fnvar (delay (unzip1 hoistable)) allocated)
 	  (debugging 'o "direct leaf routine/allocation" fnvar allocated) )
       (set! dirty #t)
-      (let ([params (node-parameters n)]
-	    [klambdas '()] 
-	    [sites (get db fnvar 'call-sites)] 
-	    [ksites '()] )
+      (let* ([params (node-parameters n)]
+	     [argc (length (third params))]
+	     [klambdas '()] 
+	     [sites (get db fnvar 'call-sites)] 
+	     [ksites '()] )
 	(match params
 	  [(id _ (kvar vars ...) _)
 	   ;; Remove continuation argument:
@@ -1235,6 +1255,8 @@
 		      (cond [(eq? fnvar (first fnp))
 			     (set! ksites (alist-cons #f n ksites))
 			     (cond [(eq? kvar (first arg0p))
+				    (unless (= argc (length (cdr subs)))
+				      (warning "known procedure called recursively with wrong number of arguments: `~A'" fnvar) )
 				    (node-class-set! n '##core#recurse)
 				    (node-parameters-set! n (list #t id))
 				    (node-subexpressions-set! n (cddr subs)) ]
@@ -1242,6 +1264,8 @@
 				    => (lambda (a)
 					 (let* ([klam (cdr a)]
 						[kbody (first (node-subexpressions klam))] )
+					   (unless (= argc (length (cdr subs)))
+					     (warning "known procedure called recursively with wrong number of arguments: `~A'" fnvar) )
 					   (node-class-set! n 'let)
 					   (node-parameters-set! n (take (third (node-parameters klam)) 1))
 					   (node-subexpressions-set!
@@ -1270,6 +1294,8 @@
 	    (lambda (site)
 	      (let* ([n (cdr site)]
 		     [nsubs (node-subexpressions n)] )
+		(unless (= argc (length (cdr nsubs)))
+		  (warning "known procedure called with wrong number of arguments: `~A'" fnvar) )
 		(node-subexpressions-set!
 		 n
 		 (list (second nsubs)
@@ -1523,7 +1549,8 @@
 	    [walked '()] )
 	(define (walk gn)
 	  (let ([name (car gn)])
-	    (unless (memq name walked)
+	    ;; Hm. To handle liftables that are called recursively (but indirect) I use this kludge. Is it safe?
+	    (unless (> (count (cut eq? name <>) walked) 1)
 	      (set! walked (cons name walked))
 	      (let ([callees (cddr gn)])
 		(for-each (lambda (c) (walk (assq c graph))) callees)
@@ -1651,16 +1678,16 @@
       (let ([g (build-call-graph cs)])
 	(debugging 'p "eliminating non-liftables...")
 	(let ([g2 (eliminate cs g)])
-	  #;(when (debugging 'l "call-graph:") (pretty-print g2))
+	  (when (debugging 'l "call-graph:") (pretty-print g2))
 	  (debugging 'p "computing access-lists...")
 	  (let ([al (collect-accessibles g2)])
-	    #;(when (debugging 'l "accessibles:") (pretty-print al))
+	    (when (debugging 'l "accessibles:") (pretty-print al))
 	    (debugging 'p "eliminating liftables by access-lists and non-liftable callees...")
 	    (let ([ls (eliminate3 (eliminate4 g2))]) ;(eliminate2 g2 al)))])
 	      (debugging 'o "liftable local procedures" (delay (unzip1 ls)))
 	      (debugging 'p "gathering extra parameters...")
 	      (let ([extra (compute-extra-variables ls)])
-		#;(when (debugging 'l "additional parameters:") (pretty-print extra))
+		(when (debugging 'l "additional parameters:") (pretty-print extra))
 		(debugging 'p "changing call sites...")
 		(extend-call-sites! extra)
 		(debugging 'p "removing local bindings...")
