@@ -1,6 +1,6 @@
 ;;;; parameters.scm - Miscellaneous system parameters
 ;
-; Copyright (c) 2000-2003, Felix L. Winkelmann
+; Copyright (c) 2000-2004, Felix L. Winkelmann
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -28,8 +28,8 @@
 ; felix@call-with-current-continuation.org
 ;
 ; Felix L. Winkelmann
-; Steinweg 1A
-; 37130 Gleichen, OT Weissenborn
+; Unter den Gleichen 1
+; 37130 Gleichen
 ; Germany
 
 
@@ -68,6 +68,7 @@
 (define-constant inline-table-size 301)
 (define-constant constant-table-size 301)
 (define-constant real-name-table-size 997)
+(define-constant user-level-optimizer-table-size 301)
 
 (define-constant default-literal-compression-threshold 50)
 
@@ -77,13 +78,32 @@
 (define-constant cygwin-default-dynamic-load-libraries '("cygstuffed-chicken-0" "cygsrfi-chicken-0" "cygchicken-0"))
 (define-constant msvc-default-dynamic-load-libraries '("stuffed-chicken" "srfi-chicken" "chicken"))
 
+(define-constant macosx-load-library-extension ".dylib")
 (define-constant windows-load-library-extension ".dll")
-(define-constant unix-load-library-extension ".so")
+(define-constant default-load-library-extension ".so")
 
 (define-constant environment-table-size 301)
 
 (define-constant source-file-extension ".scm")
+(define-constant setup-file-extension ".setup")
 
-(define-constant registry-filename "REGISTRY")
-(define-constant registry-environment-variable "CHICKEN_REGISTRY")
-(define-constant registry-local-filename ".chicken-registry")
+(define-constant repository-environment-variable "CHICKEN_REPOSITORY")
+(define-constant home-environment-variable "CHICKEN_HOME")
+
+(define-constant core-srfi-modules '(srfi-1 srfi-4 srfi-14 srfi-18 srfi-25 srfi-37 srfi-13))
+(define-constant core-library-modules '(extras format lolevel tinyclos utils tcp regex posix match))
+
+(define-constant specially-handled-syntax-files
+  '((srfi-13 srfi-13 "srfi-13-syntax.scm")
+    (match match-support "chicken-match-macros.scm") ) )
+
+(define-constant special-syntax-files
+  '(chicken-match-macros chicken-more-macros chicken-default-entry-points chicken-highlevel-macros
+			 test-infrastructure chicken-entry-points chicken-ffi-macros) )
+
+(define-constant builtin-features
+  '(chicken srfi-22 srfi-23 srfi-30 srfi-39 srfi-8 srfi-6 srfi-2 srfi-0 srfi-10 srfi-9) )
+
+(define-constant installed-executables
+  '("chicken" "csc" "csi" "rcsi" "chicken-setup" "chicken-profile" "scheme-r4rs" "scheme-r5rs" "scheme-srfi-0"
+    "scheme-ieee-1178-1990" "scheme-chicken" "scheme-chicken-hygienic") )
