@@ -1,6 +1,6 @@
 ;;;; posix.scm - Miscellaneous file- and process-handling routines
 ;
-; Copyright (c) 2000-2003, Felix L. Winkelmann
+; Copyright (c) 2000-2004, Felix L. Winkelmann
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -28,17 +28,17 @@
 ; felix@call-with-current-continuation.org
 ;
 ; Felix L. Winkelmann
-; Steinweg 1A
-; 37130 Gleichen, OT Weissenborn
+; Unter den Gleichen 1
+; 37130 Gleichen
 ; Germany
 
 
 (declare
   (unit posix)
-  (uses scheduler regex extras script-utils)
+  (uses scheduler regex extras utils)
   (disable-interrupts)
   (usual-integrations)
-  (hide ##sys#stat yield make-nonblocking group-member)
+  (hide ##sys#stat yield make-nonblocking group-member _get-groups _ensure-groups)
   (foreign-declare #<<EOF
 #include <signal.h>
 #include <errno.h>
@@ -46,7 +46,7 @@
 static int C_not_implemented(void);
 int C_not_implemented() { return -1; }
 
-static int C_wait_status;
+static C_TLS int C_wait_status;
 
 #include <unistd.h>
 #include <sys/types.h>
@@ -57,7 +57,11 @@ static int C_wait_status;
 #include <fcntl.h>
 #include <dirent.h>
 #include <pwd.h>
+
+#ifdef HAVE_GRP_H
 #include <grp.h>
+#endif
+
 #include <sys/mman.h>
 #include <time.h>
 
@@ -96,19 +100,19 @@ static int C_wait_status;
 extern char **environ;
 #endif
 
-static char *C_exec_args[ ARG_MAX ];
-static struct utsname C_utsname;
-static struct flock C_flock;
-static DIR *temphandle;
-static struct passwd *C_user;
-static struct group *C_group;
-static int C_pipefds[ 2 ];
-static time_t C_secs;
-static struct tm C_tm;
-static fd_set C_fd_sets[ 2 ];
-static struct timeval C_timeval;
-static char C_hostbuf[ 256 ];
-static struct stat C_statbuf;
+static C_TLS char *C_exec_args[ ARG_MAX ];
+static C_TLS struct utsname C_utsname;
+static C_TLS struct flock C_flock;
+static C_TLS DIR *temphandle;
+static C_TLS struct passwd *C_user;
+static C_TLS struct group *C_group;
+static C_TLS int C_pipefds[ 2 ];
+static C_TLS time_t C_secs;
+static C_TLS struct tm C_tm;
+static C_TLS fd_set C_fd_sets[ 2 ];
+static C_TLS struct timeval C_timeval;
+static C_TLS char C_hostbuf[ 256 ];
+static C_TLS struct stat C_statbuf;
 
 #define C_mkdir(str)        C_fix(mkdir(C_c_string(str), S_IRWXU | S_IRWXG | S_IRWXO))
 #define C_chdir(str)        C_fix(chdir(C_c_string(str)))
@@ -209,8 +213,8 @@ void C_fcall C_free_exec_args() {
 
 #define C_execvp(f)         C_fix(execvp(C_data_pointer(f), C_exec_args))
 
-#if defined(__FreeBSD__) || defined(C_MACOSX) || defined(__NetBSD__) || defined(__OpenBSD__)
-static int C_uw;
+#if defined(__FreeBSD__) || defined(C_MACOSX) || defined(__NetBSD__) || defined(__OpenBSD__) || defined(__sgi__) || defined(sgi)
+static C_TLS int C_uw;
 # define C_WIFEXITED(n)      (C_uw = C_unfix(n), C_mk_bool(WIFEXITED(C_uw)))
 # define C_WIFSIGNALED(n)    (C_uw = C_unfix(n), C_mk_bool(WIFSIGNALED(C_uw)))
 # define C_WIFSTOPPED(n)     (C_uw = C_unfix(n), C_mk_bool(WIFSTOPPED(C_uw)))
@@ -241,14 +245,15 @@ static int C_uw;
 # define FILENAME_MAX          1024
 #endif
 
-static sigset_t C_sigset;
+static C_TLS sigset_t C_sigset;
 #define C_sigemptyset(d)    (sigemptyset(&C_sigset), C_SCHEME_UNDEFINED)
 #define C_sigaddset(s)      (sigaddset(&C_sigset, C_unfix(s)), C_SCHEME_UNDEFINED)
 #define C_sigprocmask(d)    C_fix(sigprocmask(SIG_SETMASK, &C_sigset, NULL))
 
-#define C_open(fn, fl, m)   C_fix(open((char *)C_data_pointer(fn), C_unfix(fl), C_unfix(m)))
+#define C_open(fn, fl, m)   C_fix(open(C_c_string(fn), C_unfix(fl), C_unfix(m)))
 #define C_read(fd, b, n)    C_fix(read(C_unfix(fd), C_data_pointer(b), C_unfix(n)))
 #define C_write(fd, b, n)   C_fix(write(C_unfix(fd), C_data_pointer(b), C_unfix(n)))
+#define C_mkstemp(t)        C_fix(mkstemp(C_c_string(t)))
 
 #define C_ftell(p)            C_fix(ftell(C_port_file(p)))
 #define C_fseek(p, n, w)      C_mk_nbool(fseek(C_port_file(p), C_unfix(n), C_unfix(w)))
@@ -267,6 +272,12 @@ static sigset_t C_sigset;
 #else
 # define C_asctime(v)        (memset(&C_tm, 0, sizeof(struct tm)), C_tm.tm_sec = C_unfix(C_block_item(v, 0)), C_tm.tm_min = C_unfix(C_block_item(v, 1)), C_tm.tm_hour = C_unfix(C_block_item(v, 2)), C_tm.tm_mday = C_unfix(C_block_item(v, 3)), C_tm.tm_mon = C_unfix(C_block_item(v, 4)), C_tm.tm_year = C_unfix(C_block_item(v, 5)), C_tm.tm_wday = C_unfix(C_block_item(v, 6)), C_tm.tm_yday = C_unfix(C_block_item(v, 7)), C_tm.tm_isdst = (C_block_item(v, 8) != C_SCHEME_FALSE), C_tm.tm_gmtoff = C_unfix(C_block_item(v, 9)), asctime(&C_tm) )
 #endif
+
+static gid_t *C_groups = NULL;
+
+#define C_get_gid(n)      C_fix(C_groups[ C_unfix(n) ])
+#define C_set_gid(n, id)  (C_groups[ C_unfix(n) ] = C_unfix(id), C_SCHEME_UNDEFINED)
+#define C_set_groups(n)   C_fix(setgroups(C_unfix(n), C_groups))
 EOF
 ) )
 
@@ -389,7 +400,7 @@ EOF
 	(##sys#check-string filename 'file-open)
 	(##sys#check-exact flags 'file-open)
 	(##sys#check-exact mode 'file-open)
-	(let ([fd (##core#inline "C_open" (##sys#make-c-string filename) flags mode)])
+	(let ([fd (##core#inline "C_open" (##sys#make-c-string (##sys#expand-home-path filename)) flags mode)])
 	  (when (eq? -1 fd)
 	    (##sys#update-errno)
 	    (##sys#signal-hook #:file-error 'file-open "can not open file" filename flags mode) )
@@ -429,6 +440,18 @@ EOF
 	  (##sys#signal-hook #:file-error 'file-write "can not write to file" fd size) )
 	n) ) ) )
 
+(define file-mkstemp
+  (let ([string-length string-length])
+    (lambda (template)
+      (##sys#check-string template 'file-mkstemp)
+      (let* ([buf (##sys#make-c-string template)]
+             [fd (##core#inline "C_mkstemp" buf)]
+             [path-length (string-length buf)])
+        (when (eq? -1 fd)
+          (##sys#update-errno)
+          (##sys#signal-hook #:file-error 'file-mkstemp "can not create temporary file" template) )
+        (values fd (##sys#substring buf 0 (fx- path-length 1) ) ) ) ) ) )
+
 
 ;;; I/O multiplexing:
 
@@ -500,13 +523,13 @@ EOF
 (define seek/cur _seek_cur)
 
 (define-foreign-variable _stat_st_size int "C_statbuf.st_size")
-(define-foreign-variable _stat_st_mtime int "C_statbuf.st_mtime")
+(define-foreign-variable _stat_st_mtime double "C_statbuf.st_mtime")
 (define-foreign-variable _stat_st_uid int "C_statbuf.st_uid")
 (define-foreign-variable _stat_st_mode int "C_statbuf.st_mode")
 
 (define (##sys#stat file)
   (let ([r (cond [(fixnum? file) (##core#inline "C_fstat" file)]
-		 [(string? file) (##core#inline "C_stat" (##sys#make-c-string file))]
+		 [(string? file) (##core#inline "C_stat" (##sys#make-c-string (##sys#expand-home-path file)))]
 		 [else (##sys#signal-hook #:type-error "bad argument type - not a fixnum or string" file)] ) ] )
     (when (fx< r 0)
       (##sys#update-errno)
@@ -535,7 +558,7 @@ EOF
     (let ([whence (if (pair? whence) (car whence) _seek_set)])
       (##sys#check-exact pos 'set-file-position!)
       (##sys#check-exact whence 'set-file-position!)
-      (when (fx< pos 0) (##sys#error 'set-file-position! "invalid negative port position" pos port))
+      (when (fx< pos 0) (##sys#signal-hook #:bounds-error 'set-file-position! "invalid negative port position" pos port))
       (unless (cond [(port? port) 
 		     (and (eq? (##sys#slot port 7) 'stream)
 			  (##core#inline "C_fseek" port pos whence) ) ]
@@ -550,21 +573,21 @@ EOF
 (define create-directory 
   (lambda (name)
     (##sys#check-string name 'create-directory)
-    (unless (zero? (##core#inline "C_mkdir" (##sys#make-c-string name)))
+    (unless (zero? (##core#inline "C_mkdir" (##sys#make-c-string (##sys#expand-home-path name))))
       (##sys#update-errno)
       (##sys#signal-hook #:file-error 'create-directory "can not create directory" name) ) ) )
 
 (define change-directory
   (lambda (name)
     (##sys#check-string name 'change-directory)
-    (unless (zero? (##core#inline "C_chdir" (##sys#make-c-string name)))
+    (unless (zero? (##core#inline "C_chdir" (##sys#make-c-string (##sys#expand-home-path name))))
       (##sys#update-errno)
       (##sys#signal-hook #:file-error 'change-directory "can not change current directory" name) ) ) )
 
 (define delete-directory
   (lambda (name)
     (##sys#check-string name 'delete-directory)
-    (unless (zero? (##core#inline "C_rmdir" (##sys#make-c-string name)))
+    (unless (zero? (##core#inline "C_rmdir" (##sys#make-c-string (##sys#expand-home-path name))))
       (##sys#update-errno)
       (##sys#signal-hook #:file-error 'delete-directory "can not delete directory" name) ) ) )
   
@@ -578,7 +601,7 @@ EOF
       (let ([buffer (make-string 256)]
 	    [handle (##sys#make-pointer)]
 	    [entry (##sys#make-pointer)] )
-	(##core#inline "C_opendir" (##sys#make-c-string spec) handle)
+	(##core#inline "C_opendir" (##sys#make-c-string (##sys#expand-home-path spec)) handle)
 	(if (##sys#null-pointer? handle)
 	    (begin
 	      (##sys#update-errno)
@@ -595,7 +618,7 @@ EOF
 
 (define (directory? fname)
   (##sys#check-string fname 'directory?)
-  (let ((info (##sys#file-info fname)))
+  (let ((info (##sys#file-info (##sys#expand-home-path fname))))
     (and info (fx= 1 (##sys#slot info 4))) ) )
 
 (define current-directory
@@ -832,13 +855,13 @@ EOF
 		 (##sys#check-string user 'user-information)
 		 (##core#inline "C_getpwnam" (##sys#make-c-string user)) ) ) ] )
     (and r
-	 (values _user-name
-		 _user-passwd
-		 _user-uid
-		 _user-gid
-		 _user-gecos
-		 _user-dir
-		 _user-shell) ) ) )
+	 (list _user-name
+	       _user-passwd
+	       _user-uid
+	       _user-gid
+	       _user-gecos
+	       _user-dir
+	       _user-shell) ) ) )
 
 (define-foreign-variable _group-name nonnull-c-string "C_group->gr_name")
 (define-foreign-variable _group-passwd nonnull-c-string "C_group->gr_passwd")
@@ -864,6 +887,54 @@ EOF
 			 (cons n (rec (fx+ i 1)))
 			 '() ) ) ) ) ) ) )
 
+(define _get-groups
+  (foreign-lambda* int ([int n])
+    "return(getgroups(n, C_groups));") )
+
+(define _ensure-groups
+  (foreign-lambda* bool ([int n])
+    "if(C_groups != NULL) C_free(C_groups);"
+    "C_groups = (gid_t *)C_malloc(sizeof(gid_t) * n);"
+    "if(C_groups == NULL) return(0);"
+    "else return(1);") )
+
+(define (get-groups)
+  (let ([n (foreign-value "getgroups(0, C_groups)" int)])
+    (when (fx< n 0)
+      (##sys#update-errno)
+      (##sys#error 'get-groups "can not retrieve supplementary group ids") )
+    (unless (_ensure-groups n)
+      (##sys#error 'get-groups "out of memory") )
+    (when (fx< (_get-groups n) 0)
+      (##sys#update-errno)
+      (##sys#error 'get-groups "can not retrieve supplementary group ids") )
+    (let loop ([i 0])
+      (if (fx>= i n)
+	  '()
+	  (cons (##core#inline "C_get_gid" i) (loop (fx+ i 1))) ) ) ) )
+
+(define (set-groups! lst0)
+  (unless (_ensure-groups (length lst0))
+    (##sys#error 'set-groups! "out of memory") )
+  (do ([lst lst0 (##sys#slot lst 1)]
+       [i 0 (fx+ i 1)] )
+      ((null? lst) 
+       (when (fx< (##core#inline "C_set_groups" i) 0)
+	 (##sys#update-errno)
+	 (##sys#error 'set-groups! "can not set supplementary group ids" lst0) ) )
+    (let ([n (##sys#slot lst 0)])
+      (##sys#check-exact n 'set-groups!)
+      (##core#inline "C_set_gid" i n) ) ) ) 
+
+(define initialize-groups
+  (let ([init (foreign-lambda int "initgroups" c-string int)])
+    (lambda (user id)
+      (##sys#check-string user 'initialize-groups)
+      (##sys#check-exact id 'initialize-groups)
+      (when (fx< (init user id) 0)
+	(##sys#update-errno)
+	(##sys#error 'initialize-groups "can not initialize supplementrary group ids" user id) ) ) ) )
+
 
 ;;; More errno codes:
 
@@ -923,7 +994,7 @@ EOF
   (lambda (fname m)
     (##sys#check-string fname 'change-file-mode)
     (##sys#check-exact m 'change-file-mode)
-    (when (fx< (##core#inline "C_chmod" (##sys#make-c-string fname) m) 0)
+    (when (fx< (##core#inline "C_chmod" (##sys#make-c-string (##sys#expand-home-path fname)) m) 0)
       (##sys#update-errno)
       (##sys#signal-hook #:file-error 'change-file-mode "can not change file mode" fname m) ) ) )
 
@@ -932,7 +1003,7 @@ EOF
     (##sys#check-string fn 'change-file-owner)
     (##sys#check-exact uid 'change-file-owner)
     (##sys#check-exact gid 'change-file-owner)
-    (when (fx< (##core#inline "C_chown" (##sys#make-c-string fn) uid gid) 0)
+    (when (fx< (##core#inline "C_chown" (##sys#make-c-string (##sys#expand-home-path fn)) uid gid) 0)
       (##sys#update-errno)
       (##sys#signal-hook #:file-error 'change-file-owner "can not change file owner" fn uid gid) ) ) )
 
@@ -954,7 +1025,7 @@ EOF
 (let ()
   (define (check filename acc loc)
     (##sys#check-string filename loc)
-    (let ([r (fx= 0 (##core#inline "C_access" (##sys#make-c-string filename) acc))])
+    (let ([r (fx= 0 (##core#inline "C_access" (##sys#make-c-string (##sys#expand-home-path filename)) acc))])
       (unless r (##sys#update-errno))
       r) )
   (set! file-read-access? (lambda (filename) (check filename _r_ok 'file-read-access?)))
@@ -968,7 +1039,11 @@ EOF
   (lambda (old new)
     (##sys#check-string old 'create-symbolic-link)
     (##sys#check-string new 'create-symbolic-link)
-    (when (fx< (##core#inline "C_symlink" (##sys#make-c-string old) (##sys#make-c-string new)) 0)
+    (when (fx< (##core#inline 
+		"C_symlink"
+		(##sys#make-c-string (##sys#expand-home-path old))
+		(##sys#make-c-string (##sys#expand-home-path new)) )
+	       0)
       (##sys#update-errno)
       (##sys#signal-hook #:file-error 'create-symbol-link "can not create symbolic link" old new) ) ) )
 
@@ -979,7 +1054,7 @@ EOF
 	[buf (make-string (fx+ _filename_max 1))] )
     (lambda (fname)
       (##sys#check-string fname 'read-symbolic-link)
-      (let ([len (##core#inline "C_readlink" (##sys#make-c-string fname) buf)])
+      (let ([len (##core#inline "C_readlink" (##sys#make-c-string (##sys#expand-home-path fname)) buf)])
 	(when (fx< len 0)
 	  (##sys#update-errno)
 	  (##sys#signal-hook #:file-error 'read-symbolic-link "can not read symbolic link" fname) )
@@ -1052,7 +1127,7 @@ EOF
 (define file-truncate
   (lambda (fname off)
     (##sys#check-number off 'file-truncate)
-    (when (fx< (cond [(string? fname) (##core#inline "C_truncate" (##sys#make-c-string fname) off)]
+    (when (fx< (cond [(string? fname) (##core#inline "C_truncate" (##sys#make-c-string (##sys#expand-home-path fname)) off)]
 		     [(fixnum? fname) (##core#inline "C_ftruncate" fname off)]
 		     [else (##sys#error 'file-truncate "invalid file" fname)] )
 	       0)
@@ -1079,7 +1154,7 @@ EOF
       (##sys#make-structure 'lock port start len) ) )
   (define (err msg lock loc)
     (##sys#update-errno)
-    (##sys#error loc msg (##sys#slot lock 1) (##sys#slot lock 2) (##sys#slot lock 3)) )
+    (##sys#signal-hook #:file-error loc msg (##sys#slot lock 1) (##sys#slot lock 2) (##sys#slot lock 3)) )
   (set! file-lock
     (lambda (port . args)
       (let ([lock (setup port args 'file-lock)])
@@ -1114,14 +1189,14 @@ EOF
     (##sys#check-string fname 'create-fifo)
     (let ([mode (if (pair? mode) (car mode) (bitwise-ior _s_irwxu (bitwise-ior _s_irwxg _s_irwxo)))])
       (##sys#check-exact mode 'create-fifo)
-      (when (fx< (##core#inline "C_mkfifo" (##sys#make-c-string fname) mode) 0)
+      (when (fx< (##core#inline "C_mkfifo" (##sys#make-c-string (##sys#expand-home-path fname)) mode) 0)
 	(##sys#update-errno)
 	(##sys#signal-hook #:file-error 'create-fifo "can not create FIFO" fname mode) ) ) ) )
 
 (define fifo?
   (lambda (filename)
     (##sys#check-string filename 'fifo?)
-    (let ([v (##sys#file-info filename)])
+    (let ([v (##sys#file-info (##sys#expand-home-path filename))])
       (##sys#update-errno)
       (if v
 	  (fx= 3 (##sys#slot v 4))
@@ -1210,11 +1285,11 @@ EOF
 ;;; Time related things:
 
 (define (seconds->local-time secs)
-  (##sys#check-exact secs 'seconds->local-time)
+  (##sys#check-number secs 'seconds->local-time)
   (##sys#decode-seconds secs #f) )
 
 (define (seconds->utc-time secs)
-  (##sys#check-exact secs 'seconds->utc-time)
+  (##sys#check-number secs 'seconds->utc-time)
   (##sys#decode-seconds secs #t) )
 
 (define seconds->string 
@@ -1340,7 +1415,7 @@ EOF
 	     [i 1 (fx+ i 1)] )
 	    ((null? al) 
 	     (setarg i #f 0)
-	     (let ([r (##core#inline "C_execvp" (##sys#make-c-string filename))])
+	     (let ([r (##core#inline "C_execvp" (##sys#make-c-string (##sys#expand-home-path filename)))])
 	       (##sys#update-errno)
 	       (when (fx= r -1)
 		 (freeargs)
@@ -1441,7 +1516,8 @@ EOF
 		     (unless (fx= fileno/stdout out2)
 		       (duplicate-fileno out2 fileno/stdout)
 		       (file-close out2) )
-		     (process-run cmd) ) ) ] 
+		     (let ([shell (or (getenv "SHELL") "/bin/sh")])
+		       (process-execute shell (list "-c" cmd)) ) ) ) ]
 		 [iclosed #f]
 		 [oclosed #f] )
 	    (file-close in1)
