@@ -1,6 +1,6 @@
 ; scheduler.scm - Basic scheduler for multithreading
 ;
-; Copyright (c) 2000-2003, Felix L. Winkelmann
+; Copyright (c) 2000-2004, Felix L. Winkelmann
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -28,8 +28,8 @@
 ; felix@call-with-current-continuation.org
 ;
 ; Felix L. Winkelmann
-; Steinweg 1A
-; 37130 Gleichen, OT Weissenborn
+; Unter den Gleichen 1
+; 37130 Gleichen
 ; Germany
 
 
@@ -38,7 +38,6 @@
   (unit scheduler)
   (disable-interrupts)
   (usual-integrations)
-  (compile-time-macros-only)
   (hide ##sys#ready-queue-head ##sys#ready-queue-tail ##sys#timeout-list
 	##sys#update-thread-state-buffer ##sys#restore-thread-state-buffer
 	##sys#remove-from-ready-queue ##sys#unblock-threads-for-i/o
@@ -55,11 +54,15 @@
 # include <time.h>
 static C_word C_msleep(C_word ms);
 C_word C_msleep(C_word ms) {
+#ifdef __CYGWIN__
+  usleep(ms * 1000);
+#else
   struct timespec ts;
   unsigned long mss = C_unfix(ms);
   ts.tv_sec = mss / 1000;
   ts.tv_nsec = (mss % 1000) * 1000000;
   nanosleep(&ts, NULL);
+#endif
   return C_SCHEME_UNDEFINED;
 }
 #endif
@@ -110,7 +113,7 @@ EOF
 		  (if (eq? tmo1 tmo2)
 		      (if (>= now tmo1)
 			  (begin
-			    (##sys#thread-unblock! tto)
+			    (##sys#thread-basic-unblock! tto)
 			    (loop (cdr lst)) )
 			  (begin
 			    (set! ##sys#timeout-list lst) 
@@ -219,23 +222,34 @@ EOF
        (lambda (t2)
 	 (dbg "  checking: " t2 " (" (##sys#slot t2 3) ") -> " (##sys#slot t2 11))
 	 (when (eq? (##sys#slot t2 11) t)
-	   (##sys#thread-unblock! t2) ) )
+	   (##sys#thread-basic-unblock! t2) ) )
        rs) ) )
   (##sys#setislot t 12 '()) )
 
-(define (##sys#thread-unblock! t)
+(define (##sys#thread-basic-unblock! t)
   (dbg "unblocking: " t)
   (##sys#setislot t 3 'ready)
   (##sys#setislot t 11 #f)
   (##sys#setislot t 4 #f)
   (##sys#add-to-ready-queue t) )
 
-(define (##sys#default-exception-handler arg)
-  (let ([ct ##sys#current-thread])
-    (dbg "exception: " ct " -> " (if (##sys#structure? arg 'condition) (##sys#slot arg 2) arg))
-    (##sys#setslot ct 7 arg)
-    (##sys#thread-kill! ct 'terminated)
-    (##sys#schedule) ) )
+(define ##sys#default-exception-handler
+  (let ([print-error-message print-error-message]
+	[display display]
+	[open-output-string open-output-string]
+	[get-output-string get-output-string] )
+    (lambda (arg)
+      (let ([ct ##sys#current-thread])
+	(dbg "exception: " ct " -> " (if (##sys#structure? arg 'condition) (##sys#slot arg 2) arg))
+	(when ##sys#warnings-enabled
+	  (let ([o (open-output-string)])
+	    (display "Warning (" o)
+	    (display ct o)
+	    (display "): " o)
+	    (print-error-message arg ##sys#standard-error (get-output-string o)) ) )
+	(##sys#setslot ct 7 arg)
+	(##sys#thread-kill! ct 'terminated)
+	(##sys#schedule) ) ) ) )
 
 
 ;;; `select()'-based blocking:
@@ -319,7 +333,7 @@ EOF
 			  (let* ([t (car threads)]
 				 [p (##sys#slot t 11)] )
 			    (when (and (pair? p) (eq? fd (car p)))
-			      (##sys#thread-unblock! t) )
+			      (##sys#thread-basic-unblock! t) )
 			    (loop2 (cdr threads)) ) ) )
 		    (cons a (loop n (cdr lst))) ) ) ) ) ) )
     (##sys#fdset-restore) ) )
@@ -351,3 +365,17 @@ EOF
   (set! ##sys#ready-queue-tail (##sys#slot vec 1))
   (set! ##sys#fd-list (##sys#slot vec 2))
   (set! ##sys#timeout-list (##sys#slot vec 3)) )
+
+
+;;; Unblock thread cleanly:
+
+(define (##sys#thread-unblock! t)
+  (when (eq? 'blocked (##sys#slot t 3))
+    (set! ##sys#timeout-list (##sys#delq t ##sys#timeout-list))
+    (set! ##sys#fd-list 
+      (let loop ([fdl ##sys#fd-list])
+	(cond [(null? fdl)]
+	      [(eq? t (##sys#slot (##sys#slot fdl 0) 1)) (##sys#slot fdl 1)]
+	      [else (cons (##sys#slot fdl 0) (loop (##sys#slot fdl 1)))] ) ) )
+    (##sys#setislot t 12 '())
+    (##sys#thread-basic-unblock! t) ) )
