@@ -6,9 +6,9 @@
 
 (declare
   (unit srfi-1)
-  (interrupts-disabled)
-  (fixnum)
-  (no-bound-checks)
+  (disable-interrupts)
+  (hide ##srfi1#cars+cdrs/no-test ##srfi1#cdrs ##srfi1#cars+ ##srfi1#really-append-map ##srfi1#cars+cdrs+
+	##srfi1#cars+cdrs ##srfi1#lset2<=)
   (extended-bindings)
   (standard-bindings not boolean? apply call-with-current-continuation eq? eqv? equal? pair? cons car cdr caar cadr
 		     cdar cddr caaar caadr cadar caddr cdaar cdadr cddar cdddr caaaar caaadr caadar caaddr cadaar
@@ -28,20 +28,26 @@
 		     open-output-file call-with-input-file call-with-output-file close-input-port close-output-port
 		     port? values call-with-values vector procedure? memq memv assq assv) )
 
+(cond-expand
+ [paranoia]
+ [else
+  (declare
+    (no-bound-checks) ) ] )
+
 (cond-expand
  [unsafe
   (eval-when (compile)
-    (define-macro (##sys#check-structure x y) '(##core#undefined))
-    (define-macro (##sys#check-range x y z) '(##core#undefined))
-    (define-macro (##sys#check-pair x) '(##core#undefined))
-    (define-macro (##sys#check-list x) '(##core#undefined))
-    (define-macro (##sys#check-symbol x) '(##core#undefined))
-    (define-macro (##sys#check-string x) '(##core#undefined))
-    (define-macro (##sys#check-char x) '(##core#undefined))
-    (define-macro (##sys#check-exact x) '(##core#undefined))
-    (define-macro (##sys#check-port x) '(##core#undefined))
-    (define-macro (##sys#check-number x) '(##core#undefined))
-    (define-macro (##sys#check-byte-vector x) '(##core#undefined)) ) ]
+    (define-macro (##sys#check-structure . _) '(##core#undefined))
+    (define-macro (##sys#check-range . _) '(##core#undefined))
+    (define-macro (##sys#check-pair . _) '(##core#undefined))
+    (define-macro (##sys#check-list . _) '(##core#undefined))
+    (define-macro (##sys#check-symbol . _) '(##core#undefined))
+    (define-macro (##sys#check-string . _) '(##core#undefined))
+    (define-macro (##sys#check-char . _) '(##core#undefined))
+    (define-macro (##sys#check-exact . _) '(##core#undefined))
+    (define-macro (##sys#check-port . _) '(##core#undefined))
+    (define-macro (##sys#check-number . _) '(##core#undefined))
+    (define-macro (##sys#check-byte-vector . _) '(##core#undefined)) ) ]
  [else] )
 
 (register-feature! 'srfi-1)
@@ -275,14 +281,14 @@
 
 (define (make-list len . maybe-elt)
 ;  (check-arg (lambda (n) (and (integer? n) (>= n 0))) len make-list)
-  (##sys#check-exact len)
+  (##sys#check-exact len 'make-list)
   (let ((elt (cond ((null? maybe-elt) #f) ; Default value
 		   ((null? (cdr maybe-elt)) (car maybe-elt))
-		   (else (error "Too many arguments to MAKE-LIST"
+		   (else (##sys#error 'make-list "Too many arguments to MAKE-LIST"
 				(cons len maybe-elt))))))
-    (do ((i len (- i 1))
+    (do ((i len (fx- i 1))
 	 (ans '() (cons elt ans)))
-	((<= i 0) ans))))
+	((fx<= i 0) ans))))
 
 
 ;(define (list . ans) ans)	; R4RS
@@ -293,10 +299,10 @@
 (define (list-tabulate len proc)
 ;  (check-arg (lambda (n) (and (integer? n) (>= n 0))) len list-tabulate)
 ;  (check-arg procedure? proc list-tabulate)
-  (##sys#check-exact len)
-  (do ((i (- len 1) (- i 1))
+  (##sys#check-exact len 'list-tabulate)
+  (do ((i (fx- len 1) (fx- i 1))
        (ans '() (cons (proc i) ans)))
-      ((< i 0) ans)))
+      ((fx< i 0) ans)))
 
 ;;; (cons* a1 a2 ... an) = (cons a1 (cons a2 (cons ... an)))
 ;;; (cons* a1) = a1	(cons* a1 a2 ...) = (cons a1 (cons* a2 ...))
@@ -320,16 +326,15 @@
 ;;; IOTA count [start step]	(start start+step ... start+(count-1)*step)
 
 (define (iota count . maybe-start+step)
-  (##sys#check-exact count)
 ;  (check-arg integer? count iota)
-  (##sys#check-exact count)
-  (if (< count 0) (error "Negative step count" iota count))
+  (##sys#check-number count 'iota)
+  (if (< count 0) (##sys#error 'iota "Negative step count" iota count))
   (let ((start (:optional maybe-start+step 0))
 	(step (if (pair? maybe-start+step)
 		  (:optional (cdr maybe-start+step) 1)
 		  1) ) )
-    (##sys#check-exact start)
-    (##sys#check-exact step)
+    (##sys#check-number start 'iota)
+    (##sys#check-number step 'iota)
 ;    (check-arg number? start iota)
 ;    (check-arg number? step iota)
     (let ((last-val (+ start (* (- count 1) step))))
@@ -395,7 +400,9 @@
 ;;; Note that this definition rules out circular lists -- and this
 ;;; function is required to detect this case and return false.
 
-(define (proper-list? x)
+(define proper-list? list?)
+
+#;(define (proper-list? x)
   (let lp ((x x) (lag x))
     (if (pair? x)
 	(let ((x (cdr x)))
@@ -471,11 +478,11 @@
   (let lp ((x x) (lag x) (len 0))
     (if (pair? x)
 	(let ((x (cdr x))
-	      (len (+ len 1)))
+	      (len (fx+ len 1)))
 	  (if (pair? x)
 	      (let ((x   (cdr x))
 		    (lag (cdr lag))
-		    (len (+ len 1)))
+		    (len (fx+ len 1)))
 		(and (not (eq? x lag)) (lp x lag len)))
 	      len))
 	len)))
@@ -531,30 +538,30 @@
 (define (tenth   x) (cadr (cddddr (cddddr x))))
 
 (define (car+cdr pair)
-  (##sys#check-pair pair)
+  (##sys#check-pair pair 'car+cdr)
   (values (##sys#slot pair 0) (##sys#slot pair 1)) )
 
 ;;; take & drop
 
 (define (take lis k)
-  (##sys#check-exact k)
+  (##sys#check-exact k 'take)
 ;  (check-arg integer? k take)
   (let recur ((lis lis) (k k))
-    (if (= 0 k) '()
+    (if (eq? 0 k) '()
 	(cons (car lis)
-	      (recur (cdr lis) (- k 1))))))
+	      (recur (cdr lis) (fx- k 1))))))
 
 (define (drop lis k)
-  (##sys#check-exact k)
+  (##sys#check-exact k 'drop)
 ;  (check-arg integer? k drop)
   (let iter ((lis lis) (k k))
-    (if (= 0 k) lis (iter (cdr lis) (- k 1)))))
+    (if (eq? 0 k) lis (iter (cdr lis) (fx- k 1)))))
 
 (define (take! lis k)
-  (##sys#check-exact k)
+  (##sys#check-exact k 'take!)
 ;  (check-arg integer? k take!)
-  (if (= 0 k) '()
-      (begin (set-cdr! (drop lis (- k 1)) '())
+  (if (eq? 0 k) '()
+      (begin (set-cdr! (drop lis (fx- k 1)) '())
 	     lis)))
 
 ;;; TAKE-RIGHT and DROP-RIGHT work by getting two pointers into the list, 
@@ -635,18 +642,18 @@
 ;      (list-tail lis k)))
 
 (define (split-at x k)
-  (##sys#check-exact k)
+  (##sys#check-exact k 'split-at)
 ;  (check-arg integer? k split-at)
   (let recur ((lis x) (k k))
-    (if (= 0 k) (values '() lis)
-	(receive (prefix suffix) (recur (cdr lis) (- k 1))
+    (if (eq? 0 k) (values '() lis)
+	(receive (prefix suffix) (recur (cdr lis) (fx- k 1))
 	  (values (cons (car lis) prefix) suffix)))))
 
 (define (split-at! x k)
-  (##sys#check-exact k)
+  (##sys#check-exact k 'split-at!)
 ;  (check-arg integer? k split-at!)
-  (if (= 0 k) (values '() x)
-      (let* ((prev (drop x (- k 1)))
+  (if (eq? 0 k) (values '() x)
+      (let* ((prev (drop x (fx- k 1)))
 	     (suffix (cdr prev)))
 	(set-cdr! prev '())
 	(values x suffix))))
@@ -849,12 +856,12 @@
 	    (receive (as ds) (##srfi1#cars+cdrs lists)
 	      (if (null? as) i
 		  (lp (cdr list1) ds
-		      (if (apply pred (car list1) as) (+ i 1) i))))))
+		      (if (apply pred (car list1) as) (fx+ i 1) i))))))
 
       ;; Fast path
       (let lp ((lis list1) (i 0))
 	(if (null-list? lis) i
-	    (lp (cdr lis) (if (pred (car lis)) (+ i 1) i))))))
+	    (lp (cdr lis) (if (pred (car lis)) (fx+ i 1) i))))))
 
 
 ;;; fold/unfold
@@ -1456,12 +1463,12 @@
 	(receive (heads tails) (##srfi1#cars+cdrs lists)
 	  (and (pair? heads)
 	       (if (apply pred heads) n
-		   (lp tails (+ n 1))))))
+		   (lp tails (fx+ n 1))))))
 
       ;; Fast path
       (let lp ((lis lis1) (n 0))
 	(and (not (null-list? lis))
-	     (if (pred (car lis)) n (lp (cdr lis) (+ n 1)))))))
+	     (if (pred (car lis)) n (lp (cdr lis) (fx+ n 1)))))))
 
 ;;; Reverse
 ;;;;;;;;;;;
@@ -1604,7 +1611,7 @@
 	    ;; Compute a-b and a^b, then compute b-(a^b) and
 	    ;; cons it onto the front of a-b.
 	    (receive (a-b a-int-b)   (lset-diff+intersection = a b)
-	      (cond ((null? a-b)     (lset-difference b a =))
+	      (cond ((null? a-b)     (lset-difference = b a))
 		    ((null? a-int-b) (append b a))
 		    (else (fold (lambda (xb ans)
 				  (if (member xb a-int-b =) ans (cons xb ans)))
@@ -1626,7 +1633,7 @@
 	    ;; Compute a-b and a^b, then compute b-(a^b) and
 	    ;; cons it onto the front of a-b.
 	    (receive (a-b a-int-b)   (lset-diff+intersection! = a b)
-	      (cond ((null? a-b)     (lset-difference! b a =))
+	      (cond ((null? a-b)     (lset-difference! = b a))
 		    ((null? a-int-b) (append! b a))
 		    (else (pair-fold (lambda (b-pair ans)
 				       (if (member (car b-pair) a-int-b =) ans
