@@ -8,6 +8,7 @@
   (hide %string-prefix? %string-hash %finish-string-concatenate-reverse %string-suffix-length %string-prefix-length
 	%string-map %string-copy! %string-compare %substring/shared %string-suffix? %multispan-repcopy!
 	%string-prefix-length-ci %string-suffix-length-ci %string-prefix-ci? %string-suffix-ci?
+	##srfi13#traverse
 	%string-titlecase! %string-map! %kmp-search %string-compare-ci ##srfi13#string-fill!)
   (standard-bindings not boolean? apply call-with-current-continuation eq? eqv? equal? pair? cons car cdr caar cadr
 		     cdar cddr caaar caadr cadar caddr cdaar cdadr cddar cdddr caaaar caaadr caadar caaddr cadaar
@@ -1373,30 +1374,33 @@
 ;;; comparison testing with fancier implementations.
 ;;; See below for fast KMP version.
 
-;(define (string-contains string substring . maybe-starts+ends)
-;  (let-string-start+end2 (start1 end1 start2 end2) 
-;                         string-contains string substring maybe-starts+ends
-;    (let* ((len (- end2 start2))
-;	   (i-bound (- end1 len)))
-;      (let lp ((i start1))
-;	(and (< i i-bound)
-;	     (if (string= string substring i (+ i len) start2 end2)
-;		 i
-;		 (lp (+ i 1))))))))
+(define (string-contains string substring . maybe-starts+ends)
+  (let-string-start+end2 (start1 end1 start2 end2) 
+                         string-contains string substring maybe-starts+ends
+    (let* ((len (fx- end2 start2))
+	   (i-bound (fx- end1 len)))
+      (let lp ((i start1))
+	(and (fx<= i i-bound)
+	     (if (string= string substring i (fx+ i len) start2 end2)
+		 i
+		 (lp (fx+ i 1))))))))
+
+(define (string-contains-ci string substring . maybe-starts+ends)
+  (let-string-start+end2 (start1 end1 start2 end2) 
+                         string-contains string substring maybe-starts+ends
+    (let* ((len (fx- end2 start2))
+	   (i-bound (fx- end1 len)))
+      (let lp ((i start1))
+	(and (fx<= i i-bound)
+	     (if (string-ci= string substring i (fx+ i len) start2 end2)
+		 i
+		 (lp (fx+ i 1))))))))
 
 
 ;;; Searching for an occurrence of a substring
 ;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
 
-(define (string-contains text pattern . maybe-starts+ends)
-  (let-string-start+end2 (t-start t-end p-start p-end)
-                         string-contains text pattern maybe-starts+ends
-    (%kmp-search pattern text char=? p-start p-end t-start t-end)))
-
-(define (string-contains-ci text pattern . maybe-starts+ends)
-  (let-string-start+end2 (t-start t-end p-start p-end)
-                         string-contains-ci text pattern maybe-starts+ends
-    (%kmp-search pattern text char-ci=? p-start p-end t-start t-end)))
+; this is broken. Thanks, Olin! ;-)
 
 
 ;;; Knuth-Morris-Pratt string searching
@@ -1587,6 +1591,8 @@
 	(string-set! s j ci)))))
 
 
+#| this is already available in library.scm:
+
 (define (reverse-list->string clist)
   (let* ((len (length clist))
 	 (s (make-string len)))
@@ -1594,6 +1600,7 @@
 	((not (pair? clist)))
       (string-set! s i (car clist)))
     s))
+|#
 
 
 ;(define (string->list s . maybe-start+end)
