@@ -1,6 +1,6 @@
 ;;;; srfi-18.scm - Simple thread unit - felix
 ;
-; Copyright (c) 2000-2003, Felix L. Winkelmann
+; Copyright (c) 2000-2004, Felix L. Winkelmann
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -28,8 +28,8 @@
 ; felix@call-with-current-continuation.org
 ;
 ; Felix L. Winkelmann
-; Steinweg 1A
-; 37130 Gleichen, OT Weissenborn
+; Unter den Gleichen 1
+; 37130 Gleichen
 ; Germany
 
 
@@ -38,7 +38,7 @@
  (uses scheduler)
  (disable-interrupts)
  (usual-integrations)
- (hide ##sys#compute-time-limit ##sys#delq) )
+ (hide ##sys#compute-time-limit) )
 
 (cond-expand
  [paranoia]
@@ -77,12 +77,6 @@
 
 ;;; Helper routines:
 
-(define (##sys#delq x lst)
-  (let loop ([lst lst])
-    (cond ((null? lst) lst)
-	  ((eq? x (##sys#slot lst 0)) (##sys#slot lst 1))
-	  (else (cons (##sys#slot lst 0) (loop (##sys#slot lst 1)))) ) ) )
-
 (define ##sys#compute-time-limit
   (let ([truncate truncate])
     (lambda (tm)
@@ -94,10 +88,28 @@
 
 ;;; Time objects:
 
-(define (current-time) 
-  (let* ([s (##sys#fudge 20)]
-	 [ms (##sys#fudge 21)] )
-    (##sys#make-structure 'time (##sys#fudge 22) s ms) ) )
+(declare
+  (foreign-declare #<<EOF
+static C_TLS long C_ms;
+#define C_get_seconds   C_seconds(&C_ms)
+EOF
+) )
+
+(define-foreign-variable C_get_seconds double)
+(define-foreign-variable C_startup_time_seconds double)
+(define-foreign-variable C_ms long)
+
+(define (current-time)
+  (let* ([s C_get_seconds]
+	 [ss C_startup_time_seconds] 
+	 [ms C_ms] )
+    (##sys#make-structure
+     'time
+     (inexact->exact (truncate (+ (* (- s ss) 1000) C_ms)))
+     s
+     C_ms) ) )
+
+(define srfi-18:current-time current-time)
 
 (define (time->seconds tm)
   (##sys#check-structure tm 'time 'time->seconds)
@@ -105,12 +117,15 @@
 
 (define (seconds->time n)
   (##sys#check-number n 'seconds->time)
-  (let* ([n2 (- n (##sys#fudge 23))]	; seconds since startup
-	 [n3 (inexact->exact (truncate (* n2 1000)))] )	; milliseconds since startup
-    (##sys#make-structure 'time n3 (inexact->exact (truncate n)) 0) ) )
+  (let* ([n2 (- n C_startup_time_seconds)] ; seconds since startup
+	 [ms (truncate (* 1000 (##sys#flonum-fraction n)))] ; milliseconds
+	 [n3 (inexact->exact (truncate (+ (* n2 1000) ms)))] ) ; milliseconds since startup
+    (##sys#make-structure 'time n3 (truncate n) (inexact->exact ms)) ) )
 
 (define (time? x) (##sys#structure? x 'time))
 
+(define srfi-18:time? time?)
+
 
 ;;; Exception handling:
 
@@ -415,7 +430,7 @@
 	     [t0s (##sys#slot t0 3)] )
 	(##sys#setslot cvar 2 (##sys#slot ts 1))
 	(when (or (eq? t0s 'blocked) (eq? t0s 'sleeping))
-	  (##sys#thread-unblock! t0) ) ) ) ) )
+	  (##sys#thread-basic-unblock! t0) ) ) ) ) )
 
 (define (condition-variable-broadcast! cvar)
   (##sys#check-structure cvar 'condition-variable 'condition-variable-broadcast!)
@@ -424,7 +439,7 @@
    (lambda (ti)
      (let ([tis (##sys#slot ti 3)])
        (when (or (eq? tis 'blocked) (eq? tis 'sleeping))
-	 (##sys#thread-unblock! ti) ) ) )
+	 (##sys#thread-basic-unblock! ti) ) ) )
    (##sys#slot cvar 2) )
   (##sys#setislot cvar 2 '()) )
 
