@@ -1,36 +1,4 @@
 ;;;; srfi-25.scm - Jussi Piitulainen's array library
-;
-; Copyright (c) 2000-2003, Felix L. Winkelmann
-; All rights reserved.
-;
-; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
-; conditions are met:
-;
-;   Redistributions of source code must retain the above copyright notice, this list of conditions and the following
-;     disclaimer. 
-;   Redistributions in binary form must reproduce the above copyright notice, this list of conditions and the following
-;     disclaimer in the documentation and/or other materials provided with the distribution. 
-;   Neither the name of the author nor the names of its contributors may be used to endorse or promote
-;     products derived from this software without specific prior written permission. 
-;
-; THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS
-; OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
-; AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS OR
-; CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
-; CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
-; SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
-; THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR
-; OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
-; POSSIBILITY OF SUCH DAMAGE.
-;
-; Send bugs, suggestions and ideas to: 
-;
-; felix@call-with-current-continuation.org
-;
-; Felix L. Winkelmann
-; Steinweg 1A
-; 37130 Gleichen, OT Weissenborn
-; Germany
 
 
 (declare
@@ -127,13 +95,19 @@
 (define (array? obj)
    (array:array? obj))
 
+(define-macro (check-array x loc)
+  (let ([var (gensym)])
+    `(let ([,var ,x])
+       (or (##core#check (array:array? ,var))
+	   (##sys#signal-hook #:type-error ,loc "bad argument type - not an array" ,var) ) ) ) )
+
 ;;; (make-array shape)
 ;;; (make-array shape obj)
 ;;; makes array of `shape' with each cell containing `obj' initially.
 
 (define (make-array shape . rest)
   (or (##core#check (array:good-shape? shape))
-      (##sys#signal-hook #:type-error 'make-array "shape is not a shape" shape))
+      (##sys#signal-hook #:type-error 'make-array "bad argument type - not a valid shape" shape))
   (apply array:make-array shape rest))
 
 (define (array:make-array shape . rest)
@@ -174,7 +148,7 @@
 
 (define (array shape . elts)
   (or (##core#check (array:good-shape? shape))
-      (##sys#signal-hook #:type-error 'array "not a shape" shape) )
+      (##sys#signal-hook #:type-error 'array "bad argument type - not a valid shape" shape) )
   (let ((size (array:size shape)))
     (let ((vector (list->vector elts)))
       (or (##core#check (= (vector-length vector) size))
@@ -193,7 +167,8 @@
 ;;; returns the number of dimensions of `array'.
 
 (define (array-rank array)
-   (quotient (vector-length (array:shape array)) 2))
+  (check-array array 'array-rank)
+  (quotient (vector-length (array:shape array)) 2))
 
 ;;; (array-start array k)
 ;;; returns the lower bound index of array along dimension k. This is
@@ -201,6 +176,7 @@
 ;;; empty.
 
 (define (array-start array d)
+  (check-array array 'array-start)
   (vector-ref (array:shape array) (+ d d)))
 
 ;;; (array-end array k)
@@ -209,6 +185,7 @@
 ;;; the lower bound along it.
 
 (define (array-end array d)
+  (check-array array 'array-end)
   (vector-ref (array:shape array) (+ d d 1)))
 
 ;;; (share-array array shape proc)
@@ -1276,7 +1253,7 @@
 		     "bad index object: "
 		     x)))))))
 
-(define (array:make-locative a x)
+(define (array:make-locative a x weak?)
   (or (##core#check (array:array? a))
       (##sys#signal-hook #:type-error 'array:make-locative "not an array"))
   (let ((shape (array:shape a)))
@@ -1290,21 +1267,23 @@
 		 (##sys#signal-hook #:type-error 'array:make-locative "not an index object" x)))))
     (if (vector? x)
 	(##core#inline_allocate
-	 ("C_a_i_make_locative" 4)
+	 ("C_a_i_make_locative" 5)
 	 0
 	 (array:vector a)
 	 (array:index/vector
 	  (quotient (vector-length shape) 2)
 	  (array:index a)
-	  x) )
+	  x) 
+	 weak?)
 	(if (##core#check (array:array? x))
 	    (##core#inline_allocate
-	     ("C_a_i_make_locative" 4)
+	     ("C_a_i_make_locative" 5)
 	     0
 	     (array:vector a)
 	     (array:index/array
 	      (quotient (vector-length shape) 2)
 	      (array:index a)
 	      (array:vector x)
-	      (array:index x)) )
+	      (array:index x)) 
+	     weak?)
 	    (##sys#signal-hook #:type-error 'array:make-locative "bad index object: " x)))))
