@@ -1,6 +1,6 @@
 ;;;; srfi-4.scm - Homogeneous numeric vectors
 ;
-; Copyright (c) 2000-2003, Felix L. Winkelmann
+; Copyright (c) 2000-2004, Felix L. Winkelmann
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -28,8 +28,8 @@
 ; felix@call-with-current-continuation.org
 ;
 ; Felix L. Winkelmann
-; Steinweg 1A
-; 37130 Gleichen, OT Weissenborn
+; Unter den Gleichen 1
+; 37130 Gleichen
 ; Germany
 
 
@@ -39,9 +39,9 @@
  (usual-integrations)
  (hide ##sys#u8vector-set! ##sys#s8vector-set! ##sys#u16vector-set! ##sys#s16vector-set!
        ##sys#u32vector-set! ##sys#s32vector-set! ##sys#f32vector-set! ##sys#f64vector-set!
-       ##sys#u8vector-ref ##sys#s8vector-ref ##sys#u16vector-ref ##sys#s16vector-ref
+       ##sys#u8vector-ref ##sys#s8vector-ref ##sys#u16vector-ref ##sys#s16vector-ref subvector
        ##sys#u32vector-ref ##sys#s32vector-ref ##sys#f32vector-ref ##sys#f64vector-ref)
- (foreign-declare "
+ (foreign-declare #<<EOF
 #define C_u8peek(b, i)         C_fix(((unsigned char *)C_data_pointer(b))[ C_unfix(i) ])
 #define C_s8peek(b, i)         C_fix(((char *)C_data_pointer(b))[ C_unfix(i) ])
 #define C_u16peek(b, i)        C_fix(((unsigned short *)C_data_pointer(b))[ C_unfix(i) ])
@@ -56,7 +56,11 @@
 #define C_s32poke(b, i, x)     ((((C_s32 *)C_data_pointer(b))[ C_unfix(i) ] = C_num_to_int(x)), C_SCHEME_UNDEFINED)
 #define C_f32poke(b, i, x)     ((((float *)C_data_pointer(b))[ C_unfix(i) ] = C_flonum_magnitude(x)), C_SCHEME_UNDEFINED)
 #define C_f64poke(b, i, x)     ((((double *)C_data_pointer(b))[ C_unfix(i) ] = C_flonum_magnitude(x)), C_SCHEME_UNDEFINED)
-") )
+#define C_copy_subvector(to, from, start_from, bytes)   \
+  (C_memcpy((C_char *)C_data_pointer(to), (C_char *)C_data_pointer(from) + C_unfix(start_from), C_unfix(bytes)), \
+    C_SCHEME_UNDEFINED)
+EOF
+) )
 
 (cond-expand
  [paranoia]
@@ -473,20 +477,20 @@
 (set! ##sys#user-read-hook
   (let ([old-hook ##sys#user-read-hook]
 	[read read]
-	[consers (list "u8" list->u8vector
-		       "s8" list->s8vector
-		       "u16" list->u16vector
-		       "s16" list->s16vector
-		       "u32" list->u32vector
-		       "s32" list->s32vector
-		       "f32" list->f32vector
-		       "f64" list->f64vector) ] )
+	[consers (list 'u8 list->u8vector
+		       's8 list->s8vector
+		       'u16 list->u16vector
+		       's16 list->s16vector
+		       'u32 list->u32vector
+		       's32 list->s32vector
+		       'f32 list->f32vector
+		       'f64 list->f64vector) ] )
     (lambda (char port)
       (if (memq char '(#\u #\s #\f #\U #\S #\F))
 	  (let* ([x (read port)]
-		 [tag (and (symbol? x) (##sys#symbol->string x))] )
-	    (cond [(string=? tag "f") #f]
-		  [(member tag consers) => (lambda (c) ((##sys#slot (##sys#slot c 1) 0) (read port)))]
+		 [tag (and (symbol? x) x)] )
+	    (cond [(or (eq? tag 'f) (eq? tag 'F)) #f]
+		  [(memq tag consers) => (lambda (c) ((##sys#slot (##sys#slot c 1) 0) (read port)))]
 		  [else (##sys#error "illegal bytevector syntax" tag)] ) )
 	  (old-hook char port) ) ) ) )
 
@@ -512,4 +516,29 @@
 	      (else (old-hook x readable port)) ) ) ) ) )
 
 
+;;; Subvectors:
+
+(define (subvector v t es from to loc)
+  (##sys#check-structure v t loc)
+  (let* ([bv (##sys#slot v 1)]
+	 [len (##sys#size bv)]
+	 [ilen (fx/ len es)] )
+    (##sys#check-range from 0 ilen loc)
+    (let* ([size2 (fx* es (fx- to from))]
+	   [bv2 (##sys#allocate-vector size2 #t #f #t)] )
+      (##core#inline "C_string_to_bytevector" bv2)
+      (let ([v (##sys#make-structure t bv2)])
+	(##core#inline "C_copy_subvector" bv2 bv (fx* from es) size2)
+	v) ) ) )
+
+(define (subu8vector v from to) (subvector v 'u8vector 1 from to 'subu8vector))
+(define (subu16vector v from to) (subvector v 'u16vector 2 from to 'subu16vector))
+(define (subu32vector v from to) (subvector v 'u32vector 4 from to 'subu32vector))
+(define (subs8vector v from to) (subvector v 's8vector 1 from to 'subs8vector))
+(define (subs16vector v from to) (subvector v 's16vector 2 from to 'subs16vector))
+(define (subs32vector v from to) (subvector v 's32vector 4 from to 'subs32vector))
+(define (subf32vector v from to) (subvector v 'f32vector 4 from to 'subf32vector))
+(define (subf64vector v from to) (subvector v 'f64vector 8 from to 'subf64vector))
+
+
 (register-feature! 'srfi-4)
