@@ -1,6 +1,6 @@
 ;;;; support.scm - Miscellaneous support code for the CHICKEN compiler
 ;
-; Copyright (c) 2000-2003, Felix L. Winkelmann
+; Copyright (c) 2000-2004, Felix L. Winkelmann
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -28,8 +28,8 @@
 ; felix@call-with-current-continuation.org
 ;
 ; Felix L. Winkelmann
-; Steinweg 1A
-; 37130 Gleichen, OT Weissenborn
+; Unter den Gleichen 1
+; 37130 Gleichen
 ; Germany
 
 
@@ -42,24 +42,25 @@
   non-foldable-standard-bindings foldable-standard-bindings non-foldable-extended-bindings foldable-extended-bindings
   standard-bindings-that-never-return-false side-effect-free-standard-bindings-that-never-return-false
   installation-home optimization-iterations compiler-cleanup-hook decompose-lambda-list
-  foreign-type-table-size file-io-only
+  foreign-type-table-size file-io-only banner custom-declare-alist
   unit-name insert-timer-checks used-units source-filename pending-canonicalizations
   foreign-declarations block-compilation analysis-database-size line-number-database-size
   target-heap-size target-stack-size try-harder default-installation-home 
   default-default-target-heap-size default-default-target-stack-size verbose-mode original-program-size
   current-program-size line-number-database-2 foreign-lambda-stubs immutable-constants foreign-variables
   rest-parameters-promoted-to-vector inline-table inline-table-used constant-table constants-used mutable-constants
-  dependency-list broken-constant-nodes inline-substitutions-enabled
-  always-bound-to-procedure block-variable-literal copy-node!
+  dependency-list broken-constant-nodes inline-substitutions-enabled no-c-syntax-checks
+  always-bound-to-procedure block-variable-literal copy-node! valid-c-identifier?
   direct-call-ids foreign-type-table first-analysis scan-sharp-greater-string enable-sharp-greater-read-syntax
   expand-profile-lambda profile-lambda-list profile-lambda-index profile-info-vector-name
   initialize-compiler canonicalize-expression expand-foreign-lambda update-line-number-database scan-toplevel-assignments
   perform-cps-conversion analyze-expression simplifications perform-high-level-optimizations perform-pre-optimization!
-  reorganize-recursive-bindings substitution-table simplify-named-call
+  reorganize-recursive-bindings substitution-table simplify-named-call check-c-syntax
   perform-closure-conversion prepare-for-code-generation compiler-source-file create-foreign-stub expand-foreign-lambda*
-  transform-direct-lambdas! finish-foreign-result compressable-literal
+  transform-direct-lambdas! finish-foreign-result compressable-literal csc-control-file
   debugging-chicken warnings-enabled bomb check-signature posq stringify symbolify flonum? build-lambda-list
-  string->c-identifier c-ify-string words words->bytes check-and-open-input-file close-checked-input-file fold-inner constant?
+  string->c-identifier c-ify-string words words->bytes check-and-open-input-file close-checked-input-file fold-inner
+  constant? compute-module-exports register-optimizer user-level-optimizers
   collapsable-literal? immediate? canonicalize-begin-body extract-mutable-constants string->expr get get-all
   put! collect! count! get-line get-line-2 find-lambda-container display-analysis-database varnode qnode 
   build-node-graph build-expression-tree fold-boolean inline-lambda-bindings match-node expression-has-side-effects?
@@ -67,17 +68,18 @@
   pprint-expressions-to-file foreign-type-check estimate-foreign-result-size scan-used-variables scan-free-variables
   topological-sort print-version print-usage initialize-analysis-database estimate-foreign-result-location-size
   real-name real-name-table set-real-name! real-name2 display-real-name-table display-line-number-database
-  product copyright compiler-features default-declarations units-used-by-default words-per-flonum
+  copyright compiler-features default-declarations units-used-by-default words-per-flonum
   foreign-string-result-reserve parameter-limit eq-inline-operator optimizable-rest-argument-operators
   membership-test-operators membership-unfold-limit valid-compiler-options valid-compiler-options-with-argument
-  default-optimization-iterations chop-separator chop-extension follow-without-loop
-  generate-code make-variable-list make-argument-list generate-foreign-stubs foreign-type-declaration
-  foreign-argument-conversion foreign-result-conversion final-foreign-type debugging
-  make-random-name foreign-type-convert-result foreign-type-convert-argument}
+  default-optimization-iterations chop-separator chop-extension follow-without-loop dump-exported-globals
+  generate-code make-variable-list make-argument-list generate-foreign-stubs foreign-type-declaration optimizers-used
+  foreign-argument-conversion foreign-result-conversion final-foreign-type debugging export-list block-globals
+  make-random-name foreign-type-convert-result foreign-type-convert-argument process-custom-declaration}
 
 
 (include "parameters")
 (include "tweaks")
+(include "banner")
 
 
 ;;; Debugging and error-handling stuff:
@@ -86,6 +88,7 @@
 
 (define debugging-chicken '())
 (define warnings-enabled #t)
+(define optimizers-used #f)
 
 (define (bomb . msg-and-args)
   (if (pair? msg-and-args)
@@ -181,6 +184,14 @@
 			(loop (cdr chars)) )
 		(cons c (loop (cdr chars))) ) ) ) ) ) ) )
 
+(define (valid-c-identifier? name)
+  (let ([str (string->list (->string name))])
+    (and (pair? str)
+	 (let ([c0 (car str)])
+	   (and (or (char-alphabetic? c0) (char=? #\_ c0))
+		(any (lambda (c) (or (char-alphabetic? c) (char-numeric? c) (char=? #\_ c)))
+		     (cdr str) ) ) ) ) ) )
+
 (eval-when (load)
   (define words (foreign-lambda int "C_bytestowords" int)) 
   (define words->bytes (foreign-lambda int "C_wordstobytes" int)) )
@@ -225,11 +236,13 @@
       (char? x)
       (string? x)
       (boolean? x)
+      (eof-object? x)
       (and (pair? x) (eq? 'quote (car x))) ) )
 
 (define (collapsable-literal? x)
   (or (boolean? x)
       (char? x)
+      (eof-object? x)
       (number? x)
       (symbol? x) ) )
 
@@ -237,6 +250,7 @@
   (or (fixnum? x)
       (eq? (##core#undefined) x)
       (null? x)
+      (eof-object? x)
       (char? x)
       (boolean? x) ) )
 
@@ -459,10 +473,16 @@
 
 ;;; Node creation and -manipulation:
 
+;; Note: much of this stuff will be overridden by the inline-definitions in "tweaks.scm".
+
 (define-record node
   class					; symbol
   parameters				; (value...)
-  subexpressions)			; (node...)
+  subexpressions			; (node...)
+  user-field)				; anything
+
+(define (make-node c p s)
+  (##sys#make-structure 'node c p s #f) ) ; this kludge is for allowing the inlined `make-node'
 
 (define (varnode var) (make-node '##core#variable (list var) '()))
 (define (qnode const) (make-node 'quote (list const) '()))
@@ -494,8 +514,16 @@
 				 (append (map (lambda (b) (walk (cadr b))) (cadr x))
 					 (list (walk body)) ) ) ) ) )
 	       ((lambda) (make-node 'lambda (list (cadr x)) (list (walk (caddr x)))))
-	       ((set! ##core#inline ##core#callunit ##core#primitive ##core#proc) 
+	       ((##core#primitive)
+		(let ([arg (cadr x)])
+		  (make-node
+		   (car x)
+		   (list (if (and (pair? arg) (eq? 'quote (car arg))) (cadr arg) arg))
+		   (map walk (cddr x)) ) ) )
+	       ((set! ##core#inline ##core#callunit) 
 		(make-node (car x) (list (cadr x)) (map walk (cddr x))) )
+	       ((##core#proc)
+		(make-node '##core#proc (list (cadr x) #t) '()) )
 	       ((##core#set!) (make-node 'set! (list (cadr x)) (map walk (cddr x))))
 	       ((##core#foreign-callback-wrapper)
 		(let ([name (cadr (second x))])
@@ -582,7 +610,9 @@
 	(if rest
 	    (make-node
 	     'let (list rest)
-	     (list (make-node '##core#inline_allocate (list "C_a_i_list" (* 3 (length rargs))) rargs)
+	     (list (if (null? rargs)
+		       (qnode '())
+		       (make-node '##core#inline_allocate (list "C_a_i_list" (* 3 (length rargs))) rargs) )
 		   body) )
 	    body)
 	(take vars argc)
@@ -713,6 +743,18 @@
      (printf ";   known call sites: \t~s\n" sites) 
      (printf ";   database entries: \t~s\n" entries) ) ) )
 
+(define (dump-exported-globals db)
+  (unless block-compilation
+    (##sys#hash-table-for-each
+     (lambda (sym plist)
+       (when (and (assq 'global plist) 
+		  (assq 'assigned plist)
+		  (or (and export-list (memq sym export-list))
+		      (not (memq sym block-globals)) ) )
+	 (write sym) 
+	 (newline) ) )
+     db) ) )
+
 
 ;;; Pretty-print expressions:
 
@@ -741,7 +783,7 @@
        (lambda (t next)
 	 (case t
 	   [(char unsigned-char) (if unsafe param `(##sys#foreign-char-argument ,param))]
-	   [(int unsigned-int short unsigned-short) (if unsafe param `(##sys#foreign-fixnum-argument ,param))]
+	   [(int unsigned-int short unsigned-short byte unsigned-byte) (if unsafe param `(##sys#foreign-fixnum-argument ,param))]
 	   [(float double) (if unsafe param `(##sys#foreign-flonum-argument ,param))]
 	   [(pointer byte-vector)
 	    (let ([tmp (gensym)])
@@ -800,16 +842,24 @@
 		   => (lambda (t)
 			(next (if (vector? t) (vector-ref t 0) t)) ) ]
 		  [(pair? t)
-		   (case (car t)
-		     [(ref pointer function c-pointer)
+		   (match t
+		     [((or 'ref 'pointer 'function 'c-pointer) . _)
 		      (let ([tmp (gensym)])
 			`(let ([,tmp ,param])
 			   (if ,tmp
 			       (##sys#foreign-pointer-argument ,tmp)
 			       '#f) ) )  ]
-		     [(nonnull-pointer nonnull-c-pointer)
+		     [('instance . _)
+		      (let ([tmp (gensym)])
+			`(let ([,tmp ,param])
+			   (if ,tmp
+			       (slot-ref ,param 'this)
+			       '#f) ) ) ]
+		     [('nonnull-instance . _)
+		      `(slot-ref ,param 'this) ]
+		     [((or 'nonnull-pointer 'nonnull-c-pointer) . _)
 		      `(##sys#foreign-pointer-argument ,param) ]
-		     [else param] ) ]
+		     [_ param] ) ]
 		  [else param] ) ] ) )
        (lambda () (quit "foreign type `~S' refers to itself" type)) ) ) ) )
 
@@ -848,10 +898,12 @@
    type
    (lambda (t next)
      (case t
-       ((char int short bool void unsigned-short scheme-object unsigned-char unsigned-int) 0)
-       ((c-string nonnull-c-string c-pointer nonnull-c-pointer unsigned-integer long integer unsigned-long
+       ((char int short bool void unsigned-short scheme-object unsigned-char unsigned-int byte unsigned-byte) 0)
+       ((c-string nonnull-c-string c-pointer nonnull-c-pointer
 		  c-string* nonnull-c-string*)
 	(words->bytes 3) )
+       ((unsigned-integer long integer unsigned-long)
+	(words->bytes 4) )
        ((float double) 
 	(words->bytes 4) )		; possibly 8-byte aligned 32-bit double
        (else
@@ -860,7 +912,9 @@
 		    (next (if (vector? t2) (vector-ref t2 0) t2)) ) ]
 	      [(pair? t)
 	       (case (car t)
-		 [(ref nonnull-pointer pointer c-pointer nonnull-c-pointer function) (words->bytes 3)]
+		 [(ref nonnull-pointer pointer c-pointer nonnull-c-pointer function instance
+		       nonnull-instance) 
+		  (words->bytes 3) ]
 		 [else 0] ) ]
 	      [else 0] ) ) ) )
    (lambda () (quit "foreign type `~S' refers to itself" type)) ) )
@@ -872,7 +926,7 @@
    type
    (lambda (t next)
      (case t
-       ((char int short bool unsigned-short unsigned-char unsigned-int long unsigned-long
+       ((char int short bool unsigned-short unsigned-char unsigned-int long unsigned-long byte unsigned-byte
 	      c-pointer pointer nonnull-c-pointer unsigned-integer integer float c-string
 	      nonnull-c-string c-string* nonnull-c-string*) 
 	(words->bytes 1) )
@@ -898,7 +952,13 @@
     [(nonnull-c-string) `(##sys#peek-nonnull-c-string ,body '0)]
     [(c-string*) `(##sys#peek-and-free-c-string ,body '0)]
     [(nonnull-c-string*) `(##sys#peek-and-free-nonnull-c-string ,body '0)]
-    [else body] ) )
+    [else
+     (match type
+       [('instance cname sname)
+	`(##tinyclos#make-instance-from-pointer ,body ,sname) ]
+       [('nonnull-instance cname sname)
+	`(make ,sname 'this ,body) ]
+       [_ body] ) ] ) )
 
 
 ;;; Scan expression-node for variable usage:
@@ -1018,8 +1078,9 @@
 
 ;;; Print version/usage information:
 
-(define (print-version)
-  (printf "This is ~a.~%~a~%~a" product build-information copyright) )
+(define (print-version . b)
+  (when (:optional b #f) (printf "~A~%" banner))
+  (printf "~A~%~A" build-information copyright) )
 
 (define (print-usage)
   (print-version)
@@ -1040,6 +1101,8 @@ Usage: chicken FILENAME OPTION ...
   File and pathname options:
 
     -output-file FILENAME       specifies output-filename, default is 'out.c'
+    -split NUMBER               split the output into smaller files
+    -split-level NUMBER         how hard the compiler should try partitioning the output
     -include-path PATHNAME      specifies alternative path for included files
     -to-stdout                  write compiled file to stdout instead of file
 
@@ -1047,22 +1110,23 @@ Usage: chicken FILENAME OPTION ...
 
     -feature SYMBOL             register feature identifier
     -no-feature SYMBOL          unregister feature identifier
-    -srfi-7                     process source file as SRFI-7 configuration
     -ffi-define SYMBOL          define preprocessor macro for ``easy'' FFI parser
+    -ffi-include-path PATH      set include path for ``easy'' FFI parser
+    -ffi                        compile C/C++ code and generate Scheme bindings
     -r5rs                       equivalent to `-hygienic -strict'
 
   Syntax related options:
 
     -case-insensitive           don't preserve case of read symbols
-    -hygienic                   use syntax-case macro package
-    -keyword-style STYLE        allow alternative keyword syntax (prefix or suffix)
+    -hygienic -syntax           use syntax-case macro package
+    -keyword-style STYLE        allow alternative keyword syntax (none, prefix or suffix)
     -strict                     disable non-standard macros and symbolsyntax
     -strict-srfi-0              disable non-standard macros except `cond-expand'
     -strict-reader              disable non-standard read syntax
     -strict-letrec              enable fully R5RS compliant `letrec'
-    -compile-time-macros-only   low-level macros are not made available at run-time
-    -hygienic-at-run-time       enable hygienic macros at run-time
-    -visit FILENAME             reads and macroexpands source file FILENAME
+    -run-time-macros            low-level macros are made available at run-time
+    -hygienic-at-run-time -syntax-at-run-time
+                                enable hygienic macros at run-time
 
   Translation options:
 
@@ -1075,7 +1139,7 @@ Usage: chicken FILENAME OPTION ...
     -no-warnings                disable warnings
     -debug-level NUMBER         set level of available debugging information
     -no-trace                   disable tracing information
-    -profile FILENAME           executable emits profiling information 
+    -profile                    executable emits profiling information 
 
   Optimization options:
 
@@ -1088,6 +1152,7 @@ Usage: chicken FILENAME OPTION ...
     -disable-interrupts         disable interrupts in compiled code
     -fixnum-arithmetic          assume all numbers are fixnums
     -benchmark-mode             fixnum mode, no interrupts and opt.-level 3
+    -disable-stack-overflow-checks  disables detection of stack-overflows.
 
   Configuration options:
 
@@ -1110,6 +1175,9 @@ Usage: chicken FILENAME OPTION ...
 
     -debug MODES                display debugging output for the given modes
     -compress-literals NUMBER   compile literals above threshold as strings
+    -disable-c-syntax-checks    disable syntax check of C code fragments
+    -unsafe-libraries           marks the generated file as being linked
+                                with the unsafe runtime system
 
 EOF
 ) )
@@ -1179,13 +1247,6 @@ EOF
 (define (string-null? x) (string-null? x))
 
 
-;;; Handle inline- and constant-definitions in visited files:
-
-(set! ##sys#visit-define-hook
-  (lambda (x)
-    (set! pending-canonicalizations (cons x pending-canonicalizations)) ) )
-
-
 ;;; Dump node structure:
 
 (define (dump-nodes n)
@@ -1208,21 +1269,64 @@ EOF
     (let ([old-hook ##sys#user-read-hook])
       (lambda (char port)
 	(if (char=? #\> char)	       
-	    (let ([_ (read-char port)])	; swallow #\>
+	    (let ([_ (read-char port)]		; swallow #\>
+		  [decl #f]
+		  [parse #f]
+		  [swig #f]
+		  [exec #f] )
 	      (case (peek-char port)
 		[(#\!)
 		 (read-char port)
-		 (let ([text (scan-sharp-greater-string port)])
-		   `(declare (foreign-declare ,text)
-			     (foreign-parse ,text) ) ) ]
+		 (set! parse #t)
+		 (set! decl #t) ]
 		[(#\?)
 		 (read-char port)
-		 (let ([text (scan-sharp-greater-string port)])
-		   `(declare (foreign-parse ,text)) ) ]
-		[else
-		 `(declare 
-		    (foreign-declare 
-		     ,(scan-sharp-greater-string port) ) ) ] ) )
+		 (set! parse #t) ]
+		[(#\:)
+		 (read-char port)
+		 (set! exec #t) ]
+		[(#\()
+		 (let ([head (read port)])
+		   (match head
+		     [('swig module args ...) (set! swig (cdr head))]
+		     [_ (for-each
+			 (lambda (t)
+			   (case t
+			     [(declare) (set! decl #t)]
+			     [(parse) (set! parse #t)]
+			     [(execute) (set! exec #t)]
+			     [else (quit "invalid tag in `#>(...) ...<#' form" t)] ) )
+			 head) ] ) ) ]
+		[else (set! decl #t)] )
+	      (let ([text (scan-sharp-greater-string port)])
+		(check-c-syntax text)
+		(when swig (set! no-c-syntax-checks #t))
+		`(begin
+		   ,@(if swig
+			 (let* ([mod (car swig)]
+				[modname (->string mod)]
+				[init (string-append modname "_swig_init")] )
+			   `((declare 
+			       (foreign-declare
+				,(sprintf "C_extern void ~A(int,C_word,C_word) C_noret;" init) ) 
+			       (custom-declare
+				(swig ,mod ,(string-append modname "_swig.i") ,@(cdr swig))
+				,text) )
+			     ((##core#primitive ,init)) ) )
+			 '() )
+		   ,@(if decl
+			 `((declare (foreign-declare ,text)))
+			 '() )
+		   ,@(if parse
+			 `((declare (foreign-parse ,text)))
+			 '() )
+		   ,@(if exec
+			 (let ([tmp (gensym 'code_)])
+			   `((declare 
+			       (foreign-declare
+				,(sprintf "static C_word ~A() { ~A; return C_SCHEME_UNDEFINED; }\n" tmp text) ) )
+			     (##core#inline ,tmp) ) )
+			 '() ) ) ) )
 	    (old-hook char port) ) ) ) ) )
 
 (define (scan-sharp-greater-string port)
@@ -1247,3 +1351,63 @@ EOF
 	       (loop) ] ) ) ) ) )
 
 (enable-sharp-greater-read-syntax)
+
+
+;;; Custom declarations:
+
+(define (process-custom-declaration spec strings)
+  (let* ([tag (car spec)]
+	 [name (cadr spec)]
+	 [fname (caddr spec)]
+	 [args (cdddr spec)] 
+	 [id (cons tag name)]
+	 [a (assoc id custom-declare-alist)] )
+    (unless a
+      (let ([out (open-output-file fname)])
+	(set! a (cons id out))
+	(set! custom-declare-alist (cons a custom-declare-alist))
+	(set! compiler-cleanup-hook
+	  (let ([old compiler-cleanup-hook])
+	    (lambda ()
+	      (close-output-port out)
+	      (old) ) ) )
+	(unless csc-control-file
+	  (set! csc-control-file (open-output-file (pathname-replace-extension source-filename "csc")))
+	  (display "#%csc\n" csc-control-file) 
+	  (set! compiler-cleanup-hook
+	    (let ([old compiler-cleanup-hook])
+	      (lambda ()
+		(close-output-port csc-control-file)
+		(old) ) ) ) )
+	(fprintf csc-control-file "~S~%" (cons* tag name fname args)) ) )
+    (for-each (cute display <> (cdr a)) strings) ) )
+
+
+;;; Compute real names for identifiers exported from modules:
+
+(define (compute-module-exports mexps exps)
+  (define (real-names syms)
+    (append-map
+     (lambda (sym)
+       (let ([sym2 (macroexpand sym)])
+	 (if (eq? sym2 sym)
+	     '()
+	     `((##core#set! ,sym ,sym2)) ) ) )
+     syms) )
+  (concatenate
+   (cons
+    (if exps 
+	(real-names exps)
+	'() )
+    (map (match-lambda
+	   [(mod . exps) 
+	    (macroexpand `(import ,mod))
+	    (real-names exps) ] )
+	 mexps) ) ) ) 
+
+
+;;; User-defined optimizers:
+
+(define (register-optimizer sym proc)
+  (set! optimizers-used #t)
+  (##sys#hash-table-set! user-level-optimizers sym proc) )
