@@ -1,6 +1,6 @@
 ;;;; tcp.scm - Networking stuff
 ;
-; Copyright (c) 2000-2003, Felix L. Winkelmann
+; Copyright (c) 2000-2004, Felix L. Winkelmann
 ; All rights reserved.
 ;
 ; Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following
@@ -28,8 +28,8 @@
 ; felix@call-with-current-continuation.org
 ;
 ; Felix L. Winkelmann
-; Steinweg 1A
-; 37130 Gleichen, OT Weissenborn
+; Unter den Gleichen 1
+; 37130 Gleichen
 ; Germany
 
 
@@ -43,8 +43,8 @@
 	  tcp-abandon-port tcp-listener-port)
   (bound-to-procedure
    ##net#socket ##net#bind ##net#connect ##net#listen ##net#accept 
-   ##net#close ##net#recv ##net#send ##net#select ##net#gethostaddr ##net#io-ports ##sys#update-errno
-   ##sys#error ##sys#signal-hook ##net#getservbyname ##net#parse-host
+   ##net#close ##net#recv ##net#send ##net#select ##net#select-write ##net#gethostaddr ##net#io-ports ##sys#update-errno
+   ##sys#error ##sys#signal-hook ##net#getservbyname ##net#parse-host ##net#fresh-addr
    ##net#bind-socket ##net#shutdown)
   (foreign-declare #<<EOF
 #include <errno.h>
@@ -104,6 +104,7 @@ EOF
 (define-foreign-variable _ipproto_tcp int "IPPROTO_TCP")
 (define-foreign-variable _invalid_socket int "INVALID_SOCKET")
 (define-foreign-variable _ewouldblock int "EWOULDBLOCK")
+(define-foreign-variable _einprogress int "EINPROGRESS")
 
 (define ##net#socket (foreign-lambda int "socket" int int int))
 (define ##net#bind (foreign-lambda int "bind" int pointer int))
@@ -178,6 +179,16 @@ EOF
      if(select(fd + 1, &in, NULL, NULL, &tm) == -1) return(-1);
      else return(FD_ISSET(fd, &in) ? 1 : 0);") )
 
+(define ##net#select-write
+  (foreign-lambda* int ((int fd))
+    "fd_set out;
+     struct timeval tm;
+     FD_ZERO(&out);
+     FD_SET(fd, &out);
+     tm.tv_sec = tm.tv_usec = 0;
+     if(select(fd + 1, NULL, &out, NULL, &tm) == -1) return(-1);
+     else return(FD_ISSET(fd, &out) ? 1 : 0);") )
+
 (define ##net#gethostaddr
   (foreign-lambda* bool ((pointer saddr) (c-string host) (unsigned-short port))
     "struct hostent *he = gethostbyname(host);"
@@ -211,46 +222,52 @@ EOF
 			    [p (##net#getservbyname s proto)] )
 		       (when (eq? 0 p)
 			 (##sys#update-errno)
-			 (##sys#signal-hook #:network-error "can not compute port from service" s) )
+			 (##sys#signal-hook #:network-error 'tcp-connect "can not compute port from service" s) )
 		       p) )
 		    (loop (fx+ i 1)) ) ) ) ) ) ) ) )
 
-(define (##net#bind-socket port style)
+(define ##net#fresh-addr
+  (foreign-lambda* void ((pointer saddr) (unsigned-short port))
+    "struct sockaddr_in *addr = (struct sockaddr_in *)saddr;"
+    "memset(addr, 0, sizeof(struct sockaddr_in));"
+    "addr->sin_family = AF_INET;"
+    "addr->sin_port = htons(port);"
+    "addr->sin_addr.s_addr = htonl(INADDR_ANY);") )
+
+(define (##net#bind-socket port style host)
   (##sys#check-exact port)
   (let ([s (##net#socket _af_inet style 0)])
     (when (eq? _invalid_socket s)
       (##sys#update-errno)
       (##sys#error "can not create socket") )
+    ;; PLT makes this an optional arg to tcp-listen. Should we as well?
     (when (eq? -1 ((foreign-lambda* int ((int socket)) 
 		     "int yes = 1; 
                       return(setsockopt(socket, SOL_SOCKET, SO_REUSEADDR, (const char *)&yes, sizeof(int)));") 
 		   s) )
       (##sys#update-errno)
-      (##sys#signal-hook #:network-error "error while setting up socket" s) )
+      (##sys#signal-hook #:network-error 'tcp-listen "error while setting up socket" s) )
     (let ([addr (make-string _sockaddr_in_size)])
-      ((foreign-lambda* void ((pointer saddr) (unsigned-short port))
-	 "struct sockaddr_in *addr = (struct sockaddr_in *)saddr;"
-	 "memset(addr, 0, sizeof(struct sockaddr_in));"
-	 "addr->sin_family = AF_INET;"
-	 "addr->sin_port = htons(port);"
-	 "addr->sin_addr.s_addr = htonl(INADDR_ANY);")
-       addr port)
+      (if host
+	  (unless (##net#gethostaddr addr host port)
+	    (##sys#signal-hook #:network-error 'tcp-listen "getting listener host IP failed" host port) )
+	  (##net#fresh-addr addr port) )
       (let ([b (##net#bind s addr _sockaddr_in_size)])
 	(when (eq? -1 b)
 	  (##sys#update-errno)
-	  (##sys#signal-hook #:network-error "can not bind to socket" s port) )
+	  (##sys#signal-hook #:network-error 'tcp-listen "can not bind to socket" s port) )
 	(values s addr) ) ) ) )
 
 (define-constant default-backlog 10)
 
-(define (tcp-listen port . wait)
-  (let-values ([(s addr) (##net#bind-socket port _sock_stream)])
-    (let ([w (if (pair? wait) (car wait) default-backlog)])
+(define (tcp-listen port . more)
+  (let-optionals more ([w default-backlog] [host #f])
+    (let-values ([(s addr) (##net#bind-socket port _sock_stream host)])
       (##sys#check-exact w)
       (let ([l (##net#listen s w)])
 	(when (eq? -1 l)
 	  (##sys#update-errno)
-	  (##sys#signal-hook #:network-error "can not listen on socket" s port) )
+	  (##sys#signal-hook #:network-error 'tcp-listen "can not listen on socket" s port) )
 	(##sys#make-structure 'tcp-listener s) ) ) ) )
 
 (define (tcp-listener? x) 
@@ -262,7 +279,7 @@ EOF
   (let ([s (##sys#slot tcpl 1)])
     (when (fx= -1 (##net#close s))
       (##sys#update-errno)
-      (##sys#signal-hook #:network-error "can not close TCP socket" tcpl) ) ) )
+      (##sys#signal-hook #:network-error 'tcp-close "can not close TCP socket" tcpl) ) ) )
 
 (define-constant buffer-size 1024)
 
@@ -355,7 +372,7 @@ EOF
   (let ([fd (##net#accept (##sys#slot tcpl 1) #f #f)])
     (when (eq? -1 fd)
       (##sys#update-errno)
-      (##sys#signal-hook #:network-error "could not accept from listener" tcpl) )
+      (##sys#signal-hook #:network-error 'tcp-accept "could not accept from listener" tcpl) )
     (##net#io-ports fd) ) )
 
 (define (tcp-accept-ready? tcpl)
@@ -363,7 +380,7 @@ EOF
   (let ([f (##net#select (##sys#slot tcpl 1))])
     (when (eq? -1 f)
       (##sys#update-errno)
-      (##sys#signal-hook #:network-error "can not check socket for input" tcpl) )
+      (##sys#signal-hook #:network-error 'tcp-accept-ready? "can not check socket for input" tcpl) )
     (eq? 1 f) ) )
 
 (define (tcp-connect host . more)
@@ -371,19 +388,32 @@ EOF
     (##sys#check-string host)
     (unless port
       (set!-values (host port) (##net#parse-host host "tcp"))
-      (unless port (##sys#signal-hook #:network-error "no port specified" host)) )
+      (unless port (##sys#signal-hook #:network-error 'tcp-connect "no port specified" host)) )
     (##sys#check-exact port)
     (let ([addr (make-string _sockaddr_in_size)]
 	  [s (##net#socket _af_inet _sock_stream 0)] )
-      (when (eq? -1 s)
-	(##sys#update-errno)
-	(##sys#signal-hook #:network-error "can not create socket" host port) )
-      (unless (##net#gethostaddr addr host port)
-	(##sys#signal-hook #:network-error "can not find host address" host) )
-      (when (eq? -1 (##net#connect s addr _sockaddr_in_size))
+      (define (fail)
 	(##net#close s)
 	(##sys#update-errno)
-	(##sys#signal-hook #:network-error "can not connect to socket" host port) )
+	(##sys#signal-hook #:network-error 'tcp-connect "can not connect to socket" host port) )
+      (when (eq? -1 s)
+	(##sys#update-errno)
+	(##sys#signal-hook #:network-error 'tcp-connect "can not create socket" host port) )
+      (unless (##net#gethostaddr addr host port)
+	(##sys#signal-hook #:network-error 'tcp-connect "can not find host address" host) )
+      (unless (##net#make-nonblocking s)
+	(##sys#update-errno)
+	(##sys#signal-hook #:network-error 'tcp-connect "fcntl() failed") )
+      (when (eq? -1 (##net#connect s addr _sockaddr_in_size))
+	(if (eq? errno _einprogress)
+	    (let loop ()
+	      (let ([f (##net#select-write s)])
+		(when (eq? f -1) (fail))
+		(unless (eq? f 1)
+		  ;(##sys#thread-block-for-i/o! ##sys#current-thread s #t)
+		  (yield)
+		  (loop) ) ) )
+	    (fail) ) )
       (##net#io-ports s) ) ) )
 
 (define (##sys#tcp-port->fileno p)
@@ -400,16 +430,16 @@ EOF
   (let ([fd (##sys#tcp-port->fileno p)])
     (values 
      (or (##net#getsockname fd)
-	 (##sys#signal-hook #:network-error "can not compute local address" p) )
+	 (##sys#signal-hook #:network-error 'tcp-addresses "can not compute local address" p) )
      (or (##net#getpeername fd)
-	 (##sys#signal-hook #:network-error "can not compute remote address" p) ) ) ) )
+	 (##sys#signal-hook #:network-error 'tcp-addresses "can not compute remote address" p) ) ) ) )
 
 (define (tcp-listener-port tcpl)
   (##sys#check-structure tcpl 'tcp-listener)
   (let* ([fd (##sys#slot tcpl 1)]
 	 [port (##net#getsockport fd)] )
     (when (eq? -1 port)
-      (##sys#signal-hook #:network-error "can not obtain listener port" tcpl fd) )
+      (##sys#signal-hook #:network-error 'tcp-listener-port "can not obtain listener port" tcpl fd) )
     port) )
 
 (define (tcp-abandon-port p)
