@@ -31,7 +31,7 @@
   %allocate-instance %allocate-entity get-field set-field! lookup-slot-info the-slots-of-a-class
   getters-n-setters-for-class generic-invocation-generics applicable? more-specific? 
   make-primitive-class entity-tag method-cache-tag method-cache-lookup make-method-cache
-  hash-arg-list method-caching-enabled 
+  hash-arg-list method-caching-enabled make-instance-from-pointer
   quick-getl symbol-vector}
 
 
@@ -45,7 +45,7 @@
   (disable-interrupts)
   (hide every1 every2 every2-eq? method-caching-enabled method-cache-tag symbol-vector ##tinyclos#slot-ref
 	symbol-vector getl make-method-cache set-field! method-cache-lookup %allocate-entity
-	hash-arg-list std-tie-breaker build-constraints %allocate-instance compute-std-cpl
+	hash-arg-list std-tie-breaker build-constraints %allocate-instance 
 	get-field ensure-generic ##tinyclos#slot-set! lookup-slot-info
 	getters-n-setters-for-class the-slots-of-a-class make-primitive-class quick-getl
 	build-transitive-closure applicable? more-specific? method-cache-tag top-sort
@@ -171,6 +171,7 @@
 (define (every2 test list1 list2)
   (let loop ([list1 list1] [list2 list2])
     (or (null? list1)
+	(null? list2)
 	(and (test (##sys#slot list1 0) (##sys#slot list2 0))
 	     (loop (##sys#slot list1 1) (##sys#slot list2 1)) ) ) ) )
 
@@ -506,7 +507,7 @@
 	[(symbol? x)     <symbol>]
 	[(vector? x)     <vector>]
 	[(pair? x)       <pair>]
-	[(number? x)     <inexact>]
+	[(number? x)     (if (integer? x) <integer> <inexact>)]
 	[(string? x)     <string>]
 	[(procedure? x)  <procedure>]
 	[(port? x) (if (input-port? x) <input-port> <output-port>)]
@@ -1203,7 +1204,8 @@
 (define <vector>    (make-primitive-class "vector"))
 (define <pair>      (make-primitive-class "pair"))
 (define <number>    (make-primitive-class "number"))
-(define <exact>    (make-primitive-class "exact" <number>))
+(define <integer>    (make-primitive-class "integer" <number>))
+(define <exact>    (make-primitive-class "exact" <integer>))
 (define <inexact>    (make-primitive-class "inexact" <number>))
 (define <string>    (make-primitive-class "string"))
 (define <port>      (make-primitive-class "port"))
@@ -1317,3 +1319,8 @@
 
 (define (subclass? x y)
   (if (memq y (compute-cpl x)) #t #f) )
+
+(define (make-instance-from-pointer ptr cls)
+  (and ptr 
+       (not (##sys#null-pointer? ptr))
+       (make cls 'this ptr) ) )
